#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta


from __future__ import with_statement

# FIXME raise InternalError()などの()不要?
# FIXME 同名だが片方ファイル、片方ディレクトリの場合を未考慮
# FIXME DataSyncはクラスである意味がない。
#   コンストラクタでsrc,dstなど受けとればまだ意味がある。
# FIXME 一時ファイル実装、dropboxとamazonも対応すること

import os
import re
import stat
import shutil
import errno
import locale
try:
    import email.utils as emailut
except:
    import email.Utils as emailut

import smbc

try:
    from dropbox import session, client, rest
except:
    session = None
    client = None
    rest = None

try:
    from boto.s3.connection import S3Connection
    from boto.s3.key import Key
except:
    S3Connection = None
    Key = None


DIRS_NAME = "dirs"
FILES_NAME = "files"
MTIME_NAME = "mtime"
SIZE_NAME = "size"
CHILDREN_NAME = "children"
STATE_NAME = "state"

STATE_NEW = "new"
STATE_DEL = "del"
STATE_SAME = "same"

MODE_READ = "read"
MODE_WRITE = "write"

ST_MODE_DIR = 16877
ST_MODE_FILE = 33188

MODE_CH_FILE = (stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP |
            stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)
MODE_CH_DIR = (stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)

BUF_SIZE = 65536 # FIXME 適当

MAIN_SEP = "/"

ABORT = "Abort!!"
FILE_TRANSFER_OK = "File Transfer OK"
FILE_TRANSFER_NG = "File Transfer OK"
FILE_DELETE_OK = "File Delete OK"
FILE_DELETE_NG = "File Delete NG"
DIR_CREATE_OK = "Directory Create OK"
DIR_CREATE_NG = "Directory Create OK"
DIR_DELETE_OK = "Directory Delete OK"
DIR_DELETE_NG = "Directory Delete NG"

LOG_FMT_STR = "%s -> %s.\n"

WRITE_TEMP_FILE_NAME = ".tmp_file_for_Copy"

# For local
LOCAL_SEP = "/"

# For smb
SMBC_TYPE_DIR = 7
SMB_SEP = "/"

# For dropbox
APP_KEY = "av49kj4lr28xk65"
APP_SECRET = "irh68ok5h6hqbar"
ACCESS_TYPE = 'dropbox'
DEFAULT_LOCALE = "UTF8"
DROPBOX_ENCODE = "UTF8"
CONTENTS_NAME = "contents"
PATH_NAME = "path"
MODIFIED_NAME = "modified"
BYTES_NAME = "bytes"
IS_DIR_NAME = "is_dir"
STATUS_BAD_INPUT_PARAM = 400 # FIXME どうつかうか
STATUS_ALREADY_EXISTS = 403
STATUS_NOT_FOUND = 404
STATUS_NO_SPACE = 507
DROPBOX_SEP = "/"

# For amazons3
AMAZONS3_ENCODE = "UTF8"
AMAZONS3_SEP = "/"


# FIXME for debug
def DEBUG(string):
    #print "######################## " + string + "\n"
    open("/tmp/1", "a").write("### in DataSync: " + string + "\n")


class InternalError(Exception):
    pass


class NonDiffException(Exception):
    pass


class NoEntryError(Exception):
    pass


class NoSpaceError(Exception):
    pass


class ExistsError(Exception):
    pass


class PermissionError(Exception):
    pass



class Src():
    """Base class for the source file tree."""

    def __init__(self, root, ignored_list=[], log_path="", base=""):
        """Constructer."""

        self.root = root
        self.ignored_list = ignored_list
        self.log_path = log_path

    # FIXME get_root restrictedで使うがfullではどうつかう?
    def get_root(self):
        """Return the source root."""

        return self.root 

    def get_tree(self):
        """Return the all file tree."""

        raise NotImplementedError

    def is_delta_tree_supported(self):
        """Return True if "get_delta_tree" method is suppored."""

        return False

    def get_delta_tree(self):
        """Return the changed file tree after this method is called last."""

        raise NotImplementedError

    def init_tree(self):
    
        raise NotImplementedError

class Dst:
    """Base class for the destination file tree."""

    def __init__(self, root, log_path):

        DEBUG("Dst constracter")
        self.root = root
        self.log_path = log_path

    def get_root(self):
        """Return the destination root."""

        return self.root 

    def update(self, src):

        raise NotImplementedError


class SrcFull(Src):
    """Class for the source file tree supporting full access."""

    def __init__(self, path, ignored_list=[], log_path=""):

        self.tree = None
        Src.__init__(self, path, ignored_list, log_path)

    def __set_update(self, tree, path, name):

        is_updated = False

        abspath = self.join_path([path, name])
        path_stat = self.stat(abspath)
        if stat.S_ISDIR(path_stat[stat.ST_MODE]):
            if not name in tree[DIRS_NAME]:
                tree[DIRS_NAME][name] = {
                    CHILDREN_NAME: {DIRS_NAME: {}, FILES_NAME: {}},
                    STATE_NAME: STATE_NEW}
                is_updated = True
            else:
                tree[DIRS_NAME][name][STATE_NAME] = STATE_SAME
        else:
            if not name in tree[FILES_NAME]:
                tree[FILES_NAME][name] = {
                MTIME_NAME: int(path_stat[stat.ST_MTIME]),
                SIZE_NAME: path_stat[stat.ST_SIZE],
                STATE_NAME: STATE_NEW}
                is_updated = True
            else:
                last_info = tree[FILES_NAME][name]
                new_info = {
                    MTIME_NAME: int(path_stat[stat.ST_MTIME]),
                    SIZE_NAME: path_stat[stat.ST_SIZE],
                    STATE_NAME: STATE_NEW}
                if ((tree[FILES_NAME][name][SIZE_NAME] != new_info[SIZE_NAME]) or
                    (tree[FILES_NAME][name][MTIME_NAME] != new_info[MTIME_NAME])):
                    DEBUG("debug 13")
        
                    tree[FILES_NAME][name][SIZE_NAME] = new_info[SIZE_NAME]
                    tree[FILES_NAME][name][MTIME_NAME] = new_info[MTIME_NAME]
                    tree[FILES_NAME][name][STATE_NAME] = STATE_NEW
                    is_updated = True
                else:
                    tree[FILES_NAME][name][STATE_NAME] = STATE_SAME

        return is_updated

    def __set_delete(self, tree, path):

        is_updated = False

        for file in tree[FILES_NAME].keys():
            if (not file in self.listdir(path) and
                tree[FILES_NAME][file][STATE_NAME] != STATE_DEL):
                is_updated = True
                tree[FILES_NAME][file][STATE_NAME] = STATE_DEL

        for dir in tree[DIRS_NAME].keys():
            if (not dir in self.listdir(path) and
                tree[DIRS_NAME][dir][STATE_NAME] != STATE_DEL):
                is_updated = True
                tree[DIRS_NAME][dir][STATE_NAME] = STATE_DEL

        return is_updated

    def __set_delta_tree(self, tree, path):

        is_updated = False

        for name in self.listdir(path):
            DEBUG("debug 8 listdir " + name + " in " + path)
            try:
                if name in self.ignored_list:
                    continue
            except:
                raise
    
            if self.__set_update(tree, path, name):
                is_updated = True

        if self.__set_delete(tree, path):
            is_updated = True

        for dir in tree[DIRS_NAME].keys():
            if tree[DIRS_NAME][dir][STATE_NAME] == STATE_DEL:
                continue
            if self.__set_delta_tree(
                tree[DIRS_NAME][dir][CHILDREN_NAME],
                self.join_path([path, dir])):
                is_updated = True

        return is_updated
 
    def get_tree(self, tree=None, path=None):
        """Return the all file tree."""

        if not tree:
            tree = {DIRS_NAME: {}, FILES_NAME: {}}

        if not path:
            path = self.root

        for name in self.listdir(path):
            DEBUG("debug 8 listdir " + name + " in " + path)
            try:
                if name in self.ignored_list:
                    continue
            except:
                raise
    
            abspath = self.join_path([path, name])
            path_stat = self.stat(abspath)
            if stat.S_ISDIR(path_stat[stat.ST_MODE]):
                tree[DIRS_NAME][name] = {
                    CHILDREN_NAME: {DIRS_NAME: {}, FILES_NAME: {}}}
                self.get_tree(tree[DIRS_NAME][name][CHILDREN_NAME], abspath)
            else:
                tree[FILES_NAME][name] = {
                    MTIME_NAME: int(path_stat[stat.ST_MTIME]),
                    SIZE_NAME: path_stat[stat.ST_SIZE]}
        return tree

    def set_delta_tree(self):

        return self.__set_delta_tree(self.tree, self.root)

    def join_path(self, path_list):
        """Join path."""
        
        raise NotImplementedError

    def listdir(self, path):
        """Return List dir of "path"."""

        raise NotImplementedError

    def isdir(self, entry): # FIXME 高速化するなら必要かも
        """Is dir."""
        
        raise NotImplementedError

    def stat(self, path):
        """Return status of "path"."""
        
        raise NotImplementedError

    def open(self, path, mode):
        """Return file obj of "path"."""
        
        raise NotImplementedError


class DstFull(Dst):

    def __init__(self, root, log_path):

        Dst.__init__(self, root, log_path)

    def __logging(self, relpath, name, info_list):

        file_path = relpath + [name]
        if self.log_path:
            with open(self.log_path, "a") as log_file:
                log_file.write(LOG_FMT_STR % (
                    "/".join(file_path), ", ".join(info_list)))
    
    def __tmp_file_remove(self, tmp_dst_path):

        DEBUG("tmp_file_remove 1")
        try:
            self.remove(tmp_dst_path)
        except NoEntryError:
            pass
        DEBUG("tmp_file_remove 2")

    def __exec_file_transfer(self, relpath, file, file_info):
        is_executed = False

        if file_info[STATE_NAME] != STATE_NEW:
            return is_executed

        src_path = self.src.join_path([self.src.root] + relpath + [file])
        dst_path = self.join_path([self.root] + relpath + [file])
        tmp_dst_path = self.join_path([self.root, WRITE_TEMP_FILE_NAME])

        self.__tmp_file_remove(tmp_dst_path)

        DEBUG("src_path = " + str(src_path))
        DEBUG("dst_path = " + str(dst_path))

        try:
            src_file = self.src.open(src_path, MODE_READ)
        except NoEntryError:
            file_info[STATE_NAME] = STATE_DEL
            return is_executed

        tmp_dst_file = self.open(tmp_dst_path, MODE_WRITE)
        
        # FIXME NOSPACEエラーなどの処理どうする
        try:
            try:
                tmp_dst_file.write_by_path(src_file)
            except NotImplementedError:
                try:
                    DEBUG("tmp_dst_file.get_path() = " + str(tmp_dst_file.get_path()))
                    src_file.read_by_path(tmp_dst_file)
                #except NotImplementedError:
                except: # FIXME readdirectでNoSpaceError検出できない欠陥をここのNoSpaceErrorで吸収している
                    tmp_dst_file.write(src_file)
        except NoSpaceError:
            src_file.close()
            tmp_dst_file.close()
            self.remove(tmp_dst_path)
            raise

        src_file.close()
        tmp_dst_file.close()

        DEBUG("try_file_transfer file dst_path: " + str(dst_path))
        self.chmod(tmp_dst_path, MODE_CH_FILE)

        DEBUG("try_file_transfer 1")
        try:
            self.utime(tmp_dst_path, file_info[MTIME_NAME])
            DEBUG("try_file_transfer 2")
        except NotImplementedError:
            # COMMENT dropboxのようにdst側のutimeができない場合の処理。将来不要になるかも
            # 現時点ではdropbox -> dropboxは時刻同期不可ということ
            DEBUG("try_file_transfer 3")
            dst_mtime = self.stat(tmp_dst_path)[stat.ST_MTIME]
            DEBUG("try_file_transfer 4")
            self.src.utime(src_path, dst_mtime)
            DEBUG("try_file_transfer 5")
            file_info[MTIME_NAME] = dst_mtime

        DEBUG("try_file_transfer file_info: " + str(file_info))
        file_info[STATE_NAME] = STATE_SAME
       
        self.move(tmp_dst_path, dst_path)

        is_executed = True

        return is_executed

    def __exec_file_delete(self, relpath, file, file_info):
        is_executed = False

        if file_info[STATE_NAME] == STATE_NEW:
            return is_executed


        if file_info[STATE_NAME] == STATE_DEL:
            dst_path = self.join_path([self.root] + relpath + [file])

            is_executed = True

            try:
                DEBUG("debug 4 dst file delete " + dst_path)
                self.remove(dst_path)
            except NoEntryError:
                pass

        return is_executed

    def __exec_dir_create(self, relpath, dir, dir_info):
        is_executed = False

        if dir_info[STATE_NAME] == STATE_NEW:
            dst_path = self.join_path([self.root] + relpath + [dir])

            is_executed = True

            try:
                DEBUG("debug 5 dst dir create " + dst_path)
                self.mkdir(dst_path)
            except ExistsError:
                pass

            dir_info[STATE_NAME] = STATE_SAME

        return is_executed

    def __exec_dir_delete(self, relpath, dir, dir_info):
        is_executed = False

        if dir_info[STATE_NAME] == STATE_NEW:
            return is_executed

        is_executed = False

        if dir_info[STATE_NAME] == STATE_DEL:
            dst_path = self.join_path([self.root] + relpath + [dir])
            is_executed = True
            try:
                DEBUG("debug 7 dst dir delete " + dst_path)
                self.rmdir(dst_path)
            except NoEntryError:
                pass
        
        return is_executed

    def __try_file_transfer(self, relpath, file, file_info):
        is_executed = False
        try:
            is_executed = self.__exec_file_transfer(relpath, file, file_info)
            if is_executed:
                self.__logging(relpath, file, [FILE_TRANSFER_OK])
        except NoSpaceError: 
            # FIXME 処理中断する例外
            self.__logging(relpath, file, [FILE_TRANSFER_NG, ABORT])
            raise
        except:
            # FIXME 処理中断しない例外
            # アップロードサイズ制限などもこっちに含める
            self.__logging(relpath, file, [FILE_TRANSFER_NG])
            pass
        return is_executed

    def __try_file_delete(self, relpath, file, file_info):
        is_executed = False
        try:
            is_executed = self.__exec_file_delete(relpath, file, file_info)
            if is_executed:
                self.__logging(relpath, file, [FILE_DELETE_OK])
        except:
            # FIXME 処理中断しない例外
            self.__logging(relpath, file, [FILE_DELETE_NG])
            pass
        return is_executed

    def __try_dir_create(self, relpath, dir, dir_info):
        is_executed = False
        try:
            is_executed = self.__exec_dir_create(relpath, dir, dir_info)
            if is_executed:
                self.__logging(relpath, dir, [DIR_CREATE_OK])
        except NoSpaceError: 
            # FIXME 処理中断する例外
            self.__logging(relpath, dir, [DIR_CREATE_NG, ABORT])
            raise
        except:
            # FIXME 処理中断しない例外
            self.__logging(relpath, dir, [DIR_CREATE_NG])
            pass
        return is_executed

    def __try_dir_delete(self, relpath, dir, dir_info):
        is_executed = False
        try:
            is_executed = self.__exec_dir_delete(relpath, dir, dir_info)
            if is_executed:
                self.__logging(relpath, dir, [DIR_DELETE_OK])
        except:
            # FIXME 処理中断しない例外
            self.__logging(relpath, dir, [DIR_DELETE_NG])
            pass
        return is_executed

    def __update(self, relpath, tree):
        for file in tree[FILES_NAME].keys():
            file_info = tree[FILES_NAME][file]
            
            self.__try_file_transfer(relpath, file, file_info)
            # COMMENT nasdsyncはここでdropboxのアップロード制限エラーの場合、
            # ログ残している。今回は下位で受け取りSkipErrorに抽象化して
            # 下のSkipErrorでうけとればよいかと
            if self.__try_file_delete(relpath, file, file_info):
                del tree[FILES_NAME][file]

        for dir in tree[DIRS_NAME].keys():
            dir_info = tree[DIRS_NAME][dir]

            self.__try_dir_create(relpath, dir, dir_info)
            if self.__try_dir_delete(relpath, dir, dir_info):
                del tree[DIRS_NAME][dir]
            else:
                self.__update(relpath + [dir],
                    tree[DIRS_NAME][dir][CHILDREN_NAME])

    def update(self, src):
        self.src = src
        self.__update([], src.tree)

    def mkdir(self, path):
        """Create a directory specified by "path"."""

        raise NotImplementedError
        
    def stat(self, path):

        raise NotImplementedError

    def chmod(self, path, mode):

        raise NotImplementedError

    def utime(self, path, mtime):

        raise NotImplementedError

    def remove(self, path):

        raise NotImplementedError

    def move(self, path):

        raise NotImplementedError

    def rmdir(self, path):
        
        raise NotImplementedError

    def open(self, path, mode):

        raise NotImplementedError

    def join_path(self, path):
        
        raise NotImplementedError


class FileLikeObj():

    def __init__(self, file, path):

        self.file = file
        self.path = path
        self.point = 0

    def __len__(self):

        raise NotImplementedError

    def tell(self):

        raise NotImplementedError

    def get_path(self):

        return self.path

    def read(self):

        raise NotImplementedError

    def write(self, file_like):

        raise NotImplementedError
 
    def write_by_path(self, file_like):

        raise NotImplementedError

    def read_by_path(self, file_like):

        raise NotImplementedError

    def close(self):
        
        raise NotImplementedError


class LocalFileObj(FileLikeObj):

    def __len__(self):

        return os.fstat(self.file.fileno()).st_size

    def tell(self):

        return self.file.tell()

    def seek(self, offset, whence=os.SEEK_SET):

        self.file.seek(offset, whence)

    def get_path(self):

        return self.path

    def read(self, size):

        return self.file.read(size)

    def write(self, file_like):

        while True:
            buf = file_like.read(BUF_SIZE)
            if not buf:
                break
            try:
                self.file.write(buf)
            except (IOError), err:
                if err.errno == errno.ENOSPC:
                    raise NoSpaceError()
                raise

    def close(self):

        try:
            self.file.close()
        except (IOError), err:
            if err.errno == errno.ENOSPC:
                raise NoSpaceError()
            raise
        

# FIXME システムにてos.listdir()などでとれる文字列はutf-8として限定し,
# 内部ツリーにはutf-8として保持する。
class LocalAccess():

    def __init__(self):
        self.encoding = locale.getdefaultlocale()[1]
        
    def listdir(self, path):
        try:
            return os.listdir(path)
        except (OSError), err:
            if err.errno == errno.ENOENT:
                raise NoEntryError()
            raise
      
    def open(self, path, mode):
        if mode == MODE_READ:
            try:
                return LocalFileObj(open(path, "r"), path)
            except (OSError), err:
                if err.errno == errno.ENOENT:
                    raise NoEntryError()
                raise

        elif mode == MODE_WRITE:
            try:
                return LocalFileObj(open(path, "w"), path)
            except (IOError), err:
                if err.errno == errno.ENOSPC:
                    raise NoSpaceError()
                raise

    def mkdir(self, path):
        try:
            os.mkdir(path)
        except (OSError), err:
            if err.errno == errno.EEXIST:
                raise ExistsError()
            if err.errno == errno.ENOSPC:
                # COMMENT NoSpaceでdirができる場合はrmdirすること
                raise NoSpaceError()
            raise

        self.chmod(path, MODE_CH_DIR)

        return 
        
    def stat(self, path):
        return os.stat(path)

    def chmod(self, path, mode):
        os.chmod(path, mode)

    def utime(self, path, mtime):
        if not mtime:
            raise InternalError()
        os.utime(path, (mtime, mtime)) # COMMENT nasdsyncでも(mtime, mtime)だった

    def remove(self, path):
        try:
            os.remove(path)
        except (OSError), err:
            if err.errno == errno.ENOENT:
                raise NoEntryError()
            raise

    def rmdir(self, path):
        try:
            DEBUG("rmdir")
            shutil.rmtree(path)
        except (OSError), err:
            if err.errno == errno.ENOENT:
                raise NoEntryError()
            raise

    def move(self, src, dst):
        shutil.move(src, dst)

    def join_path(self, path_list):
        return LOCAL_SEP.join(path_list)


class LocalFileTree(SrcFull, DstFull):

    def __init__(self, path, ignored_list=[], log_path=""):
        self.local_access = LocalAccess()
        SrcFull.__init__(self, path, ignored_list)
        DEBUG("LocalFileTree constracter")
        DstFull.__init__(self, path, log_path)

    def listdir(self, path):
        return self.local_access.listdir(path)

    def open(self, path, mode):
        return self.local_access.open(path, mode)

    def mkdir(self, path):
        self.local_access.mkdir(path)
        
    def stat(self, path):
        return self.local_access.stat(path)

    def chmod(self, path, mode):
        self.local_access.chmod(path, mode)

    def utime(self, path, mtime):
        self.local_access.utime(path, mtime)

    def remove(self, path):
        self.local_access.remove(path)

    def rmdir(self, path):
        self.local_access.rmdir(path)

    def move(self, src, dst):
        self.local_access.move(src, dst)

    def join_path(self, path_list):
        return self.local_access.join_path(path_list)


class SmbFileObj(FileLikeObj):

    def __len__(self):

        # FIXME len()はint型をかえす必要あり。巨大size大丈夫か
        return int(self.file.fstat()[stat.ST_SIZE])

    def tell(self):

        # FIXME amazon用。read完了しても0にならないが問題あるか?
        # とりあえずNotImplement
        #return self.point
        raise NotImplementedError

    def seek(self, offset, whence=os.SEEK_SET):

        # FIXME エラー処理どうする
        # writeの場合のpointはサポートしないがよいか
        # とりあえずNotImplement
        #if whence == os.SEEK_SET:
        #    self.point = offset
        #elif whence == os.SEEK_CUR:
        #    self.point += offset
        #elif whence == os.SEEK_END:
        #    self.point = len(self) - offset         
        raise NotImplementedError

    def get_path(self):

        # FIXME get_pathがNotImpleであることによって
        # writedirect,readdirectをつかえないことになっている
        # このような実装でよいのか
        raise NotImplementedError

    def read(self, size):

        self.point += size
        return self.file.read(size)

    def write(self, file_like):

        while True:
            buf = file_like.read(BUF_SIZE)
            if not buf:
                break
            try:
                self.file.write(buf)
            except smbc.NoSpaceError:
                raise NoSpaceError()

    def write_by_path(self, file_like):

        try:
            # FIXME 一度に書いて大丈夫か?
            self.file.writedirect(file_like.get_path())
        except smbc.NoSpaceError:
            raise NoSpaceError()

    def read_by_path(self, file_like):

        try:
            # FIXME 一度に書いて大丈夫か?
            self.file.readdirect(file_like.get_path())
        except (IOError), err:
            if err.errno == errno.ENOSPC:
                raise NoSpaceError()
            raise

    def close(self):

        try:
            self.file.close()
        except (IOError), err:
            if err.errno == errno.ENOSPC:
                raise NoSpaceError()
            raise

class SmbAccess():

    def __init__(self, user, passwd):
        auth_fn = lambda srv, share, wg, username, pw: (wg, user, passwd)
        self.__ctx = smbc.Context(auth_fn=auth_fn)
        self.encoding = locale.getdefaultlocale()[1]
        if not self.encoding:
            self.encoding = DEFAULT_LOCALE
        # FIXME インスタンスコピーの仕組みどうする。一見Fullには不要そうだが

    def listdir(self, path):
        try:
            return [
                entry.name
                for entry               
                in self.__ctx.opendir(path).getdents()
                if entry.name != "." and entry.name != ".."]
        except smbc.PermissionError:
            raise PermissionError
        
    def isdir(self, entry): # FIXME 高速化するなら必要かも
        pass

    def open(self, path, mode):
        if mode == MODE_READ:
            try:
                return SmbFileObj(self.__ctx.open(path, os.O_RDONLY), path)
            except smbc.NoEntryError:
                raise NoEntryError()

        if mode == MODE_WRITE:
            try:
                return SmbFileObj(self.__ctx.open(path, os.O_WRONLY|os.O_CREAT|os.O_TRUNC), path)
            except smbc.NoSpaceError:
                raise NoSpaceError()

    def stat(self, path):
        return self.__ctx.stat(path)

    def chmod(self, path, mode):
        self.__ctx.chmod(path, mode)

    def utime(self, path, mtime):
        if not mtime:
            raise InternalError()
        DEBUG("smb utime path: " + str(path))
        DEBUG("smb utime mtime: " + str(mtime))
        self.__ctx.utime(path, (mtime, mtime)) # COMMENT nasdsyncでも(mtime, mtime)だったのでOKかと
        DEBUG("smb utime end")

    def remove(self, path):
        try:
            self.__ctx.unlink(path)
        except smbc.NoEntryError:
            raise NoEntryError()

    def rmdir(self, path):
        try:
            self.__ctx.rmdir(path)
        except smbc.NoEntryError:
            raise NoEntryError()
        except smbc.NotEmptyError:
            for entry in self.__ctx.opendir(path).getdents():
                if entry.name == "." or entry.name == "..":
                    continue
                entry_path = os.path.join(path, entry.name)
                if entry.smbc_type != SMBC_TYPE_DIR: # is not Directory
                    self.__ctx.unlink(entry_path)
                else: # is Directory
                    self.rmdir(entry_path)
            self.__ctx.rmdir(path)

    def mkdir(self, path):
        try:
            self.__ctx.mkdir(path)
        except smbc.ExistsError:
            raise ExistsError()
        except smbc.NoSpaceError:
            # COMMENT NoSpaceでdirができる場合はrmdirすること
            raise NoSpaceError()

        self.chmod(path, MODE_CH_DIR)

        return 

    def move(self, src, dst):
        self.__ctx.rename(src, dst)

    def join_path(self, path_list):
        # FIXME path_listの内容はutf-8として限定
        return SMB_SEP.join(path_list)


class SmbFileTree(SrcFull, DstFull):
    """Class for the file tree accessed via SMB protocol.""" 

    def __init__(self, path, user, passwd, ignored_list=[], log_path=""):
        self.smb_access = SmbAccess(user, passwd)
        SrcFull.__init__(self, path, ignored_list)
        DstFull.__init__(self, path, log_path)
        # FIXME 認証どうする。
        # FIXME インスタンスコピーの仕組みどうする。一見Fullには不要そうだが

    def listdir(self, path):
        return self.smb_access.listdir(path)

    def open(self, path, mode):
        DEBUG("smb open path = " + str(path))
        DEBUG("smb open mode = " + str(mode))
        return self.smb_access.open(path, mode)

    def mkdir(self, path):
        self.smb_access.mkdir(path)
        
    def stat(self, path):
        return self.smb_access.stat(path)

    def chmod(self, path, mode):
        self.smb_access.chmod(path, mode)

    def utime(self, path, mtime):
        self.smb_access.utime(path, mtime)

    def remove(self, path):
        self.smb_access.remove(path)

    def rmdir(self, path):
        self.smb_access.rmdir(path)

    def move(self, src, dst):
        self.smb_access.move(src, dst)

    def join_path(self, path_list):
        return self.smb_access.join_path(path_list)


class DropboxFileObj(FileLikeObj):

    def __init__(self, file, path, dclient):

        self.__dclient = dclient
        FileLikeObj.__init__(self, file, path)

    def __len__(self):

        return self.file.fstat()[stat.ST_SIZE]

    def tell(self):

        # FIXME amazon用。read完了しても0にならないが問題あるか?
        # とりあえずNotImplement
        #return self.point
        raise NotImplementedError

    def seek(self, offset, whence=os.SEEK_SET):

        # FIXME エラー処理どうする
        # writeの場合のpointはサポートしないがよいか
        # とりあえずNotImplement
        #if whence == os.SEEK_SET:
        #    self.point = offset
        #elif whence == os.SEEK_CUR:
        #    self.point += offset
        #elif whence == os.SEEK_END:
        #    self.point = len(self) - offset         
        raise NotImplementedError

    def get_path(self):

        raise NotImplementedError

    def read(self, size):

        self.point += size
        return self.file.read(size)

    def write_by_path(self, file_like):

        try:
            # FIXME 一度に書いて大丈夫か?
            DEBUG("file_like len = " + str(len(file_like)))
            self.__dclient.put_file(self.path, file_like, True) # FIXME True詳細不明
        except (rest.ErrorResponse), err:
            if err.status == STATUS_NO_SPACE:
                # FIXME nasdsyncは作成途中ファイル削除していない。
                # API内部でやってくれてる?
                raise NoSpaceError()
            raise

    def close(self):
        pass
        # FIXME 未実装


class DropboxAccess():

    def __init__(self, key, secret):
        sess = session.DropboxSession(APP_KEY, APP_SECRET, ACCESS_TYPE)
        sess.set_token(key, secret)
        self.__dclient = client.DropboxClient(sess)
        self.encoding = locale.getdefaultlocale()[1]
        if not self.encoding:
            self.encoding = DEFAULT_LOCALE
 
    def listdir(self, path):
        metadata = self.__dclient.metadata(path)
        return [
            os.path.basename(content[PATH_NAME].encode(DROPBOX_ENCODE))
            for content
            in metadata.get((CONTENTS_NAME), [])] # FIXME []の詳細不明
        
    def isdir(self, entry): # FIXME 高速化するなら必要かも
        pass

    def open(self, path, mode):
        if mode == MODE_READ:
            try:
                return DropboxFileObj(self.__dclient.get_file(path), path, self.__dclient)
            except (rest.ErrorResponse), err:
                if err.status == STATUS_NOT_FOUND:
                    raise NoEntryError()
                raise

        if mode == MODE_WRITE:
            try:
                return DropboxFileObj("", path, self.__dclient)
            except (rest.ErrorResponse), err:
                if err.status == STATUS_NO_SPACE:
                    # FIXME nasdsyncは作成途中ファイル削除していない。
                    # API内部でやってくれてる?
                    raise NoSpaceError()
                raise

    def stat(self, path):
        metadata = self.__dclient.metadata(path)
        if metadata.get(IS_DIR_NAME):
            st_mode = ST_MODE_DIR
        else:
            st_mode = ST_MODE_FILE

        st_mtime = emailut.mktime_tz(
            emailut.parsedate_tz(
                metadata.get(MODIFIED_NAME)))

        # FIXME 時刻のタイムゾーンどうする。nasdsyncにならうか
        # 国際化対応するなら、各国のタイムゾーンに対応するために
        # pytzやdatetimeなどのライブラリ必要
        return {
            stat.ST_MODE: st_mode,
            stat.ST_MTIME: st_mtime,
            stat.ST_SIZE: metadata.get(BYTES_NAME)}

    def chmod(self, path, mode):
        pass

    # FIXME nasdsyncはupload,download両方共localをutimeしている。
    # dropbox側はutimeできない。時刻を基準にファイル更新判断しているので、
    # local側を更新することにより時刻一致させている。
    #def utime(self, path, mtime):
    #    pass

    def remove(self, path):
        try:
            self.__dclient.file_delete(path)
        except (rest.ErrorResponse), err:
            if err.status == STATUS_NOT_FOUND:
                raise NoEntryError()
            raise

    def rmdir(self, path):
        try:
            self.__dclient.file_delete(path)
        except (rest.ErrorResponse), err:
            if err.status == STATUS_NOT_FOUND:
                raise NoEntryError()
            raise

    def mkdir(self, path):
        try:
            self.__dclient.file_create_folder(path)
        except (rest.ErrorResponse), err:
            if err.status == STATUS_ALREADY_EXISTS:
                raise ExistsError()
            if err.status == STATUS_NO_SPACE:
                raise NoSpaceError()
            raise
        self.chmod(path, MODE_CH_DIR)
        return

    def join_path(self, path_list):
        return DROPBOX_SEP.join(path_list)


class DropboxFileTree(SrcFull, DstFull):

    def __init__(self, path, key="tscpobfe8qjf8xn", secret="9omt4qz00tenzv3", ignored_list=[], log_path=""):
        # FIXME key,secretは上位から受け取るようにすること
        self.dropbox_access = DropboxAccess(key, secret)
        SrcFull.__init__(self, path, ignored_list)
        DstFull.__init__(self, path, log_path)

    def listdir(self, path):
        return self.dropbox_access.listdir(path)

    def open(self, path, mode):
        return self.dropbox_access.open(path, mode)

    def mkdir(self, path):
        self.dropbox_access.mkdir(path)
        
    def stat(self, path):
        return self.dropbox_access.stat(path)

    def chmod(self, path, mode):
        self.dropbox_access.chmod(path, mode)

    # COMMENT いずれAPIが対応するかも
    #def utime(self, path, mtime):
    #    self.dropbox_access.utime(path, mtime)

    def remove(self, path):
        self.dropbox_access.remove(path)

    def rmdir(self, path):
        self.dropbox_access.rmdir(path)

    def join_path(self, path_list):
        return self.dropbox_access.join_path(path_list)


class AmazonS3FileObj(FileLikeObj):

    def __len__(self):

        return self.file.size

    def tell(self):

        # FIXME amazon用。read完了しても0にならないが問題あるか?
        # とりあえずNotImplement
        #return self.point
        raise NotImplementedError

    def seek(self, offset, whence=os.SEEK_SET):

        # FIXME エラー処理どうする
        # writeの場合のpointはサポートしないがよいか
        # とりあえずNotImplement
        #if whence == os.SEEK_SET:
        #    self.point = offset
        #elif whence == os.SEEK_CUR:
        #    self.point += offset
        #elif whence == os.SEEK_END:
        #    self.point = len(self) - offset         
        raise NotImplementedError

    def get_path(self):

        raise NotImplementedError

    def read(self, size):

        self.point += size
        return self.file.read(size)

    def write_by_path(self, file_like):

        try:
            # FIXME 一度に書いて大丈夫か?
            DEBUG("file_like len = " + str(len(file_like)))
            self.file.set_contents_from_file(file_like)
        # FIXME 例外種類
        except (rest.ErrorResponse), err:
            if err.status == STATUS_NO_SPACE:
                raise NoSpaceError()
            raise

    def close(self):
        # FIXME NOSPACE例外
        self.file.close()


class AmazonS3Access():

    def __init__(self, key, secret):
        self.__conn = S3Connection(key, secret)
        self.encoding = locale.getdefaultlocale()[1]
        if not self.encoding:
            self.encoding = DEFAULT_LOCALE
 
    def listdir(self, path):
        # FIXME 文字コード大丈夫か
        path_list = path.strip("/").rstrip("/").split("/", 1)
        # FIXME 自分は本当に0要素目か
        bucket_name = path_list[0]
        bucket = self.__conn.get_bucket(bucket_name)
        if len(path_list) == 2:
            key_path = path_list[1]
            keys = bucket.get_all_keys(prefix=key_path + "/")[1:]
            return [
                key.name.replace(key_path + "/", "", 1).rstrip("/").encode(AMAZONS3_ENCODE)
                for key
                in keys
                if len(key.name.replace(key_path + "/", "", 1).rstrip("/").split("/")) < 2]
        elif len(path_list) == 1:
            keys = bucket.get_all_keys()
            return [
                key.name.split("/")[0].encode(AMAZONS3_ENCODE)
                for key
                in keys
                if len(key.name.rstrip("/").split("/")) < 2]
        # FIXME path="/"などの場合にget_all_buckets()などするか?
        # FIXME elseはlen(path_list) == 1を想定している
        # それ以外の場合の処理未実装。必要か
       
    def isdir(self, entry): # FIXME 高速化するなら必要かも
        pass

    def open(self, path, mode):
        if mode == MODE_READ:
            try:
                # FIXME read writeでの区別できない?
                # FIXME 文字コード大丈夫か
                bucket_name, key_path = path.strip("/").rstrip("/").split("/", 1)
                bucket = self.__conn.get_bucket(bucket_name)
                key = bucket.get_key(key_path)
                return AmazonS3FileObj(key, path)
            # FIXME 例外種類
            except (rest.ErrorResponse), err:
                if err.status == STATUS_NOT_FOUND:
                    raise NoEntryError()
                raise

        if mode == MODE_WRITE:
            try:
                # FIXME 文字コード大丈夫か
                bucket_name, key_path = path.strip("/").rstrip("/").split("/", 1)
                bucket = self.__conn.get_bucket(bucket_name)
                # FIXME get_keyでは不在のときにNoneypeが返ってきたが、ほんとにnew_keyでよいか
                key = bucket.new_key(key_path)
                return AmazonS3FileObj(key, path)
            # FIXME 例外種類
            except (rest.ErrorResponse), err:
                if err.status == STATUS_NO_SPACE:
                    raise NoSpaceError()
                raise

    def stat(self, path):
        # FIXME 文字コード大丈夫か
        path_list = path.strip("/").rstrip("/").split("/", 1)
        bucket_name = path_list[0]
        bucket = self.__conn.get_bucket(bucket_name)
        if len(path_list) == 2:
            key_path = path_list[1]
            key = bucket.get_key(key_path)
            if not key:
                key = bucket.get_key(key_path + "/")
                if not key:
                    raise NoEntryError()
                else:
                    st_mode = ST_MODE_DIR
                    st_mtime = None
                    st_size = 0
            else:
                st_mode = ST_MODE_FILE
                # FIXME 時刻が.0ついていて、float?かなにかのよう。
                # ついていない方式もある。動作上問題ないかもしれないが型統一するべき
                st_mtime = emailut.mktime_tz(
                    emailut.parsedate_tz(
                        key.last_modified))
                st_size = key.size
        elif len(path_list) == 1:
            # COMMENT bucketはDIR扱い
            st_mode = ST_MODE_DIR
            st_mtime = None
            st_size = 0

        DEBUG("st_mode = " + str(st_mode))
        # FIXME 時刻のタイムゾーンどうする。nasdsyncにならうか
        # 国際化対応するなら、各国のタイムゾーンに対応するために
        # pytzやdatetimeなどのライブラリ必要
        return {
            stat.ST_MODE: st_mode,
            stat.ST_MTIME: st_mtime,
            stat.ST_SIZE: st_size}

    def chmod(self, path, mode):
        pass

    # FIXME nasdsyncはupload,download両方共localをutimeしている。
    # dropbox側はutimeできない。時刻を基準にファイル更新判断しているので、
    # local側を更新することにより時刻一致させている。
    #def utime(self, path, mtime):
    #    pass

    def remove(self, path):
        bucket_name, key_path = path.strip("/").rstrip("/").split("/", 1)
        # FIXME key_pathもある前提だが。ない場合のエラー処理はどうする
        bucket = self.__conn.get_bucket(bucket_name)
        try:
            bucket.delete_key(key_path)
        # FIXME 例外種類
        except (rest.ErrorResponse), err:
            if err.status == STATUS_NOT_FOUND:
                raise NoEntryError()
            raise

    def rmdir(self, path):
        path_list = path.strip("/").rstrip("/").split("/", 1)
        bucket_name = path_list[0]
        bucket = self.__conn.get_bucket(bucket_name)
        if len(path_list) == 2:
            key_path = path_list[1]
            key_list = bucket.get_all_keys(prefix=key_path + "/")
            try:
                bucket.delete_keys(key_list)
            # FIXME 例外種類
            except (rest.ErrorResponse), err:
                if err.status == STATUS_NOT_FOUND:
                    raise NoEntryError()
                raise
        elif len(path_list) == 1:
            try:
                bucket.delete_keys(bucket.get_all_keys())
                bucket.delete()
            # FIXME 例外種類
            except (rest.ErrorResponse), err:
                if err.status == STATUS_NOT_FOUND:
                    raise NoEntryError()
                raise
 
    def mkdir(self, path):
        path_list = path.strip("/").rstrip("/").split("/", 1)
        bucket_name = path_list[0]
        if len(path_list) == 2:
            bucket = self.__conn.get_bucket(bucket_name)
            key_path = path_list[1]
            try:
                key = bucket.new_key(key_path + "/")
                key.set_contents_from_string("")
            # FIXME 例外種類
            except:
                raise
        elif len(path_list) == 1:
            try:
                self.__conn.create_bucket(bucket_name)
            # FIXME 例外種類
            except (rest.ErrorResponse), err:
                if err.status == STATUS_ALREADY_EXISTS:
                    raise ExistsError()
                if err.status == STATUS_NO_SPACE:
                    raise NoSpaceError()
                raise
        # FIXME chmod不要かと
        #self.chmod(path, MODE_CH_DIR)

        return

    def join_path(self, path_list):
        return AMAZONS3_SEP.join(path_list)


class AmazonS3FileTree(SrcFull, DstFull):

    def __init__(self, bucket, key="tscpobfe8qjf8xn", secret="9omt4qz00tenzv3", log_path=""):
        # FIXME key,secretは上位から受け取るようにすること
        self.amazons3_access = AmazonS3Access(key, secret)
        # FIXME bucket単位限定?bucket以下のpath指定できる?
        SrcFull.__init__(self, bucket, ignored_list)
        DstFull.__init__(self, bucket, log_path)

    def listdir(self, path):
        return self.amazons3_access.listdir(path)

    def open(self, path, mode):
        return self.amazons3_access.open(path, mode)

    def mkdir(self, path):
        self.amazons3_access.mkdir(path)
        
    def stat(self, path):
        return self.amazons3_access.stat(path)

    def chmod(self, path, mode):
        self.amazons3_access.chmod(path, mode)

    # COMMENT いずれAPIが対応するかも
    #def utime(self, path, mtime):
    #    self.amazons3_access.utime(path, mtime)

    def remove(self, path):
        self.amazons3_access.remove(path)

    def rmdir(self, path):
        self.amazons3_access.rmdir(path)

    def join_path(self, path_list):
        return self.amazons3_access.join_path(path_list)


class SrcRestricted(Src):
    """Class for the source file tree supporting restricted access.""" 

    def get_tree(self):
        """Return the all file tree."""

        print "Restricted get tree from " + self.root

        # return path for this sample
        return "sub_dir"

    def get_delta_tree(self):
        """Return the all file tree."""

        print "Restricted get delta tree from " + self.root
        return self.get_tree()

    def read_tree(self, path):
        """Return a object for tree data."""

        raise NotImplemented()


class DstRestricted(Dst):
    """Class for the destination file tree supporting restricted access.""" 

    def update(self, src):
        
        # update_tree is file path for this sample
        self.write_tree(src.read_tree())

    def write_tree(self, path, tree_data):
        """Write a tree data."""

        raise NotImplemented()


class RsyncInfo:
    # FIXME 途中
    """Class for the information to execute "rsync".""" 

    def __init__(self, path, name, password):
        self.path = path # FIXME self.rootへ代入では?
        self.name = name            # For authorization
        self.password = password    # For authorization


class RsyncFileTree(SrcRestricted, DstRestricted):
    # FIXME 途中
    """Class for the source file tree supporting restricted access.""" 

    def __init__(self, path, name="", password="", base=""):
        self.name = name            # For authorization
        self.password = password    # For authorization
        SrcRestricted.__init__(self, path)

    def read_tree(self, path):
        """Return a RsyncInfo instance for source."""

        print "Return Rsync Info"
        return RsyncInfo(os.path.join(self.root, path.lstrip("/")),
                self.name, self.password)

    def write_tree(self, path, rsync_info):
        """Execute rsync for tree_data."""

        print "Execute rsync from " + rsync_info.path + " to " + \
            os.path.join(self.root, path.lstrip("/"))


class TreeMgr():

    def __set_sync_simp_tree_exists_in_base(self, base_tree, target_tree):

        DEBUG("12")
        for dir in base_tree[DIRS_NAME].keys():
            if dir in target_tree[FILES_NAME]:
                base_tree[FILES_NAME][dir] = {STATE_NAME: STATE_DEL}
            if dir in target_tree[DIRS_NAME]:
                base_tree[DIRS_NAME][dir][STATE_NAME] = STATE_SAME
                self.__set_sync_simp_tree_exists_in_base(
                    base_tree[DIRS_NAME][dir][CHILDREN_NAME],
                    target_tree[DIRS_NAME][dir][CHILDREN_NAME])

        DEBUG("13")
        for file in base_tree[FILES_NAME].keys():
            DEBUG("14")
            if file in target_tree[DIRS_NAME]:
                base_tree[DIRS_NAME][file] = {STATE_NAME: STATE_DEL}
            DEBUG("15")
            if file in target_tree[FILES_NAME]:
                base_info = base_tree[FILES_NAME][file]
                target_info = target_tree[FILES_NAME][file]
                DEBUG("16")
                if (base_info[MTIME_NAME] == target_info[MTIME_NAME] and
                    base_info[SIZE_NAME] == target_info[SIZE_NAME]):
                    base_tree[FILES_NAME][file][STATE_NAME] = STATE_SAME
                else:
                    base_info[STATE_NAME] = STATE_NEW
                DEBUG("17")
            else:
                DEBUG("18")
                base_tree[FILES_NAME][file][STATE_NAME] = STATE_NEW
                DEBUG("19")

    def __set_sync_simp_tree_noexists_in_base(self, base_tree, target_tree):

        for dir in target_tree[DIRS_NAME].keys():
            if not dir in base_tree[DIRS_NAME]:
                base_tree[DIRS_NAME][dir] = {
                    STATE_NAME: STATE_DEL}
            else:
                self.__set_sync_simp_tree_noexists_in_base(
                    base_tree[DIRS_NAME][dir][CHILDREN_NAME],
                    target_tree[DIRS_NAME][dir][CHILDREN_NAME])

        for file in target_tree[FILES_NAME].keys():
            if not file in base_tree[FILES_NAME]:
                base_tree[FILES_NAME][file] = {
                    STATE_NAME: STATE_DEL}

    def __set_sync_simp_tree(self, src_tree, dst_tree):

        DEBUG("10")
        self.__set_sync_simp_tree_exists_in_base(src_tree, dst_tree)
        DEBUG("11")
        self.__set_sync_simp_tree_noexists_in_base(src_tree, dst_tree)

    def __set_sync_dup_tree(self, base_tree, target_tree):

        for dir in base_tree[DIRS_NAME].keys():
            if dir in target_tree[DIRS_NAME]:
                if base_tree[DIRS_NAME][dir][STATE_NAME] == STATE_NEW:
                    if target_tree[DIRS_NAME][dir][STATE_NAME] == STATE_DEL:
                        target_tree[DIRS_NAME][dir][STATE_NAME] == STATE_SAME
                    if dir in target_tree[FILES_NAME]:
                        base_tree[FILES_NAME][file] = {
                            STATE_NAME: STATE_DEL}
                if base_tree[DIRS_NAME][dir][STATE_NAME] != STATE_DEL:
                    self.__set_sync_dup_tree(
                        base_tree[DIRS_NAME][dir][CHILDREN_NAME],
                        target_tree[DIRS_NAME][dir][CHILDREN_NAME])

        for file in base_tree[FILES_NAME].keys():
            if file in target_tree[FILES_NAME]:
                if base_tree[FILES_NAME][file][STATE_NAME] == STATE_NEW:
                    if target_tree[FILES_NAME][file][STATE_NAME] == STATE_DEL:
                        target_tree[FILES_NAME][file][STATE_NAME] = STATE_SAME
                    elif target_tree[FILES_NAME][file][STATE_NAME] == STATE_NEW:
                        base_info = base_tree[FILES_NAME][file]
                        target_info = target_tree[FILES_NAME][file]

                        # COMMENT mtime同じでsize違う場合、
                        # 後にset_sync_dup_treeのbase_treeとして
                        # 渡された方が反映される。
                        if (base_info[MTIME_NAME] >= target_info[MTIME_NAME] or
                            base_info[SIZE_NAME] != target_info[SIZE_NAME]):
                            target_info[SIZE_NAME] = base_info[SIZE_NAME]
                            target_info[MTIME_NAME] = base_info[MTIME_NAME]
                            base_info[STATE_NAME] = STATE_NEW
                            target_info[STATE_NAME] = STATE_SAME
                    elif file in target_tree[DIRS_NAME]:
                        base_tree[DIRS_NAME][file][STATE_NAME] = STATE_DEL

# FIXME 以前のロジック。上記の新ロジックは未テストなのでのこしてある　
#    def __set_sync_dup_tree(self, base_tree, target_tree):
#
#        for dir in base_tree[DIRS_NAME].keys():
#            if dir in target_tree[DIRS_NAME]:
#                if base_tree[DIRS_NAME][dir][STATE_NAME] == STATE_NEW:
#                    if (dir in target_tree[FILES_NAME] and
#                        target_tree[FILES_NAME][file][STATE_NAME] != STATE_DEL):
#                        target_tree[FILES_NAME][file][STATE_NAME] = STATE_DEL
#                self.__set_sync_dup_tree(
#                    base_tree[DIRS_NAME][dir][CHILDREN_NAME],
#                    target_tree[DIRS_NAME][dir][CHILDREN_NAME])
#
#        for file in base_tree[FILES_NAME].keys():
#            if file in target_tree[FILES_NAME]:
#                if base_tree[FILES_NAME][file][STATE_NAME] == STATE_NEW:
#                    if target_tree[FILES_NAME][file][STATE_NAME] == STATE_DEL:
#                        target_tree[FILES_NAME][file][STATE_NAME] = STATE_SAME
#                    elif target_tree[FILES_NAME][file][STATE_NAME] == STATE_NEW:
#                        base_info = base_tree[FILES_NAME][file]
#                        target_info = target_tree[FILES_NAME][file]
#                        if base_info[STATE_NAME] == STATE_NEW:
#                            if base_info[MTIME_NAME] >= target_info[MTIME_NAME]:
#                                target_info[SIZE_NAME] = base_info[SIZE_NAME]
#                                target_info[MTIME_NAME] = base_info[MTIME_NAME]
#                                base_info[STATE_NAME] = STATE_NEW
#                                target_info[STATE_NAME] = STATE_SAME
#                    elif (file in target_tree[DIRS_NAME] and
#                        target_tree[DIRS_NAME][file][STATE_NAME] != STATE_DEL):
#                        target_tree[DIRS_NAME][file][STATE_NAME] = STATE_DEL

    def __set_new_state(self, tree):

        DEBUG("7")
        for file in tree[FILES_NAME].keys():
            tree[FILES_NAME][file][STATE_NAME] = STATE_NEW

        DEBUG("8")
        for dir in tree[DIRS_NAME].keys():
            tree[DIRS_NAME][dir][STATE_NAME] = STATE_NEW
            self.__set_new_state(tree[DIRS_NAME][dir][CHILDREN_NAME])

    def __set_delete_state(self, tree):

        for file in tree[FILES_NAME].keys():
            tree[FILES_NAME][file][STATE_NAME] = STATE_DEL

        for dir in tree[DIRS_NAME].keys():
            tree[DIRS_NAME][dir][STATE_NAME] = STATE_DEL
            self.__set_new_state(tree[DIRS_NAME][dir][CHILDREN_NAME])

    def set_copy_tree(self, src):

        src.tree = src.get_tree()
        self.__set_new_state(src.tree)

    def set_update_copy_tree(self, src):

        DEBUG("4")
        is_updated = False

        if not src.tree:
            is_updated = True
            DEBUG("5")
            src.tree = src.get_tree()
            DEBUG("6")
            self.__set_new_state(src.tree)
        else:
            if src.set_delta_tree():
                is_updated = True

        DEBUG("9")
        return is_updated

    def set_sync_simp_tree(self, src, dst):

        is_updated = False

        DEBUG("2")
        is_src_updated = self.set_update_copy_tree(src)
        DEBUG("3")
        is_dst_updated = self.set_update_copy_tree(dst)
        if is_src_updated or is_dst_updated:
            is_updated = True
            self.__set_sync_simp_tree(src.tree, dst.tree)
        return is_updated

    def set_sync_dup_tree(self, src, dst):

        is_updated = False

        is_src_updated = self.set_update_copy_tree(src)
        is_dst_updated = self.set_update_copy_tree(dst)
        if is_src_updated or is_dst_updated:
            is_updated = True
            self.__set_sync_dup_tree(src.tree, dst.tree)
            self.__set_sync_dup_tree(dst.tree, src.tree)
        return is_updated


class DataSync():
    def update(self, src, dst):
       
        if isinstance(src, SrcFull):
            if isinstance(dst, DstFull):
                dst.update(src)
            elif isinstance(dst, DstRestricted):
                local_dst = LocalFileTree("local_temp")
                restricted_src = dst.__class__(local_dst.get_root(), base=dst)
    
                local_dst.update(src)
                dst.update(restricted_src, restricted_src.get_tree())
            else:
                raise NotImplementedError
    
        elif isinstance(src, SrcRestricted):
            if isinstance(dst, DstFull):
                restricted_dst = src.__class__("local_temp", base=src)
                local_src = LocalFileTree(restricted_dst.get_root())
    
                restricted_dst.update(src)
                dst.update(local_src, local_src.get_tree())
            elif isinstance(dst, DstRestricted):
                dst.update(src)
            else:
                raise NotImplementedError

    def copy(self, src, dst, tree_mgr):
        tree_mgr.set_copy_tree(src)
        self.update(src, dst)

    def update_copy(self, src, dst, tree_mgr):
        try:
            if not tree_mgr.set_update_copy_tree(src):
                raise NonDiffException
            self.update(src, dst)
        except NonDiffException: # 差異無しの場合など
            # FIXME
            DEBUG("NonDiff")
        except:
            # FIXME
            # 想定していないエラー発生時の出力どうする。
            # 方式によって受け取り型ことなるが。
            raise

    def sync_simp(self, src, dst, tree_mgr):
        try:
            DEBUG("1")
            if not tree_mgr.set_sync_simp_tree(src, dst):
                raise NonDiffException
            self.update(src, dst)
        except NonDiffException: # 差異無しの場合など
            # FIXME
            DEBUG("NonDiff")
        except:
            # FIXME
            # 想定していないエラー発生時の出力どうする。
            # 方式によって受け取り型ことなるが。
            raise

    def sync_dup(self, src, dst, tree_mgr):
        try:
            if not tree_mgr.set_sync_dup_tree(src, dst):
                raise NonDiffException
            self.update(src, dst)
            self.update(dst, src)
        except NonDiffException: # 差異無しの場合など
            # FIXME
            DEBUG("NonDiff")

