#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Backup."""

from __future__ import with_statement

import os
import re
import shutil

from iobase.xmldata import ANY_CHILD_PATH
from iobase.xmldata import XmlNode
from iobase.exception import NotFoundError
import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)
from albg_info import ALBG_BACKUPS_PATH
from libpal.crond import ctrl as crond
import pal_config as config
from pal_config import (DATE_CMD)
import ionas.fileaccess
from ionas.fileaccess import SmbAccess, LocalAccess


DAY_NAME = "day"
ID_NAME = "id"
HOUR_PATH = "time/hour"
MINUTE_PATH = "time/minute"
SCHEDULE_PATH = "job/schedule"
DATE_PATH = "job/schedule/date"
NAME_PATH = "//job/name"
ID_PATH = "//job/id"

FMT_ENTRYNAME = "%s_%s"
FMT_CMD_ARGS = '/usr/local/bin/exec_backup %s "%s"'

FMT_ADD_DATE_CMD_ARGS = \
   '[ `' + DATE_CMD + ' +\\%%Y` -eq %s ] && /usr/local/bin/exec_backup %s "%s"'


COMPLETE_STATUS = "COMPLETE"
NOT_EXEC_STATUS = "NOT_EXEC"

DISABLE_REMAIN_TIMES = "0"

OPTION_NAME = "option"
NAME_NAME = "name"
GENERATIONLIST_NAME = "generationlist"
GENERATION_NAME = "generation"
LOCALHOST_NAME = "localhost"
JOB_NAME = "job"
DESTINATION_NAME = "destination"
USERNAME_NAME = "user"
PASSWD_NAME = "passwd"
DETAIL_ATTR = "detail"
TIMES_ATTR = "times"

YEAR_NAME = "year"
MONTH_NAME = "month"
HOUR_NAME = "hour"
MINUTE_NAME = "minute"

UNC_RE = r"^\\\\[^\\]+\\.+"
UNC_HOST_RE = r"^\\\\([^\\]+)\\"
UNC_SHARE_RE = r"^\\\\[^\\]+\\(.+)"
GENERATION_ID_RE = "(^[0-9]{14})"

SMB_PREFIX = "smb://"
SMB_SEP = "/"

UNC_SEP = "\\"

def get_tip_node(node, fullpath):

    for path in fullpath.split("/"):
        try:
            node.get_node(path)
        except NotFoundError:
            node.append_child(path)
        finally:
            node = node.get_node(path)
    return node


class PalBackups(palcmd.PalCmd):
    """Pal Backup command implementation class."""

    def __add_schedule_entry(self, category, job_name, job_id, schedule_node):

        if not schedule_node.is_enabled():
            return

        week = ",".join([node.get_text()
            for node
            in schedule_node.get_node_list(DAY_NAME)])
        hour = schedule_node.get_node(HOUR_PATH).get_text()
        minutes = schedule_node.get_node(MINUTE_PATH).get_text()

        crond.add_entry(
            FMT_ENTRYNAME % (category, job_id),
            FMT_CMD_ARGS % (category, job_name),
            user="root",
            week=week,
            hour=hour,
            minutes=minutes)

    def __add_date_entry(self, category, job_name, job_id, date_node):

        year = date_node.get_node(YEAR_NAME).get_text()
        month = date_node.get_node(MONTH_NAME).get_text()
        day = date_node.get_node(DAY_NAME).get_text()
        hour = date_node.get_node(HOUR_NAME).get_text()
        minutes = date_node.get_node(MINUTE_NAME).get_text()

        crond.add_entry(
            FMT_ENTRYNAME % (category, job_id),
            FMT_ADD_DATE_CMD_ARGS % (year, category, job_name),
            user="root",
            minutes=minutes,
            hour=hour,
            day=day,
            month=month)

    def __add_cron_entry(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        job_name = body.get_node(NAME_PATH).get_text(False)
        job_id = body.get_node(ID_PATH).get_text()

        try:
            self.__add_schedule_entry(category, job_name, job_id,
                    body.get_node(category).get_node(SCHEDULE_PATH))
        except NotFoundError:
            pass

        try:
            self.__add_date_entry(category, job_name, job_id,
                    body.get_node(category).get_node(DATE_PATH))
        except NotFoundError:
            pass

        entry.set_result(Results.OK)
        return entry

    def __delete_cron_entry(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        job_id = body.get_node(ID_PATH).get_text()

        crond.delete_entry(FMT_ENTRYNAME % (category, job_id))

        entry.set_result(Results.OK)
        return entry

    def __get_file_info(self, file_name):

        file_path = os.path.join(config.BACKUP_STATUS_DIR, file_name)
        with open(file_path, "rb") as f:
            info_list = f.read().split(":", 1)
            if len(info_list) == 2:
                status, result = info_list
            else:
                status = info_list[0]
                result = ""

        timestamp = str(int(os.stat(file_path).st_mtime))

        return (status, result, timestamp)

    def __set_status(self, category, id, status):

        file_name = category + "_" + id
        tmp_file_path = os.path.join(
            config.TEMP_DIR, file_name + "." + str(os.getpid()))
        file_path = os.path.join(
            config.BACKUP_STATUS_DIR, file_name)

        with open(tmp_file_path, "wb") as tmp_file:
            tmp_file.write(status)

        shutil.move(tmp_file_path, file_path)

    def __set_remain_times(self, category, id, times):

        file_name = category + "_" + id
        tmp_file_path = os.path.join(
            config.TEMP_DIR, file_name + "." + str(os.getpid()))
        file_path = os.path.join(
            config.BACKUP_REMAIN_TIMES_DIR, file_name)

        with open(tmp_file_path, "wb") as tmp_file:
            tmp_file.write(str(times))

        shutil.move(tmp_file_path, file_path)

    def __set_status_to_xml(self, status_fname, entry):

        status, result, timestamp = self.__get_file_info(status_fname)
        category, job_id = status_fname.split("_")

        joblist_node = get_tip_node(entry.get(), category + "/joblist")
        joblist_node.append_child("job")
        job_node = joblist_node.get_node_list("job")[-1]

        job_node.append_child("id", job_id)
        job_node.append_child("status", status)

        if status == COMPLETE_STATUS:
            job_node.append_child("result", result)
            job_node.append_child("timestamp", timestamp)

    def __set_xml_to_status(self, category, job_node):

        id = job_node.get_value("id")
        status = job_node.get_value("status")

        self.__set_status(category, id, status)

    def __set_accesser(self, accesser):

        self.dst_listdir = lambda path:\
            accesser.listdir(path)

    def __set_remote_accesser(self, username, passwd):

        return self.__set_accesser(SmbAccess(username, passwd))

    def __set_local_accesser(self):

        return self.__set_accesser(LocalAccess())

    def __get_remote_abspath(self, unc, job_name):

        share_path = \
            re.match(UNC_SHARE_RE, unc).group(1).replace(UNC_SEP, SMB_SEP)
        host = re.match(UNC_HOST_RE, unc).group(1)
        return SMB_PREFIX + SMB_SEP.join([host, share_path, job_name])

    def _get_remote_dst_user_info(self, node):

        try:
            username = node.get_node(USERNAME_NAME).get_text()
        except NotFoundError:
            username = ""

        try:
            passwd = node.get_node(PASSWD_NAME).get_text()
        except NotFoundError:
            passwd = ""

        return (username, passwd)

    def __get_generation_dir_list(self, abspath):
        """Get all generation list."""

        gen_dir_list = []
        for gen_dir in self.dst_listdir(abspath):
            match = re.match(GENERATION_ID_RE, gen_dir)
            if match:
                gen_dir_list.append(match.group(1))
        return gen_dir_list

    def __get_generation_list_node(self, opt_job_node):

        UNC_PATH = "destination/unc"
        DST_PATH_PATH = "destination/path"

        unc = opt_job_node.get_node(UNC_PATH).get_text()
        job_name = opt_job_node.get_node(NAME_NAME).get_text()

        if re.match(UNC_RE, unc):
            share = re.match(UNC_SHARE_RE, unc).group(1)
            host = re.match(UNC_HOST_RE, unc).group(1)
            username, passwd = \
                self._get_remote_dst_user_info(
                    opt_job_node.get_node(DESTINATION_NAME))
            abspath = self.__get_remote_abspath(unc, job_name)
            self.__set_remote_accesser(username, passwd)
        else:
            share = unc
            host = LOCALHOST_NAME
            unc = "\\\\%s\\%s" % (host, share)
            username, passwd = ("", "")
            abspath = os.path.join(opt_job_node.get_node(
                          DST_PATH_PATH).get_text(), job_name)
            self.__set_local_accesser()

        generationlist_node = XmlNode(GENERATIONLIST_NAME)
        for gen_dir in self.__get_generation_dir_list(abspath):
            generationlist_node.append_child(GENERATION_NAME)
            generation_node = \
                generationlist_node.get_node_list(GENERATION_NAME)[-1]
            generation_node.append_child(ID_NAME, gen_dir)

        return generationlist_node

    def __append_generation_list_node(self, entry):

        OPT_JOBLIST_PATH = "./option/joblist"
        JOB_PATH = "backup/joblist/job"
        JOB_ID_PATH_TMPL = "./job[id='%s']"

        opt_joblist_node = entry.get().get_node(OPT_JOBLIST_PATH)

        for job_node in entry.get().get_node_list(JOB_PATH):

            job_id = job_node.get_value(ID_NAME)

            try:
                opt_job_node = \
                    opt_joblist_node.get_node(JOB_ID_PATH_TMPL % job_id)
            except NotFoundError:
                continue

            try:
                job_node.append_child(
                    self.__get_generation_list_node(opt_job_node))
            except:
                continue  # fixme

    def __init_status(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        id = body.get_value(ID_PATH)

        self.__set_status(category, id, NOT_EXEC_STATUS)

    def __init_remain_times(self, entry):

        try:
            body = entry.get()
            category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
            id = body.get_value(ID_PATH)

            self.__set_remain_times(category, id, DISABLE_REMAIN_TIMES)
        except:
            pass

    def __update_remain_times(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()

        try:
            schedule_node = body.get_node(category).get_node(SCHEDULE_PATH)
            if not schedule_node.is_enabled():
                return
            times = schedule_node.get_attr_value(TIMES_ATTR)
        except NotFoundError:
            times = "0"

        id = body.get_value(ID_PATH)
        self.__set_remain_times(category, id, times)

    def __execute_handler(self, entry):

        return self.send_entry(entry, ALBG_BACKUPS_PATH)

    def __add_handler(self, entry):

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        job_id = body.get_node(ID_PATH).get_text()
        if job_id and not job_id.isdigit():
            self.__init_status(entry)

        self.__update_remain_times(entry)
        return self.__add_cron_entry(entry)

    def __modify_handler(self, entry):

        self.__init_status(entry)

        body = entry.get()
        category = body.get_node_list(ANY_CHILD_PATH)[0].get_node_name()
        schedule_node = body.get_node(category).get_node(SCHEDULE_PATH)

        if schedule_node.is_enabled():
            self.__update_remain_times(entry)
            return self.__add_cron_entry(entry)
        else:
            return self.__delete_cron_entry(entry)

    def __delete_handler(self, entry):

        self.__init_status(entry)

        self.__init_remain_times(entry)

        return self.__delete_cron_entry(entry)

    def __query_handler(self, entry):

        try:
            for status_fname in os.listdir(config.BACKUP_STATUS_DIR):
                self.__set_status_to_xml(status_fname, entry)

            try:
                if entry.get().get_node(
                       OPTION_NAME).is_set_to_true(DETAIL_ATTR):
                    self.__append_generation_list_node(entry)
            except NotFoundError:
                pass

            entry.set_result(Results.OK)

        except:
            self.exception("Query Error.")
            entry.set_result(Results.QUERY_ERROR)

        return entry

    def __apply_handler(self, entry):

        try:
            for category_node in entry.get().get_node_list(ANY_CHILD_PATH):
                for job_node in category_node.get_node_list("joblist/job"):
                    self.__set_xml_to_status(
                        category_node.get_node_name(), job_node)
            entry.set_result(Results.OK)

        except:
            self.exception("Apply Error.")
            entry.set_result(Results.APPLY_ERROR)

        return entry

    def __restore_handler(self, entry):

        return self.send_entry(entry, ALBG_BACKUPS_PATH)

    def __stop_handler(self, entry):

        return self.send_entry(entry, ALBG_BACKUPS_PATH)

    def __init__(self, option=None):
        """Constructor."""

        self.dst_listdir = None

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.EXECUTE, self.__execute_handler)
        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.MODIFY, self.__modify_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.RESTORE, self.__restore_handler)
        self.set_handler(ReqTypes.STOP, self.__stop_handler)

    def post_process(self, response_data):

        self.dst_listdir = None
        reload(ionas.fileaccess)
        return palcmd.PalCmd.post_process(self, response_data)

#
# Main
#

if __name__ == "__main__":
    PalBackups(option=palcmd.PalCmdOption()).handle_request()
