#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for BUZZER."""

from __future__ import with_statement

import os

import pal_config as config
import libpal.palcmd as palcmd
from iobase.xmldata import XmlNode
from iobase.exception import (NotFoundError, InvalidArgError)
from pal_info import (ReqTypes, Results)

### XML Tag Name ###
BUZZER_NAME = "buzzer"
SCORE_NAME = "score"
STATUS_NAME = "status"

MELODY_ON_FLG = "1"
MELODY_ON = "on"
MELODY_OFF = "off"

BUZZER_QUERY_XML_TMPL = "<buzzer><status>%s</status></buzzer>"


class PalBuzzer(palcmd.PalCmd):
    """PalBuzzer command implementation class."""

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        node_name = entry.get().get_node_name()
        if BUZZER_NAME != node_name:
            self.error("Invalid root node name %s" % node_name)
            raise InvalidArgError("Invalid root node name")

    def __is_exist_node(self, node, node_path):
        """Check whether node_path exists in specified node"""

        try:
            node.get_node(node_path)
        except NotFoundError:
            return False
        return True

    def __is_melody_stoppping_now(self):

        try:
            with open(config.PROC_MELODY_FILE, "r") as fh:
                if MELODY_ON_FLG == fh.read().rstrip():
                    return True
        except:
            pass

        return False

    def __play_buzzer(self, score):
        """Play buzzer"""

        try:
            if (not score.startswith(config.REPEATE_MELODY_PREFIX) and
                not os.path.exists(config.BOOT_COMPLETE_FILE)):

                return
        except:
            pass

        with open(config.PROC_MELODY_FILE, "w") as fh:
            fh.write(score)

    def __stop_buzzer(self):
        """Stop buzzer"""

        with open(config.PROC_MELODY_FILE, "w") as fh:
            fh.write("S")
        return

    #
    # Request handlers
    #

    def __query_handler(self, entry):
        """Handle "query" request."""

        if self.__is_melody_stoppping_now():
            entry.set(XmlNode(BUZZER_QUERY_XML_TMPL % MELODY_ON))
        else:
            entry.set(XmlNode(BUZZER_QUERY_XML_TMPL % MELODY_OFF))

        entry.set_result(Results.OK)
        return entry

    def __play_handler(self, entry):
        """Handle "play" request."""

        self.__validate_root_tag_name(entry)
        score = entry.get().get_node(SCORE_NAME).get_text()
        self.__play_buzzer(score)

        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.PLAY, self.__play_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)

#
# Main
#

if __name__ == "__main__":
    PalBuzzer(option=palcmd.PalCmdOption()).handle_request()
