#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for setting time and date."""


import os
from datetime import datetime

import iobase.xmldata as xmldata
import libpal.palcmd as palcmd
from pal_info import (ReqTypes, Results)
from iobase.exception import (NotFoundError, InvalidArgError)
from iobase.cmdexec import (CmdExec, NULL)
import pal_config as config
from libpal.crond import ctrl as crond


DATETIME = "datetime"
MANUAL = "manual"
DATE = "date"
TIME = "time"
YEAR = "year"
MONTH = "month"
DAY = "day"
HOUR = "hour"
MINUTE = "min"
SECOND = "sec"

DATETIME_SYNC = "datetime_sync"


class PalDatetime(palcmd.PalCmd):
    """Pal Datetime command implementation class."""

    def __query_handler(self, entry):
        """Handle query request and set entry to result."""

        time_node = entry.get()
        if time_node.get_node_name() != DATETIME:
            raise InvalidArgError(time_node.get_node_name())

        try:
            today = datetime.today()
        except:
            entry.set_result(Results.GETTIME_ERROR)
            return entry

        try:
            time_node.clear()
        except NotFoundError:
            pass

        time_node.append_child(DATE, "")
        time_node.append_child(TIME, "")
        date_node = time_node.get_node(DATE)
        time_node = time_node.get_node(TIME)
        date_node.append_child(YEAR, str(today.year))
        date_node.append_child(MONTH, str(today.month))
        date_node.append_child(DAY, str(today.day))
        time_node.append_child(HOUR, str(today.hour))
        time_node.append_child(MINUTE, str(today.minute))
        time_node.append_child(SECOND, str(today.second))

        entry.set_result(Results.OK)
        return entry

    def __execute_handler(self, entry):
        proc = CmdExec(config.NTPSYNC_CMD, stdout=NULL, stderr=NULL).execute()
        if proc.wait():
            self.error("Failed to synchronizing with server")
            raise ExecError(config.NTPSYNC_CMD)

        entry.set_result(Results.OK)
        return entry

    def __apply_handler(self, entry):
        """Handle apply request and set entry to result."""

        set_node = entry.get()
        if set_node.get_node_name() != DATETIME:
            raise InvalidArgError(set_node.get_node_name())

        ntp = set_node.get_node("ntp")
        if ntp.is_enabled():
            # ntp
            if ntp.get_node("sync/time").is_enabled():
                week = "0,1,2,3,4,5,6"
                hour = ntp.get_node("sync/time/hour").get_text()
                minutes = ntp.get_node("sync/time/minute").get_text()

                crond.add_entry(DATETIME_SYNC, config.NTPSYNC_CMD,
                        user="root", week=week, hour=hour, minutes=minutes)
            else:
                try:
                    crond.delete_entry(DATETIME_SYNC)
                except:
                    pass
        else:
            # ntp
            try:
                crond.delete_entry(DATETIME_SYNC)
            except:
                pass

            # manual
            try:
                method = set_node.get_node(MANUAL)
                date_node = method.get_node(DATE)
                date_dic = {}
                date_dic["year"] = date_node.get_node("year").get_text()
                date_dic["month"] = date_node.get_node("month").get_text()
                date_dic["day"] = date_node.get_node("day").get_text()
                date_dic["hour"] = date_node.get_node("hour").get_text()
                date_dic["min"] = date_node.get_node("min").get_text()
                date_dic["sec"] = date_node.get_node("sec").get_text()
                for key in date_dic.keys():
                    if len(date_dic[key]) == 1:
                        date_dic[key] = "0" + date_dic[key]

                date_str = "%(Y)s-%(M)s-%(D)s" % {
                    "Y": date_dic["year"],
                    "M": date_dic["month"],
                    "D": date_dic["day"]}
                time_str = "%(H)s:%(M)s:%(S)s" % {
                    "H": date_dic["hour"],
                    "M": date_dic["min"],
                    "S": date_dic["sec"]}
                datetime_str = "%(D)s %(T)s" % {"D": date_str, "T": time_str}

                # /bin/date
                date_cmd = [config.DATE_CMD, "-s", datetime_str]
                proc = CmdExec(date_cmd, stdout=NULL, stderr=NULL).execute()
                if proc.wait():
                    self.error("Failed to set datetime manually")
                    raise ExecError(config.DATE_CMD)

                # /sbin/hwclock
                hwclock_cmd = [
                    config.HWCLOCK_CMD, "--utc", "--noadjfile", "--systohc"]
                proc = CmdExec(hwclock_cmd, stdout=NULL, stderr=NULL).execute()
                if proc.wait():
                    self.error("Failed to set hwclock")
                    raise ExecError(config.HWCLOCK_CMD)
            except NotFoundError:
                pass

        entry.set_result(Results.OK)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.EXECUTE, self.__execute_handler)

#
# Main
#

if __name__ == "__main__":
    PalDatetime(option=palcmd.PalCmdOption()).handle_request()
