#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Dropbox."""

import os
import time
import libpal.palcmd as palcmd
from iobase.xmldata import XmlConfNode
from iobase.exception import (InvalidArgError, ExecError)
from iobase.cmdexec import (CmdExec, NULL)
from pal_info import ReqTypes
from pal_info import Results
from pal_config import (NASDSYNC_CMD, DROPBOX_URL_DIR)


### XML Path ###
NAME_PATH = "share/name"

### XML Tag Name ###
SHARES_NAME = "shares"
SHARE_NAME = "share"

### XML Format str ###
DROPBOX_URL_FMT_STR = """<shareservice>
    <dropbox>
        <url>%s</url>
    </dropbox>
</shareservice>"""

### Command ###
AUTH_NASDSYNC_ARGS = [NASDSYNC_CMD, "--save-access-token",
    "--share", ""]
CLEAR_NASDSYNC_ARGS = [NASDSYNC_CMD, "--clear-access-token",
    "--share", ""]
SHARE_IDX = 3

### nasdsync RETRY_TIME ###
RETRY_TIME = 300

### nasdsync error auth url ###
ERROR_AUTH_URL_RESULT = "Error"


class PalDropbox(palcmd.PalCmd):
    """Pal Dropbox implementation class."""

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        root_tag_name = entry.get().get_node_name()
        if SHARES_NAME != root_tag_name:
            raise InvalidArgError("invalid root tag..")
        return

    def __execute_cmd(self, share_name, args, wait=False):

        cmd_args = list(args)
        cmd_args[SHARE_IDX] = share_name
        CmdExec(cmd_args, stdin=NULL, stdout=NULL,
            stderr=NULL, wait=wait, close_fds=True).execute()

    def __set_auth_url(self, entry, share_name):

        file_path = os.path.join(DROPBOX_URL_DIR, share_name)
        url = ""
        while True:
            if os.path.exists(file_path):
                with open(file_path, "rb") as file:
                    url = file.read().strip()
                break
            time.sleep(3)

        entry.get().get_node(SHARE_NAME).append_child(
            XmlConfNode(DROPBOX_URL_FMT_STR % url))

    def __is_in_time(self, file_path):

        if os.stat(file_path).st_mtime + RETRY_TIME > time.time():
            return True
        return False

    def __is_last_error(self, file_path):

        result = False

        with open(file_path, "rb") as file:
            if file.read().strip() == ERROR_AUTH_URL_RESULT:
                result = True

        if result:
            os.unlink(file_path)

        return result

    def __is_execute_auth(self, share_name):

        result = True

        file_path = os.path.join(DROPBOX_URL_DIR, share_name)

        if os.path.exists(file_path):
            if self.__is_in_time(file_path):
                result = False
            if self.__is_last_error(file_path):
                result = True

        return result
        
    #
    # Request handlers
    #

    def __auth_handler(self, entry):
        """Handle "auth" request."""

        self.__validate_root_tag_name(entry)

        share_name = entry.get().get_node(NAME_PATH).get_text()

        try:
            # URLファイルがすでに存在し、timestampが5分以内であれば
            # そのURLをそのまま返す
            if self.__is_execute_auth(share_name):
                self.__execute_cmd(share_name, AUTH_NASDSYNC_ARGS)
            self.__set_auth_url(entry, share_name)
            entry.set_result(Results.OK)

        except ExecError:
            entry.set_result(Results.AUTH_ERROR)
            
        return entry

    def __clear_handler(self, entry):

        self.__validate_root_tag_name(entry)

        share_name = entry.get().get_node(NAME_PATH).get_text()

        try:
            self.__execute_cmd(share_name, CLEAR_NASDSYNC_ARGS, wait=True)
            entry.set_result(Results.OK)
        except ExecError:
            entry.set_result(Results.CLEAR_ERROR)

        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.AUTH, self.__auth_handler)
        self.set_handler(ReqTypes.CLEAR, self.__clear_handler)

#
# Main
#

if __name__ == "__main__":
    PalDropbox(option=palcmd.PalCmdOption()).handle_request()
