#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Eco."""

import os
import pal_config as config
import libpal.palcmd as palcmd
import libpal.util as util
from iobase.exception import (InvalidArgError, ExecError)
from iobase.cmdexec import (CmdExec, NULL)
from pal_info import ReqTypes
from pal_info import Results
from pal_config import (DEV_DIR, SPINDOWN_CONF_FILE, STOR_IF_TO_VOL,
                        SATA_HDD_DEV_LIST, VOL_INFO, SPINDOWND_INITD)

### XML Path ###
MINUTE_PATH = "./time/minute"
VOLUME_PATH = "./volumelist/volume"

### XML Tag Name ###
ECO_NAME = "eco"
NAME_NAME = "name"

### Command ###
RESTART_SPINDOWND_ARG = [SPINDOWND_INITD, "restart"]


class PalEco(palcmd.PalCmd):
    """Pal Eco implementation class."""

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        root_tag_name = entry.get().get_node_name()
        if ECO_NAME != root_tag_name:
            raise InvalidArgError("invalid root tag..")
        return

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], variable_list=[]):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]

        if len(cmd_var_idx_list) != len(variable_list):
            raise ExecError

        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = variable_list[loop_idx]

        return CmdExec(cmd,
                       stdout=std_out,
                       stderr=std_err,
                       wait=is_wait,
                       raise_exc=is_enable_exception).execute()

    def __restart_spindown_daemon(self):
        """Restart spindown daemon."""

        self.__execute_cmd(True, False,
                           NULL, NULL,
                           RESTART_SPINDOWND_ARG)
        return

    def __get_vol_to_spindown_target_map(self):

        vol_dev_map = {}

        vol_dev_map.update({STOR_IF_TO_VOL["internal"][1]:
                           SATA_HDD_DEV_LIST[:util.get_cuurent_array_num()]})

        for vol in STOR_IF_TO_VOL["usb"].values():
            vol_dev_map.update({vol: (VOL_INFO[vol]["dev"],)})

        return vol_dev_map

    def __get_dev_path_list_from_volname(self, volume_name):
        """get real device path from specified volume name"""

        dev_name_list = self.__get_vol_to_spindown_target_map()[volume_name]

        dev_path_list = []
        for dev_name in dev_name_list:
            if os.path.exists(os.path.join(DEV_DIR, dev_name)):
                real_path = os.path.realpath(os.path.join(DEV_DIR, dev_name))
                dev_path_list.append(real_path)

        return dev_path_list

    #
    # Request handlers
    #

    def __apply_handler(self, entry):
        """Handle "apply" request."""

        self.__validate_root_tag_name(entry)

        minute_str = entry.get().get_node(MINUTE_PATH).get_text()
        spindown_second = int(minute_str) * 60

        vol_node_list = entry.get().get_node_list(VOLUME_PATH)
        try:
            os.makedirs(os.path.dirname(SPINDOWN_CONF_FILE))
        except:
            pass

        conf_fh = open(SPINDOWN_CONF_FILE, "w")

        for vol_node in vol_node_list:

            if not vol_node.is_enabled():
                continue

            vol_name = vol_node.get_node(NAME_NAME).get_text()
            dev_path_list = self.__get_dev_path_list_from_volname(vol_name)

            for dev_path in dev_path_list:
                conf_fh.write("%s %d\n" % (dev_path, spindown_second))

        conf_fh.close()
        self.__restart_spindown_daemon()
        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.APPLY, self.__apply_handler)

#
# Main
#

if __name__ == "__main__":
    PalEco(option=palcmd.PalCmdOption()).handle_request()
