#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for LED."""

from __future__ import with_statement

import libpal.palcmd as palcmd
from iobase.cmdexec import (CmdExec, NULL)
from iobase.exception import (ExecError, InvalidArgError, NotFoundError)
from pal_config import (SET_LED_CMD, I2CSET_CMD)
import pal_config as config
from pal_info import (Results, ReqTypes)

### XML Tag Name ###
LED_NAME = "led"
POWER_NAME = "power"
STATUS_NAME = "status"
HDD_NAME = "hdd"
BRIGHTNESS_NAME = "brightness"
ANY_CHILD_NAME = "*"

### XML Attribute ###
COLOR_ATTR = "color"
HDD_NUM_ATTR = "hddnum"

### XML Attribute value ###
HDD1_ATTR_VAL = "1"
HDD2_ATTR_VAL = "2"
HDD3_ATTR_VAL = "3"
HDD4_ATTR_VAL = "4"
HDD5_ATTR_VAL = "5"
HDD6_ATTR_VAL = "6"
RED = "red"
GREEN = "green"
BLUE = "blue"
ORANGE = "orange"
NON_COLER = "NOCOLOR"

### XML Element ###
ON_EL = "on"
BLINK_EL = "blink"
OFF_EL = "off"
BRIGHTNESS_LEVEL1_EL = "25"
BRIGHTNESS_LEVEL2_EL = "50"
BRIGHTNESS_LEVEL3_EL = "75"
BRIGHTNESS_LEVEL4_EL = "100"

### Command and argment ###
#
# XXXXX_ARG_FORMAT is the prototype list of the command which inclues VAR.
# VAR is replaced dynamically in __execute_cmd() method using
# XXXXX_VAR_INDEX_LIST and some variables._FORMAT
#
VAR = "%s"
CTRL_POWER_LED_ON_CMD = [SET_LED_CMD, "POWER", "GREEN", "ON"]
CTRL_POWER_LED_OFF_CMD = [SET_LED_CMD, "POWER", "OFF"]

CTRL_STATUS_LED_ARG = [SET_LED_CMD, "STATUS", VAR, VAR]
CTRL_STATUS_LED_VAR_INDEX_LIST = [2, 3]

CTRL_HDD_LED_ARG = [I2CSET_CMD, "-y", "0", "0x0d", VAR, VAR, "w"]
CTRL_HDD_LED_VAR_INDEX_LIST = [4, 5]

SET_HDD_LED_BRIGHTNRSS_LEVEL1_CMD = \
    [I2CSET_CMD, "-y", "0", "0x0d", "0x08", "0x0"]

SET_HDD_LED_BRIGHTNRSS_LEVEL2_CMD = \
    [I2CSET_CMD, "-y", "0", "0x0d", "0x08", "0x5"]

SET_HDD_LED_BRIGHTNRSS_LEVEL3_CMD = \
    [I2CSET_CMD, "-y", "0", "0x0d", "0x08", "0xa"]

SET_HDD_LED_BRIGHTNRSS_LEVEL4_CMD = \
    [I2CSET_CMD, "-y", "0", "0x0d", "0x08", "0xf"]

### Command option value ###
ON_CMD_OP = "ON"
BLINK_CMD_OP = "BLINK"
OFF_CMD_OP = "OFF"
RED_CMD_OP = "RED"
GREEN_CMD_OP = "GREEN"
ORANGE_CMD_OP = "ORANGE"

HDD_LED_ON_OP = "0x07"
HDD_LED_BLINK_OP = "0x09"


class PalLed(palcmd.PalCmd):
    """Pal Led command implementation class."""

    #
    # Sub functions
    #

    def __check_correct_root_node(self, entry):
        """Check whether root node is correctly.
        """

        node_name = entry.get().get_node_name()
        if node_name != LED_NAME:
            raise InvalidArgError(node_name)

    def __is_exist_node(self, node, node_path):
        """Check whether node_path exists in specified node"""

        try:
            node.get_node(node_path)
        except NotFoundError:
            return False
        return True

    def __get_value_by_key(self, target_key, dict):
        """This method is for utility.
           Find key of "dict" and return corresponding key.
        """

        for key, value in dict.iteritems():
            if target_key == key:
                return value
        raise InvalidArgError("Invalid key:" + str(target_key))

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], variable_list=[]):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]

        if len(cmd_var_idx_list) != len(variable_list):
            raise ExecError("Command list size is bad.. (__execute_cmd)")

        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = cmd[var_idx] % variable_list[loop_idx]

        return CmdExec(cmd,
                       stdout=std_out,
                       stderr=std_err,
                       wait=is_wait,
                       raise_exc=is_enable_exception).execute()

    def __transact_power_led_node(self, power_led_node):
        """Transact power node.
        """

        mode = power_led_node.get_text()

        if ON_EL == mode:
            self.__execute_cmd(True, True, NULL, NULL, CTRL_POWER_LED_ON_CMD)
            return

        if OFF_EL == mode:
            self.__execute_cmd(True, True, NULL, NULL, CTRL_POWER_LED_OFF_CMD)
            return

        raise InvalidArgError("Invalid power led mode:" + str(mode))

    def __transact_status_led_node(self, status_led_node):
        """Transact status node.
        """

        try:
            color_attr_val = status_led_node.get_attr_value(COLOR_ATTR)
            color_op = self.__get_value_by_key(color_attr_val,
                                               {RED: RED_CMD_OP,
                                                GREEN: GREEN_CMD_OP,
                                                ORANGE: ORANGE_CMD_OP})
        except NotFoundError:
            color_op = GREEN_CMD_OP

        mode_op = \
            self.__get_value_by_key(status_led_node.get_text(),
                                    {ON_EL: ON_CMD_OP,
                                     BLINK_EL: BLINK_CMD_OP,
                                     OFF_EL: OFF_CMD_OP})

        self.__execute_cmd(True, True, NULL, NULL,
                           CTRL_STATUS_LED_ARG,
                           CTRL_STATUS_LED_VAR_INDEX_LIST,
                           [color_op, mode_op])
        return

    def __get_hddnum_from_attr(self, target_key):
        """Get HDD number(integer) by specified key"""

        return self.__get_value_by_key(target_key,
                                       {HDD1_ATTR_VAL: 0,
                                        HDD2_ATTR_VAL: 1,
                                        HDD3_ATTR_VAL: 2,
                                        HDD4_ATTR_VAL: 3,
                                        HDD5_ATTR_VAL: 4,
                                        HDD6_ATTR_VAL: 5})

    def __get_cmd_hdd_map_cmd_opt(self, hdd_light_on_status_dict):

        map_opt_value = 0xf000
        for hdd_num, color_dict in hdd_light_on_status_dict.iteritems():

            if color_dict[RED] << hdd_num:
                map_opt_value += 0x01 << hdd_num * 2

            if color_dict[BLUE] << hdd_num:
                map_opt_value += 0x02 << hdd_num * 2
        return hex(int(map_opt_value))

    def __hold_hdd_led_color(self, hdd_num, color):

        REQ_COLOR_TO_HODL_COLOR = \
            {BLUE: config.HOLD_STATE_BLUE,
             RED: config.HOLD_STATE_RED,
             NON_COLER: config.HOLD_STATE_OFF}

        with open(config.HDD_LED_HOLD_STATE_FILE % hdd_num, "w") as fh:
            fh.write(REQ_COLOR_TO_HODL_COLOR[color])

    def __get_current_hdd_led_color(self, hdd_num):

        try:
            with open(config.HDD_LED_HOLD_STATE_FILE % hdd_num, "r") as fh:
                return fh.read().rstrip()
        except:
            return OFF_EL

    def __transact_hdd_led_status_node_list(self, hdd_led_status_node_list):
        """Transact ctrl hdd status led node.
           <hdd>
             <state color="red" hddnum="1" >on</state>
        """

        hdd_light_on_status_dict = \
            {0: {BLUE: False, RED: False},
             1: {BLUE: False, RED: False},
             2: {BLUE: False, RED: False},
             3: {BLUE: False, RED: False},
             4: {BLUE: False, RED: False},
             5: {BLUE: False, RED: False}}

        hdd_blink_status_dict = \
            {0: {BLUE: False, RED: False},
             1: {BLUE: False, RED: False},
             2: {BLUE: False, RED: False},
             3: {BLUE: False, RED: False},
             4: {BLUE: False, RED: False},
             5: {BLUE: False, RED: False}}

        for status_node in hdd_led_status_node_list:

            # HDD number
            hddnum_attr_val = status_node.get_attr_value(HDD_NUM_ATTR)
            hdd_num = self.__get_hddnum_from_attr(hddnum_attr_val)

            # LED mode
            mode = status_node.get_text()

            # LED color
            try:
                color = status_node.get_attr_value(COLOR_ATTR)
                if OFF_EL == mode:
                    color = NON_COLER
            except NotFoundError:
                color = NON_COLER

            # hold color to proc file
            try:
                if OFF_EL == mode:
                    self.__hold_hdd_led_color(hdd_num + 1, NON_COLER)
                else:
                    self.__hold_hdd_led_color(hdd_num + 1, color)
            except:
                self.exception("fail to __hold_hdd_led_color()")

            # set each status dict
            if ON_EL == mode:
                hdd_light_on_status_dict[hdd_num][color] = True
                hdd_blink_status_dict[hdd_num][color] = False

            elif BLINK_EL == mode:
                hdd_light_on_status_dict[hdd_num][color] = True
                hdd_blink_status_dict[hdd_num][color] = True

            elif OFF_EL == mode:
                hdd_light_on_status_dict[hdd_num][BLUE] = False
                hdd_light_on_status_dict[hdd_num][RED] = False
                hdd_blink_status_dict[hdd_num][BLUE] = False
                hdd_blink_status_dict[hdd_num][RED] = False

        # blink command
        blink_hdd_map_opt = \
            self.__get_cmd_hdd_map_cmd_opt(hdd_blink_status_dict)

        self.__execute_cmd(True, True, NULL, NULL,
                           CTRL_HDD_LED_ARG,
                           CTRL_HDD_LED_VAR_INDEX_LIST,
                           [HDD_LED_BLINK_OP, blink_hdd_map_opt])

        # blight command
        blight_hdd_map_opt = \
            self.__get_cmd_hdd_map_cmd_opt(hdd_light_on_status_dict)

        self.__execute_cmd(True, True, NULL, NULL,
                           CTRL_HDD_LED_ARG,
                           CTRL_HDD_LED_VAR_INDEX_LIST,
                           [HDD_LED_ON_OP, blight_hdd_map_opt])
        return

    def __get_hdd_led_brightness_cmdline(self, target_level):
        """Get the command line list that controls the
           brightness of HDD LED.
        """

        level_cmd_pair = \
            {BRIGHTNESS_LEVEL1_EL: SET_HDD_LED_BRIGHTNRSS_LEVEL1_CMD,
             BRIGHTNESS_LEVEL2_EL: SET_HDD_LED_BRIGHTNRSS_LEVEL2_CMD,
             BRIGHTNESS_LEVEL3_EL: SET_HDD_LED_BRIGHTNRSS_LEVEL3_CMD,
             BRIGHTNESS_LEVEL4_EL: SET_HDD_LED_BRIGHTNRSS_LEVEL4_CMD}

        return self.__get_value_by_key(target_level, level_cmd_pair)

    def __transact_hdd_brightness_node(self, hdd_led_brightness_node):
        """Transact HDD LED brightness node."""

        brightness = hdd_led_brightness_node.get_text()
        cmd = self.__get_hdd_led_brightness_cmdline(brightness)
        self.__execute_cmd(True, True, NULL, NULL, cmd)

    def __transact_hdd_led_node(self, hdd_led_node):
        """Transact hdd led node."""

        STATUS_XPATH = "statuslist/status"

        # HDD STATUS LED
        hdd_led_status_node_list = hdd_led_node.get_node_list(STATUS_XPATH)
        if len(hdd_led_status_node_list) > 0:
            self.__transact_hdd_led_status_node_list(hdd_led_status_node_list)

        return

    #
    # Request handlers
    #

    def __control_handler(self, entry):
        """Handle "control" request.
        """

        self.__check_correct_root_node(entry)

        target_led_node_list = entry.get().get_node_list(ANY_CHILD_NAME)
        for target_led_node in target_led_node_list:
            node_name = target_led_node.get_node_name()
            if POWER_NAME == node_name:
                self.__transact_power_led_node(target_led_node)
                continue

            if STATUS_NAME == node_name:
                self.__transact_status_led_node(target_led_node)
                continue

            if HDD_NAME == node_name:
                self.__transact_hdd_led_node(target_led_node)
                continue

            if BRIGHTNESS_NAME == node_name:
                self.__transact_hdd_brightness_node(target_led_node)
                continue

        entry.set_result(Results.OK)
        return entry

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.CONTROL, self.__control_handler)

#
# Main
#

if __name__ == "__main__":
    PalLed(option=palcmd.PalCmdOption()).handle_request()
