#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for setting network and getting network condition."""

from __future__ import with_statement

import os
import re
import errno
import socket
import struct
import shutil

import iobase.xmldata as xmldata
import iobase.netut as netut
import ionas.request as request

from libpal.util import RemountRoot
from iobase.exception import (ExecError, NotFoundError, InvalidArgError)
from ionas.confut import (build_config_file)
from iobase.cmdexec import (CmdExec, PIPE, NULL)
import libpal.palcmd as palcmd
import pal_config as config

albg_network_path = None
try:
    from albg_info import (ALBG_NETWORK_PATH)
    albg_network_path = ALBG_NETWORK_PATH
except:
    pass

from pal_info import (ReqTypes, Results)

IP_ADDR_SHOW_ARGS = [config.IP_CMD, "-o", "-f", "link", "addr", "show", "dev"]
FAMILY_NAME_OPT_IDX = 3

ROUTE_SHOW_ARGS = [config.ROUTE_CMD, "-n", "-A"]
GATEWAY_FLAG = "G"

ENABLE_WOL_ARGS = [config.ETHTOOL_CMD, "-s", "eth0", "wol", "g"]
ENABLE_WOL_DEVNAME_OPT_IDX = 2
ENABLE_WOL_ENABLE_OPT_IDX = 4
ENABLE_WOL_FLAG = "g"
DISABLE_WOL_FLAG = "d"

RESTART_NETWORK_ARG = [config.NETWORKING_INITD, "restart"]

FAMILY_LINK = "link"
FAMILY_INET = "inet"
FAMILY_INET6 = "inet6"

LOWER_UP_STR = "LOWER_UP"
STATE_UP = "up"
STATE_DOWN = "down"

NETWORK_NAME = "network"
FAMILYCONFIG_NAME = "familyconfig"
NETIF_LIST_NAME = "netiflist"
NETIF_NAME = "netif"
PROTOCOLLIST_NAME = "protocollist"
PROTOCOL_NAME = "protocol"
MAC_NAME = "mac"
DNS_NAME = "dns"
GATEWAY_NAME = "gateway"
DHCP_NAME = "dhcp"
ADDR_NAME = "addr"
ADDRLIST_NAME = "addrlist"
FAMILY_INET6_NAME = "inet6"
FAMILY_INET4_NAME = "inet4"

INET6_PATH = FAMILYCONFIG_NAME + "/" + FAMILY_INET6_NAME
INET4_PATH = FAMILYCONFIG_NAME + "/" + FAMILY_INET4_NAME
NETIF_PATH = NETIF_LIST_NAME + "/" + NETIF_NAME
PROTOCOL_PATH = ".//protocol"

DEV_ATTR = "dev"
STATE_ATTR = "state"
MTU_ATTR = "mtu"
FAMILY_ATTR = "family"
PREFIXLEN_ATTR = "prefixlen"
SCOPE_ATTR = "scope"
DYNAMIC_ATTR = "dynamic"
SOURCE_ATTR = "source"
NETMASK_ATTR = "netmask"
WOL_ATTR = "wol"
SPEED_ATTR = "speed"
UNKNOWN_SPEED = "unknown"

STATIC_VALUE = "static"


# templete part of conf file format
INET4_STATIC_IPADDR_FMT = \
    "iface %s inet static\n" + \
    "address %s\n" + \
    "netmask %s\n"

INET4_DHCP_FMT = \
    "iface %s inet dhcp\n"

INET4_DISABLE_FMT = \
    "iface %s inet manual\n"

MTU_FMT = "mtu %s\n"


class PalNetwork(palcmd.PalCmd):
    """PAL network implementation class."""

    def __initialize_conf_network_net_dir(self):
        """Initialize /mnt/data/conf/network/net directory."""

        if not os.path.exists(config.NETWORK_CONF_NETWORK_NET_DIR):
            os.makedirs(config.NETWORK_CONF_NETWORK_NET_DIR)
            return

        for file in os.listdir(config.NETWORK_CONF_NETWORK_NET_DIR):
            try:
                os.remove(
                    os.path.join(config.NETWORK_CONF_NETWORK_NET_DIR, file))
            except:
                pass

    def __is_inet6_only(self):
        """Check whtether """

        return config.INET6_ONLY

    def __is_inet4_enabled(self, network):
        """Check inet4 availability."""

        try:
            src_xpath = "netiflist/netif[@enable='true']" + \
                "/protocollist/protocol[@family='inet4'][@enable]"
            src_protocol_list = network.get_node_list(src_xpath)
            for src_protocol in src_protocol_list:
                if src_protocol.is_enabled():
                    return True
            return False
        except:
            return False

    def __is_inet6_enabled(self, network):
        """Check inet6 availability."""

        try:
            src_xpath = "netiflist/netif[@enable='true']" + \
                "/protocollist/protocol[@family='inet6'][@enable]"
            src_protocol_list = network.get_node_list(src_xpath)
            for src_protocol in src_protocol_list:
                if src_protocol.is_enabled():
                    return True
            return False
        except:
            return False

    def __check_body_node(self, node):
        """Check if entry's body node is valid."""

        if node.get_node_name() != NETWORK_NAME:
            raise InvalidArgError(node.get_node_name())

    def __force_make_dirs(self, dir_path):
        """make specified directory."""
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

    def __cidr2mask(self, cidr):
        """convert CIDR format subnet valut to netmask"""

        packed_subnet = \
            struct.pack('!L', 0xffffffff ^ ((1 << 32 - cidr) - 1))

        return socket.inet_ntoa(packed_subnet)

    def __get_dns_addr_list(self):
        """Return the DNS server list from "config.NETWORK_RESOLV_CONF" file.
        """

        NAMESERVER_RE = "nameserver\s+([^\s]+)"

        dns_list = []
        with open(config.NETWORK_RESOLV_CONF) as resolv_conf:
            for line in resolv_conf:
                match = re.match(NAMESERVER_RE, line)
                if not match:
                    continue

                dns_list.append(match.group(1))

        if not dns_list:
            self.debug_exception("No nameserver in " + config.RESOLV_CONF)
            raise NotFoundError(config.RESOLV_CONF, "nameserver")

        return dns_list

    def __is_family_addr(self, addr, family):
        """Return True if "addr" is specified family address format."""

        FAMILY_ADDR_CHECKER = {
            FAMILY_INET: netut.is_inet4_addr,
            FAMILY_INET6: netut.is_inet6_addr}

        try:
            if FAMILY_ADDR_CHECKER[family](addr):
                return True
        except:
            pass

        return False

    def __get_dns(self, family):
        """Return the node for the DNS setting."""

        addrlist = xmldata.XmlConfNode(ADDRLIST_NAME)
        dns = xmldata.XmlConfNode(DNS_NAME)
        dns.append_child(addrlist)

        try:
            for dns_addr in self.__get_dns_addr_list():
                if self.__is_family_addr(dns_addr, family):
                    addr = xmldata.XmlConfNode(ADDR_NAME)
                    addr.set_text(dns_addr)
                    addrlist.append_child(addr)
        except:
            self.debug_exception("No DNS address for " + family + "!!")

        return dns

    def __get_default_gateway_addr(self, family):
        """Return the node for the default gateway address."""

        DEFAULT_GATEWAY_RES = {
            FAMILY_INET: "[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+\s+" +
            "([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)+\s+" +
            "[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+\s+" +
            "([^\s]+)\s+.*",
            FAMILY_INET6: "[0-9a-zA-Z:]+/[0-9]+\s+" +
            "([0-9a-zA-Z:]+)\s+" + "([^\s]+)\s+.*"}

        cmd_args = list(ROUTE_SHOW_ARGS)
        cmd_args = cmd_args + [family]
        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL).execute()
        info_str = proc.stdout.read()

        if proc.wait() != 0:
            raise NotFoundError("gateway", "Command Error")

        for line in info_str.split("\n"):
            match = re.match(DEFAULT_GATEWAY_RES[family], line)

            if not match:
                continue

            addr_str = match.group(1)
            flags_str = match.group(2)

            if not GATEWAY_FLAG in flags_str:
                continue

            addr = xmldata.XmlConfNode(ADDR_NAME)
            addr.set_text(addr_str)
            return addr

        raise NotFoundError("gateway", "no match")

    def __get_gateway(self, family):
        """Return the node for the gateway setting."""

        gateway = xmldata.XmlConfNode(GATEWAY_NAME)

        try:
            gateway.append_child(self.__get_default_gateway_addr(family))
        except:
            self.debug_exception("No gateway address for " + family + "!!")

        return gateway

    def __get_ethtool_info(self, dev_name):
        """Get information with "ethtool" and return them as dictionaly."""

        WOL_RE = "^\s*Wake-on:\s*(.)\s*$"
        SPEED_RE = "^\s*Speed:\s*([0-9]+)[^0-9]*\s*$"
        LINK_RE =  "^\s*Link detected:\s*([a-zA-Z]+)\s*$"

        cmd_args = (config.ETHTOOL_CMD, dev_name)

        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL).execute()
        info_str = proc.stdout.read()
        if proc.wait() != 0:
            raise ExecError("__is_wol_enable")

        ethtool_info = {}
        is_link = False
        for line in info_str.split("\n"):
            match = re.match(WOL_RE, line)
            if match:
                if "g" == match.group(1):
                    ethtool_info[WOL_ATTR] = True
                else:
                    ethtool_info[WOL_ATTR] = False
                continue

            match = re.match(SPEED_RE, line)
            if match:
                ethtool_info[SPEED_ATTR] = match.group(1)
                continue

            match = re.match(LINK_RE, line)
            if match:
                if match.group(1).lower() == "yes":
                    is_link = True
                else:
                    is_link = False
                continue

        if not is_link and SPEED_ATTR in ethtool_info.keys():
            del ethtool_info[SPEED_ATTR]

        return ethtool_info

    def __is_wol_enable(self, dev_name):
        """Check wol enable/disable by ethtool command"""

        try:
            if (not dev_name in self.__ethtool_info or
                not WOL_ATTR in self.__ethtool_info[dev_name]):
                self.__ethtool_info[dev_name] = self.__get_ethtool_info(
                    dev_name)
            if WOL_ATTR in self.__ethtool_info[dev_name]:
                return self.__ethtool_info[dev_name][WOL_ATTR]
        except:
            pass

        return False

    def __get_link_speed(self, dev_name):
        """Return link speed by ethtool command"""

        try:
            if (not dev_name in self.__ethtool_info and
                not SPEED_ATTR in self.__ethtool_info[dev_name]):
                self.__ethtool_info[dev_name] = self.__get_ethtool_info(
                    dev_name)
            if SPEED_ATTR in self.__ethtool_info[dev_name]:
                return self.__ethtool_info[dev_name][SPEED_ATTR]
        except:
            pass

        return UNKNOWN_SPEED

    def __get_inet4(self):
        """Return the node for the inet4 family configuration."""

        inet4 = xmldata.XmlConfNode(FAMILY_INET4_NAME)
        inet4.append_child(self.__get_dns(FAMILY_INET))
        inet4.append_child(self.__get_gateway(FAMILY_INET))

        return inet4

    def __get_inet6(self):
        """Return the node for the inet6 family configuration."""

        inet6 = xmldata.XmlConfNode(FAMILY_INET6)
        inet6.append_child(self.__get_dns(FAMILY_INET6))
        inet6.append_child(self.__get_gateway(FAMILY_INET6))

        return inet6

    def __get_familyconfig(self):
        """Return the node for the family configuration."""

        familyconfig = xmldata.XmlConfNode(FAMILYCONFIG_NAME)
        familyconfig.append_child(self.__get_inet6())
        if not self.__is_inet6_only():
            familyconfig.append_child(self.__get_inet4())

        return familyconfig

    def __get_netdev_basic_info(self, dev_name):
        """Return the tuple of network devive basic information."""

        LINK_FAMILY_RE = ".*:\s*.*:\s*<([^>]*)>\s*mtu\s*([^\s]+).*" + \
                            "link/ether\s*([0-9a-fA-F:]+)\s*brd.*$"

        cmd_args = list(IP_ADDR_SHOW_ARGS)
        cmd_args[FAMILY_NAME_OPT_IDX] = FAMILY_LINK
        cmd_args = cmd_args + [dev_name]
        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL).execute()
        info_str = proc.stdout.read()
        if proc.wait() != 0:
            return ("", "", "")

        match = re.match(LINK_FAMILY_RE, info_str)
        if not match:
            return ("", "", "")

        flags_str = match.group(1)
        state = STATE_DOWN
        if LOWER_UP_STR in flags_str.split(","):
            state = STATE_UP
        mtu = match.group(2)
        mac_addr = match.group(3)
        return (state, mtu, mac_addr)

    def __to_dev_name(self, name):
        """Return the real device name if "name" is the logical network device
           name.
           Otherwise return the "name" value as is.
        """

        if name in config.NET_DEV_TO_RAW_DEV:
            return config.NET_DEV_TO_RAW_DEV[name]

        return name

    def __set_netif_basic_info(self, netif, name):
        """Set the "netif" to the basic information."""

        dev_name = self.__to_dev_name(name)
        (state, mtu, mac_addr) = self.__get_netdev_basic_info(dev_name)
        netif.set_attrs({DEV_ATTR: name, STATE_ATTR: state, MTU_ATTR: mtu})
        mac = xmldata.XmlConfNode(MAC_NAME)
        mac.set_text(mac_addr)
        netif.append_child(mac)

        # WOL
        if self.__is_wol_enable(dev_name):
            netif.set_bool_attr(WOL_ATTR, True)
        else:
            netif.set_bool_attr(WOL_ATTR, False)

        # Link speed
        netif.set_attrs({SPEED_ATTR: self.__get_link_speed(dev_name)})

        return netif

    def __get_netdev_inet4_info(self, dev_name):
        """Return the tuple of network device protocol information."""

        INET4_FAMILY_RE = ".*:\s*.*\s*inet\s*([0-9.]+)/([0-9]+)\s*" + \
            ".*scope\s*([^\s]+)"

        cmd_args = list(IP_ADDR_SHOW_ARGS)
        cmd_args[FAMILY_NAME_OPT_IDX] = FAMILY_INET
        cmd_args = cmd_args + [dev_name]
        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL).execute()
        info_str = proc.stdout.read()
        if proc.wait() != 0:
            raise NotFoundError("addr", "Command Error")

        ip_list = []
        for line in info_str.split("\n"):
            match = re.match(INET4_FAMILY_RE, line)
            if not match:
                continue

            ip = match.group(1)
            cidr_netmask = match.group(2)
            netmask = self.__cidr2mask(int(cidr_netmask))
            scope = match.group(3)

            ip_list.append((ip, netmask, scope))

        if not ip_list:
            raise NotFoundError("addr", "inet6")

        return ip_list

    def __get_netdev_inet6_info(self, dev_name):
        """Return the tuple of network device protocol information."""

        INET6_FAMILY_RE = ".*:\s*.*\s*inet6\s*([0-9a-fA-F:]+)/([0-9]+)\s*" + \
            "scope\s*([^\s]+)\s*(dynamic|)\s*.*$"

        cmd_args = list(IP_ADDR_SHOW_ARGS)
        cmd_args[FAMILY_NAME_OPT_IDX] = FAMILY_INET6
        cmd_args = cmd_args + [dev_name]
        proc = CmdExec(cmd_args, stdout=PIPE, stderr=NULL).execute()
        info_str = proc.stdout.read()
        if proc.wait() != 0:
            raise NotFoundError("addr", "Command Error")

        ip_list = []
        for line in info_str.split("\n"):
            match = re.match(INET6_FAMILY_RE, line)
            if not match:
                continue

            ip = match.group(1)
            prefixlen = match.group(2)
            scope = match.group(3)
            dynamic = "false"

            if match.group(4):
                dynamic = "true"

            ip_list.append((ip, prefixlen, scope, dynamic))

        if not ip_list:
            raise NotFoundError("addr", dev_name, "inet6")

        return ip_list

    def __set_netif_inet4_protocol_info(self, netif, name):
        """Set the "netif" to the inet4 protocol information."""

        try:
            ip_list = self.__get_netdev_inet4_info(self.__to_dev_name(name))
        except:
            return

        addrlist = xmldata.XmlConfNode(ADDRLIST_NAME)
        for ip_info in ip_list:
            (ip, netmask, scope) = ip_info

            addr = xmldata.XmlConfNode(ADDR_NAME)
            addr.set_text(ip)
            addr.set_attrs({NETMASK_ATTR: netmask, SCOPE_ATTR: scope})
            addrlist.append_child(addr)

        protocol = xmldata.XmlConfNode(PROTOCOL_NAME)
        protocol.set_attrs({FAMILY_ATTR: FAMILY_INET4_NAME})
        protocol.append_child(addrlist)

        try:
            protocollist = netif.get_node(PROTOCOLLIST_NAME)
            protocollist.append_child(protocol)
        except NotFoundError:
            protocollist = xmldata.XmlConfNode(PROTOCOLLIST_NAME)
            protocollist.append_child(protocol)
            netif.append_child(protocollist)

        return netif

    def __set_netif_inet6_protocol_info(self, netif, name):
        """Set the "netif" to the inet6 protocol information."""

        try:
            ip_list = self.__get_netdev_inet6_info(self.__to_dev_name(name))
        except:
            return

        addrlist = xmldata.XmlConfNode(ADDRLIST_NAME)
        for ip_info in ip_list:
            (ip, prefixlen, scope, dynamic) = ip_info

            addr = xmldata.XmlConfNode(ADDR_NAME)
            addr.set_text(ip)
            addr.set_attrs({PREFIXLEN_ATTR: prefixlen, SCOPE_ATTR: scope,
                                DYNAMIC_ATTR: dynamic})
            addrlist.append_child(addr)

        protocol = xmldata.XmlConfNode(PROTOCOL_NAME)
        protocol.set_attrs({FAMILY_ATTR: FAMILY_INET6})
        protocol.append_child(addrlist)
        if self.__is_inet6_only():
            netif.append_child(protocol)
        else:
            protocollist = xmldata.XmlConfNode(PROTOCOLLIST_NAME)
            protocollist.append_child(protocol)
            netif.append_child(protocollist)

        return netif

    def __set_netif_protocol_info(self, netif, name):
        """Set the "netif" to the protocol information."""

        self.__set_netif_inet6_protocol_info(netif, name)
        if not self.__is_inet6_only():
            self.__set_netif_inet4_protocol_info(netif, name)

    def __get_netif(self, name):
        """Return the node for the specified network device's information."""

        netif = xmldata.XmlConfNode(NETIF_NAME)
        self.__set_netif_basic_info(netif, name)
        self.__set_netif_protocol_info(netif, name)
        return netif

    def __get_netif_list(self):
        """Return the node for the "netif" list."""

        netif_list = xmldata.XmlConfNode(NETIF_LIST_NAME)
        for name in config.NET_DEV_LIST:
            netif = self.__get_netif(name)
            netif_list.append_child(netif)

        return netif_list

    def __query_handler(self, entry):
        """Handler for "query" request."""

        network = entry.get()
        self.__check_body_node(network)

        network.delete_children(xmldata.ANY_CHILD_PATH)
        network.append_child(self.__get_familyconfig())
        network.append_child(self.__get_netif_list())

        entry.set_result(Results.OK)
        return entry

    def __enable_netif(self, dev_name, is_netif_enable):
        """Create netif devname to enable(or disable) netif.
           The module which enabling(or disabling) netif is
           /etc/init.d/networking.
        """

        path = os.path.join(config.NETWORK_CONF_NETWORK_NET_DIR, dev_name)

        if is_netif_enable:
            self.__touch_file(path)
        else:
            self.__remove_file(path)
            self.__remove_network_interfaces_part_files(dev_name)

    def __set_mtu(self, dev_name, mtu):
        """Set MTU value to system"""

        # /mnt/data/conf/network/interfaces/interfaces.mtu.eth?
        mtu_part_file_path = os.path.join(
            config.NETWORK_CONF_NETWORK_INTERFACES_DIR,
            config.NETWORK_CONF_INTERFACES_MTU_PART + "." + dev_name)
        self.__force_make_dirs(os.path.dirname(mtu_part_file_path))
        with open(mtu_part_file_path, "w") as tmp_fh:
            tmp_fh.write(MTU_FMT % mtu)

        # /mnt/data/conf/network/mtu.eth?
        mtu_part_file_path = os.path.join(
            config.NETWORK_CONF_NETWORK_DIR,
            config.NETWORK_CONF_MTU_PART + "." + dev_name)
        with open(mtu_part_file_path, "w") as tmp_fh:
            tmp_fh.write(mtu + "\n")

    def __enable_wol(self, dev_name, is_enable_wol):
        """Enable(or disable) WOL"""

        cmd_args = list(ENABLE_WOL_ARGS)
        cmd_args[ENABLE_WOL_DEVNAME_OPT_IDX] = dev_name

        if is_enable_wol:
            cmd_args[ENABLE_WOL_ENABLE_OPT_IDX] = ENABLE_WOL_FLAG
        else:
            cmd_args[ENABLE_WOL_ENABLE_OPT_IDX] = DISABLE_WOL_FLAG

        CmdExec(cmd_args,
                stdin=NULL, stdout=NULL, stderr=NULL,
                raise_exc=True, close_fds=True).execute()

    def __touch_file(self, name):
        """Touch a file "name". It is created if it dosen't exist."""

        with open(name, "a"):
            pass

    def __remove_file(self, name):
        """Remove a file "name".  "OSError" with "ENOENT" is ignored."""

        try:
            os.remove(name)
        except OSError, err:
            if err.errno != errno.ENOENT:
                raise

    def __remove_config_for_all_sources(self, name):
        """Remove all config files for all sources."""

        for suffix in config.NET_RAW_DEV_LIST + (STATIC_VALUE,):
            self.__remove_file(name + "." + suffix)

    def __apply_inet4_enable(self, is_enabled, is_netif_enabled):
        """Apply inet4 "enable" attribute setting."""

        if is_enabled and is_netif_enabled:
            self.__touch_file(config.NETWORK_CONF_INET4_CONF)
        else:
            self.__remove_file(config.NETWORK_CONF_INET4_CONF)

    def __apply_inet6_enable(self, is_enabled, is_netif_enabled):
        """Apply inet6 "enable" attribute setting."""

        if is_enabled and is_netif_enabled:
            self.__touch_file(config.NETWORK_CONF_INET6_CONF)
        else:
            self.__remove_file(config.NETWORK_CONF_INET6_CONF)

    def __apply_dns_config(self, family, dns):
        """Apply DNS configuration."""

        path = "%(BASE)s.%(AF)s" % {
            "BASE": config.NETWORK_CONF_DNS_CONF_BASE,
            "AF": family}
        self.__remove_config_for_all_sources(path)

        try:
            source = self.__to_dev_name(dns.get_attr_value(SOURCE_ATTR))
        except NotFoundError:
            self.debug_exception("No DNS source for " + family + "!!")
            return

        conf_path = "%(PATH)s.%(SOURCE)s" % {
            "PATH": path,
            "SOURCE": source}
        with open(conf_path, "w") as conf_file:
            if source == STATIC_VALUE:
                try:
                    addrlist = dns.get_node(ADDRLIST_NAME)
                    for addr in addrlist.get_node_list(ADDR_NAME):
                        conf_file.write(addr.get_text() + "\n")
                except:
                    raise InvalidArgError(DNS_NAME, ADDR_NAME)

    def __apply_gateway_config(self, family, gateway):
        """Apply gateway configuration."""

        path = "%(BASE)s.%(AF)s" % {
            "BASE": config.NETWORK_CONF_GATEWAY_CONF_BASE,
            "AF": family}
        self.__remove_config_for_all_sources(path)

        try:
            source = self.__to_dev_name(gateway.get_attr_value(SOURCE_ATTR))
        except NotFoundError:
            self.debug_exception("No gateway source for " + family + "!!")
            return

        conf_path = "%(PATH)s.%(SOURCE)s" % {
            "PATH": path,
            "SOURCE": source}
        with open(conf_path, "w") as conf_file:
            if source == STATIC_VALUE:
                try:
                    addr = gateway.get_node(ADDR_NAME)
                    conf_file.write(addr.get_text() + "\n")
                except:
                    raise InvalidArgError(GATEWAY_NAME, ADDR_NAME)

    def __apply_familyconfig_inet4(self, network):
        """Apply familyconfig of inet4."""

        try:
            inet4 = network.get_node(INET4_PATH)
        except NotFoundError:
            self.debug_exception(
                "No network configuration for " + FAMILY_INET + "!!")
            return

        if not self.__is_inet4_enabled(network):
            return

        try:
            self.__apply_dns_config(
                FAMILY_INET4_NAME, inet4.get_node(DNS_NAME))
        except NotFoundError:
            self.debug_exception(
                "No DNS configuration for %(AF)s!!" % {
                    "AF": FAMILY_INET})

        try:
            self.__apply_gateway_config(
                FAMILY_INET4_NAME, inet4.get_node(GATEWAY_NAME))
        except NotFoundError:
            self.debug_exception(
                "No Gateway configuration for %(AF)s!!" % {
                    "AF": FAMILY_INET})

    def __apply_familyconfig_inet6(self, network):
        """Apply familyconfig of inet6."""

        try:
            inet6 = network.get_node(INET6_PATH)
        except NotFoundError:
            self.debug_exception(
                "No network configuration for %(AF)s!!" % {
                    "AF": FAMILY_INET6})
            return

        if not self.__is_inet6_enabled(network):
            return

        try:
            self.__apply_dns_config(FAMILY_INET6, inet6.get_node(DNS_NAME))
        except NotFoundError:
            self.debug_exception(
                "No DNS configuration for %(AF)s!!" % {
                    "AF": FAMILY_INET6})

        try:
            self.__apply_gateway_config(
                FAMILY_INET6, inet6.get_node(GATEWAY_NAME))
        except NotFoundError:
            self.debug_exception(
                "No Gateway configuration for %(AF)s!!" % {
                    "AF": FAMILY_INET6})

    def __create_inet4_addr_part_file(self, dev_name, file_data):
        """Create inet4 address part file."""

        path = os.path.join(
            config.NETWORK_CONF_NETWORK_INTERFACES_DIR,
            config.NETWORK_CONF_INTERFACES_ADDR_PART + "." + dev_name)

        self.__force_make_dirs(os.path.dirname(path))

        with open(path, "w") as tmp_fh:
            tmp_fh.write("\n")
            tmp_fh.write(file_data)

    def __remove_network_interfaces_part_files(self, dev_name):
        """Remove network interfaces part files."""

        TARGET_PART_LIST = (
            config.NETWORK_CONF_INTERFACES_ADDR_PART,
            config.NETWORK_CONF_INTERFACES_MTU_PART)
        for target_part in TARGET_PART_LIST:
            path = os.path.join(
                config.NETWORK_CONF_NETWORK_INTERFACES_DIR,
                target_part + "." + dev_name)
            if os.path.isfile(path):
                os.remove(path)

        # 無効用の設定ファイルを用意しておく
        file_data = INET4_DISABLE_FMT % dev_name
        self.__create_inet4_addr_part_file(dev_name, file_data)

    def __build_network_net_files(self):
        """Build files of /mnt/data/etc/net."""

        if os.path.exists(config.NETWORK_ETC_NET_DIR):
            for file in os.listdir(config.NETWORK_ETC_NET_DIR):
                try:
                    os.remove(os.path.join(config.NETWORK_ETC_NET_DIR, file))
                except:
                    pass
        else:
            os.makedirs(config.NETWORK_ETC_NET_DIR)

        for file in os.listdir(config.NETWORK_CONF_NETWORK_NET_DIR):
            try:
                shutil.copyfile(
                    os.path.join(config.NETWORK_CONF_NETWORK_NET_DIR, file),
                    os.path.join(config.NETWORK_ETC_NET_DIR, file))
            except:
                pass

    def __build_network_interfaces(self):
        """Build /etc/network/interfaces."""

        conf_dir = config.NETWORK_CONF_NETWORK_INTERFACES_DIR

        part_conf_file_list = [
            os.path.join(conf_dir, config.NETWORK_CONF_INTERFACES_BASE_PART)]

        for dev_name in config.NET_RAW_DEV_LIST:
            addr_file_path = os.path.join(
                conf_dir,
                config.NETWORK_CONF_INTERFACES_ADDR_PART + "." + dev_name)
            mtu_file_path = os.path.join(
                conf_dir,
                config.NETWORK_CONF_INTERFACES_MTU_PART + "." + dev_name)

            if not os.path.isfile(addr_file_path):
                continue
            part_conf_file_list.append(addr_file_path)

            if not os.path.isfile(mtu_file_path):
                continue
            part_conf_file_list.append(mtu_file_path)

        with RemountRoot():
            build_config_file(
                config.NETWORK_INTERFACES, part_conf_file_list)

    def __apply_inet4_dhcp_config(self, dev_name, dhcp):
        """Apply DHCP configuration for inet4."""

        if dhcp.is_enabled():
            file_data = INET4_DHCP_FMT % (dev_name)
            self.__create_inet4_addr_part_file(dev_name, file_data)
        else:
            pass

    def __apply_inet6_dhcp_config(
        self, dev_name, dhcp, is_netif_enabled, is_protocol_enabled):
        """Apply DHCP configuration for inet6."""

        path = "%(BASE)s.%(DEV)s.%(AF)s" % {
            "BASE": config.NETWORK_CONF_DHCP_CONF_BASE,
            "DEV": dev_name,
            "AF": FAMILY_INET6}
        if dhcp.is_enabled() and is_netif_enabled and is_protocol_enabled:
            self.__touch_file(path)
        else:
            self.__remove_file(path)

    def __apply_inet4_addr_config(self, dev_name, dhcp, addrlist):
        """Apply address configuration for inet4."""

        if dhcp.is_enabled():
            return

        try:
            ipaddr = addrlist.get_node(ADDR_NAME).get_text()
            netmask = addrlist.get_node(ADDR_NAME).get_attr_value(NETMASK_ATTR)
        except NotFoundError:
            self.exception("IP addr or netmask tag is not found")
            raise InvalidArgError("IP addr or netmask tag is not found")

        file_data = INET4_STATIC_IPADDR_FMT % (dev_name, ipaddr, netmask)
        self.__create_inet4_addr_part_file(dev_name, file_data)

    def __apply_inet6_addr_config(self, dev_name, addrlist):
        """Apply address configuration for inet6."""

        path = "%(BASE)s.%(DEV)s.%(AF)s" % {
            "BASE": config.NETWORK_CONF_ADDR_CONF_BASE,
            "DEV": dev_name,
            "AF": FAMILY_INET6}
        with open(path, "w") as conf_file:
            for addr in addrlist.get_node_list(ADDR_NAME):
                if not addr.get_text():
                    continue

                conf_file.write(
                    "%(ADDR)s/%(PLEN)s\n" % {
                        "ADDR": addr.get_text(),
                        "PLEN": addr.get_attr_value(PREFIXLEN_ATTR)})

    def __apply_netif_inet4_config(self, network):
        """Apply netif configuration for inet4 family."""

        for netif in network.get_node_list(NETIF_PATH):
            name = netif.get_attr_value(DEV_ATTR)

            if not name in config.NET_DEV_LIST:
                continue

            try:
                dev_name = self.__to_dev_name(name)
                protocol = netif.get_node(
                    PROTOCOL_PATH, {FAMILY_ATTR: FAMILY_INET4_NAME})
                if protocol.is_enabled():
                    self.__apply_inet4_enable(True, netif.is_enabled())
                else:
                    self.__remove_network_interfaces_part_files(dev_name)
                    continue
            except NotFoundError:
                self.debug_exception(
                    "No protocol configuration for %(AF)s!!" % {
                        "AF": FAMILY_INET})
                continue

            try:
                dhcp = protocol.get_node(DHCP_NAME)
                self.__apply_inet4_dhcp_config(dev_name, dhcp)
            except NotFoundError:
                self.debug_exception(
                    "No DHCP configuration for %(AF)s!!" % {
                        "AF": FAMILY_INET})

            try:
                self.__apply_inet4_addr_config(
                    dev_name, dhcp, protocol.get_node(ADDRLIST_NAME))
            except NotFoundError:
                self.debug_exception(
                    "No address configuration for %(AF)s!!" % {
                        "AF": FAMILY_INET})

    def __apply_netif_inet6_config(self, network):
        """Apply netif configuration for inet6 family."""

        for netif in network.get_node_list(NETIF_PATH):
            name = netif.get_attr_value(DEV_ATTR)

            if not name in config.NET_DEV_LIST:
                continue

            try:
                dev_name = self.__to_dev_name(name)
                is_netif_enabled = netif.is_enabled()
                protocol = netif.get_node(
                    PROTOCOL_PATH, {FAMILY_ATTR: FAMILY_INET6})
                is_protocol_enabled = protocol.is_enabled()
                if is_protocol_enabled == True:
                    self.__apply_inet6_enable(True, is_netif_enabled)
            except NotFoundError:
                self.debug_exception(
                    "No protocol configuration for %(AF)s!!" % {
                        "AF": FAMILY_INET6})
                continue

            try:
                dhcp = protocol.get_node(DHCP_NAME)
                self.__apply_inet6_dhcp_config(
                    dev_name, dhcp,
                    is_netif_enabled, is_protocol_enabled)
            except NotFoundError:
                self.debug_exception(
                    "No DHCP configuration for %(AF)s!!" % {
                        "AF": FAMILY_INET6})

            try:
                self.__apply_inet6_addr_config(
                    dev_name, protocol.get_node(ADDRLIST_NAME))
            except NotFoundError:
                self.debug_exception(
                    "No address configuration for %(AF)s!!" % {
                        "AF": FAMILY_INET6})

    def __apply_netif_common_config(self, network):
        """Apply netif common configuration."""

        for netif in network.get_node_list(NETIF_PATH):
            name = netif.get_attr_value(DEV_ATTR)

            if not name in config.NET_DEV_LIST:
                continue

            try:
                dev_name = self.__to_dev_name(name)
            except NotFoundError:
                self.debug_exception("netif name %s not found!!" % name)
                continue

            try:
                is_netif_enable = netif.is_enabled()
                self.__enable_netif(dev_name, is_netif_enable)
            except NotFoundError:
                pass

            if not is_netif_enable:
                return

            try:
                mtu = netif.get_attr_value(MTU_ATTR)
                self.__set_mtu(dev_name, mtu)
            except NotFoundError:
                pass

            try:
                netif.get_attr_value(WOL_ATTR)
                is_enable_wol = netif.is_set_to_true(WOL_ATTR)
                self.__enable_wol(dev_name, is_enable_wol)
            except NotFoundError:
                pass

    def __apply_handler(self, entry):
        """Handler for "apply" request."""

        network = entry.get()
        self.__check_body_node(network)
        self.__initialize_conf_network_net_dir()

        try:
            self.__apply_netif_common_config(network)

            if not self.__is_inet6_only():
                self.__apply_netif_inet4_config(network)
            self.__apply_netif_inet6_config(network)

            if not self.__is_inet6_only():
                self.__apply_familyconfig_inet4(network)
            self.__apply_familyconfig_inet6(network)

            # build real files
            self.__build_network_net_files()
            if not self.__is_inet6_only():
                self.__build_network_interfaces()

            if albg_network_path:
                restart_entry = request.Entry(ReqTypes.RESTART)
                self.send_entry(restart_entry, ALBG_NETWORK_PATH)

            entry.set_result(Results.OK)

        except ExecError:
            entry.set_result(Results.APPLY_ERROR)

        return entry

    def __restart_handler(self, entry):
        """Handler for "restart" request."""

        CmdExec(
            RESTART_NETWORK_ARG, stdin=NULL, stdout=NULL, stderr=NULL,
            raise_exc=True, close_fds=True).execute()
        entry.set_result(Results.OK)
        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.APPLY, self.__apply_handler)
        self.set_handler(ReqTypes.RESTART, self.__restart_handler)

        self.__ethtool_info = {}

#
# Main
#

if __name__ == "__main__":
    PalNetwork(option=palcmd.PalCmdOption()).handle_request()
