#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for managing packages."""


from __future__ import with_statement

import sys
import os
import re
import tarfile
import shutil
import glob
import errno

import iobase.xmldata as xmldata
from iobase.exception import (ExecError, NotFoundError, InvalidArgError)
from iobase.cmdexec import (CmdExec, PIPE, NULL)

import libpal.palcmd as palcmd
import libpal.util as util
import pal_config as config
from pal_info import (ReqTypes, Results)


PACKAGES_NAME = "packages"
INTERNALNAME_PATH = "package/internalname"


SCRIPTS_DIR = "scripts"
FILES_SUFFIX = ".files"
PRE_INSTALL_CMD = "pre_install"
INSTALL_CMD = "install"
POST_INSTALL_CMD = "post_install"
PRE_UNINSTALL_CMD = "pre_uninstall"
UNINSTALL_CMD = "uninstall"
POST_UNINSTALL_CMD = "post_uninstall"


class PalPackages(palcmd.PalCmd):
    """PAL packages implementation class."""

    def __check_body_node(self, node):
        """Check if entry's body node is valid."""

        if node.get_node_name() != PACKAGES_NAME:
            raise InvalidArgError(node.get_node_name())

    def __untar(self, tar_file, path="."):
        """Untar "tar_file" on the "path" directory."""

        tarf = tarfile.open(tar_file)
        for member in tarf.getnames():
            tarf.extract(member, path)

    def __copy_scripts(self, src_dir, internal_name):
        """Copy script files for package."""

        src = os.path.join(src_dir, SCRIPTS_DIR)
        dest = os.path.join(config.PACKAGE_SCRIPTS_DIR, internal_name)
        if os.path.isdir(src):
            try:
                shutil.rmtree(dest)
            except:
                pass

            shutil.copytree(src, dest)

    def __get_script_path(self, internal_name, script_name):
        """Return script path for "script_name"."""

        return os.path.join(config.PACKAGE_SCRIPTS_DIR, internal_name,
                                script_name)

    def __execute_script(self, internal_name, script_name, untar_dir=""):
        """Execute script file for package."""

        script_path = self.__get_script_path(internal_name, script_name)
        if not os.path.isfile(script_path):
            return

        cmd_args = [script_path]
        if untar_dir:
            cmd_args.append(untar_dir)
        CmdExec(cmd_args, cwd=os.path.dirname(script_path),
                stdin=NULL, stdout=NULL, stderr=NULL, close_fds=True).execute()

    def __get_files_path(self, internal_name):
        """Return the installed file path to package's "files"."""

        return os.path.join(config.PACKAGE_FILES_DIR,
                                internal_name + FILES_SUFFIX)

    def __copy_files_file(self, src_dir, internal_name):
        """Copy "files" file for package."""

        files_src_path = os.path.join(src_dir, internal_name + FILES_SUFFIX)
        files_path = self.__get_files_path(internal_name)
        if not os.path.exists(os.path.dirname(files_path)):
            os.makedirs(os.path.dirname(files_path))

        shutil.copyfile(files_src_path, self.__get_files_path(internal_name))

    def __remove(self, path):
        """Remove "path" file or directory."""

        try:
            if not os.path.islink(path) and os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)
        except (OSError, IOError), err:
            if err.errno != errno.ENOENT:
                raise

    def __remove_package_files(self, internal_name):
        """Remove files in package based on package's "files"."""

        files_path = self.__get_files_path(internal_name)
        files_file = open(files_path)
        dir_list = []
        for path in files_file:
            real_path = os.path.join(config.PACKAGE_ROOT_DIR, path.strip())
            if not os.path.islink(real_path) and os.path.isdir(real_path):
                dir_list.append(real_path)
            else:
                self.__remove(real_path)

        for dir in dir_list:
            if len(os.listdir(dir)) != 0:
                continue
            self.__remove(dir)

        os.remove(files_path)

    def __delete_scripts(self, internal_name):
        """Delete script files for package."""

        scripts_dir = os.path.join(config.PACKAGE_SCRIPTS_DIR, internal_name)
        if os.path.exists(scripts_dir):
            shutil.rmtree(scripts_dir)

    def __install_handler(self, entry):
        """Handler for "install" request."""

        package_node = entry.get()
        self.__check_body_node(package_node)

        with util.RemountRoot():
            package_file = entry.get_file()
            internal_name = package_node.get_node(INTERNALNAME_PATH).get_text()

            try:
                untar_dir = config.PACKAGE_UNTAR_DIR_BASE + \
                                    "." + str(os.getpid())
                if not os.path.exists(untar_dir):
                    os.makedirs(untar_dir)

                self.__untar(package_file, untar_dir)
                self.__copy_scripts(untar_dir, internal_name)
                self.__execute_script(internal_name,
                            PRE_INSTALL_CMD, untar_dir)
                if config.PACKAGE_TYPE == config.PACKAGE_TYPE_PAL:
                    self.__untar(
                            os.path.join(untar_dir, config.PACKAGE_TAR_FILE),
                            config.PACKAGE_ROOT_DIR)
                    self.__copy_files_file(untar_dir, internal_name)
                elif config.PACKAGE_TYPE == config.PACKAGE_TYPE_SELF:
                    self.__execute_script(internal_name,
                            INSTALL_CMD, untar_dir)
                else:
                    pass
                self.__execute_script(internal_name,
                            POST_INSTALL_CMD, untar_dir)
                entry.set_result(Results.OK)
            except:
                self.exception("install error: " + internal_name)
                entry.set_result(Results.INSTALL_ERROR)
            finally:
                try:
                    shutil.rmtree(untar_dir)
                except:
                    pass

        return entry

    def __uninstall_handler(self, entry):
        """Handler for "uninstall" request."""

        package_node = entry.get()
        self.__check_body_node(package_node)

        with util.RemountRoot():
            internal_name = package_node.get_node(INTERNALNAME_PATH).get_text()
            if not os.path.exists(
                    os.path.join(config.PACKAGE_SCRIPTS_DIR, internal_name)):

                self.error("not installed: " + internal_name)
                entry.set_result(Results.UNINSTALL_ERROR)
                return entry

            try:
                self.__execute_script(internal_name, PRE_UNINSTALL_CMD)
                if config.PACKAGE_TYPE == config.PACKAGE_TYPE_PAL:
                    self.__remove_package_files(internal_name)
                elif config.PACKAGE_TYPE == config.PACKAGE_TYPE_SELF:
                    self.__execute_script(internal_name, UNINSTALL_CMD)
                else:
                    pass
                self.__execute_script(internal_name, POST_UNINSTALL_CMD)
                self.__delete_scripts(internal_name)
                entry.set_result(Results.OK)
            except:
                self.exception("uninstall error: " + internal_name)
                entry.set_result(Results.UNINSTALL_ERROR)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.INSTALL, self.__install_handler)
        self.set_handler(ReqTypes.UNINSTALL, self.__uninstall_handler)


#
# Main
#

if __name__ == "__main__":
    PalPackages(option=palcmd.PalCmdOption()).handle_request()
