#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Raid."""

from __future__ import with_statement

import os
import re
import time
import glob
from iobase.xmldata import XmlNode
import libpal.palcmd as palcmd
from libpal.crond import ctrl as crond
import iobase.fileut as fileut
from iobase.cmdexec import (CmdExec, PIPE, NULL)
from iobase.exception import (ExecError, InvalidArgError, NotFoundError)
from pal_info import (ReqTypes, Results)
import pal_config as config
import ionas.device as device
import libpal.util as util

from pal_config import (MDADM_CMD, MDADM_INITD, DD_CMD, BLOCKDEV_CMD,
                        PS_CMD, CRON_ACTIVEREPAIR_CMD, KILLALL_CMD,
                        UDEVADM_CMD, LS_CMD, MKNOD_CMD)
from albg_info import ALBG_ACTIVEREPAIR_PATH
from pal_config import DEV_DIR
from pal_config import (RAID_X_SIZE_FILE, RAID_X_SLAVES_DIR,
                        RAID_X_SYNCACT_FILE, RAID_X_SYNCCOMPLETE_FILE,
                        RAID_X_LEVEL_FILE, RAID_X_ELEMENT_STATE_FILE)
from pal_config import (RAID_CACHE_FILE_TMPL)

### OPTION ###
FORCE_OPT = "force"
TRUE_OPT_VAL = "true"
FALSE_OPT_VAL = "false"
TYPE_OPT = "type"

### XML Tag Name ###
OPTION_NAME = "option"
RAIDCTRL_NAME = "raidctrl"
RAID_NAME = "raid"
NAME_NAME = "name"
LEVEL_NAME = "level"
NUMBER_NAME = "number"
SIZE_NAME = "size"
PARTITIONLIST_NAME = "partitionlist"
STATUS_NAME = "status"
PROGRESS_NAME = "progress"
DAY_NAME = "day"
TIME_NAME = "time"
HOUR_NAME = "hour"
MINUTE_NAME = "minute"
DISK_NAME = "disk"
RAIDLIST_NAME = "raidlist"
RESULT_NAME = "result"
OPERATION_NAME = "operation"
ELEMENTLIST_NAME = "elementlist"

# OPTION
OPERATION_OPT = "operation"
SCHEDULE_OPT_VALUE = "schedule"
START_SERVICE_OP = "startservice"
STOP_SERVICE_OP = "stopservice"

# XML Path
SCHEDULE_XPATH = "./activerepair/schedule"

# XML Attribute
TYPE_ATTR = "type"
RAID_ATTR = "raid"
MISSING_ATTR = "missing"
ELEMENT_ATTR = "element"
JOIN_ATTR = "join"
REDUDANCY_ATTR = "redudancy"
CRASH_ATTR = "crash"
RESYNC_ATTR = "resync"
SILENT_ATTR = "silent"
REBUILD_TARGET_ATTR = "rebuildtarget"
CRASH_ATTR_VAL = "crash"
GET_OPERATION_ATTR_VAL = "getoperation"

# XML Attribute value
TRUE_ATTR_VAL = "true"

### Command and argment ###
#
# XXXXX_ARG_FORMAT is the prototype list of the command which inclues VAR.
# VAR is replaced dynamically in __execute_cmd() method using
# XXXXX_VAR_INDEX_LIST and some variables._FORMAT
#

VAR = "%s"
# FORMAT: mdadm -E [partition_name]*
MDADM_GETINFO_ARG_FORMAT = [MDADM_CMD, "-E", VAR]
MDADM_GETINFO_VAR_INDEX_LIST = [2]

# FORMAT: mdadm -AR [raid device] [partition1] [partition2] ...
MDADM_START_ARG = [MDADM_CMD, "-AR"]

# FORMAT: mdadm -AR --force [raid device] [partition 1] [partition 2]
MDADM_FORCE_START_ARG = [MDADM_CMD, "-AR", "--force"]

# FORMAT: mdadm [raid device name] --create --level=[RAID level]
#         --raid-devices=[device count] --force [partition1] [partition2]...
MDADM_CREATE_ARG_FORMAT = [MDADM_CMD, VAR, "--create", "--level=%s",
                           "--raid-devices=%s", "--force"]
MDADM_CREATE_VAR_INDEX_LIST = [1, 3, 4]

# FORMAT: mdadm --stop [raid device]
MDADM_STOP_ARG_FORMAT = [MDADM_CMD, "--stop", VAR]
MDADM_STOP_VAR_INDEX_LIST = [2]

# FORMAT: mdadm [raid device] --add [partition]
MDADM_ADD_ARG_FORMAT = [MDADM_CMD, VAR, "--add", VAR]
MDADM_ADD_VAR_INDEX_LIST = [1, 3]

# FORMAT: mdadm [raid device] --fail [partition]
MDADM_FAIL_ARG_FORMAT = [MDADM_CMD, VAR, "--fail", VAR]
MDADM_FAIL_VAR_INDEX_LIST = [1, 3]

# FORMAT: mdadm [raid device] --remove [partition]
MDADM_REMOVE_ARG_FORMAT = [MDADM_CMD, VAR, "--remove", VAR]
MDADM_REMOVE_VAR_INDEX_LIST = [1, 3]

# FORMAT: mdadm --zero-superblock [partition]
MDADM_ZEROCLR_ARG_FORMAT = [MDADM_CMD, "--zero-superblock", VAR]
MDADM_ZEROCLR_VAR_INDEX_LIST = [2]

# FORMAT: mdadm --detail [raid_path]
MDADM_DETAIL_ARG_FORMAT = [MDADM_CMD, "--detail", VAR]
MDADM_DETAIL_VAR_INDEX_LIST = [2]

# FORMAT: ps -C [process name]
PS_ARG_FORMAT = [PS_CMD, "-C", VAR]
PS_VAR_INDEX_LIST = [2]

# FORMAT: blockdev --flushbufs [device]
BLOCKDEV_FLSUSH_ARG_FORMAT = [BLOCKDEV_CMD, "--flushbufs", VAR]
BLOCKDEV_FLSUSH_VAR_INDEX_LIST = [2]

# FORMAT: blockdev --getsize64 [device]
BLOCKDEV_GETSIZE_ARG_FORMAT = [BLOCKDEV_CMD, "--getsize64", VAR]
BLOCKDEV_GETSIZE_VAR_INDEX_LIST = [2]

# FORMAT: blockdev --getss [device]
BLOCKDEV_SECTOR_SIZE_ARG_FORMAT = [BLOCKDEV_CMD, "--getss", VAR]
BLOCKDEV_SECTOR_SIZE_VAR_INDEX_LIST = [2]

# FORMAT: dd if=/dev/zero of=[パーティションX]
#    seek=[削除対象の先頭セクタ] count=256
DD_ZERO_CLEAR_ARG_FORMAT = [DD_CMD, "if=/dev/zero", "of=%s",
                            "seek=%s", "count=%s"]
DD_ZERO_CLEAR_VAR_INDEX_LIST = [2, 3, 4]

# FORMAT: mdamd start
MDADM_START_DAEMON_ARG = [MDADM_INITD, "start"]

# FORMAT: mdamd stop
MDADM_STOP_DAEMON_ARG = [MDADM_INITD, "stop"]

# FORMAT: killall mdadm
KILLALL_MDADM_DAEMON_ARG = [KILLALL_CMD, "mdadm"]

UDEV_SETTLE_ARG = [UDEVADM_CMD, "settle"]

# ls -l [device path]
GET_DEVNODE_INFO_ARG = [LS_CMD, "-l", VAR]
GET_DEVNODE_INFO_VAR_INDEX_LIST = [2]

# mknod [node path] b [major] [minor]
MAKE_DEVNODE_ARG = [MKNOD_CMD, VAR, "b", VAR, VAR]
MAKE_DEVNODE_VAR_INDEX_LIST = [1, 3, 4]

# RAID Status
FAIL_STATUS = "fail"
DEGRADE_STATUS = "degrade"
STOP_STATUS = "stop"
RUNNING_STATUS = "running"
REBUILDING_STATUS = "rebuilding"
ACTIVEREPAIR_STATUS = "activerepair"
CRASH_STATUS = "crash"

# RAID action(RAID_X_SYNCACT_FILE)
IDLE_ACTION = "idle"
CHECK_ACTION = "check"
RECOVER_ACTION = "recover"
RESYNC_ACTION = "resync"

MISSING_DEVICE = "missing"

### Some value ###
RAID6_MODE = "raid6"
RAID5_MODE = "raid5"
RAID1_MODE = "raid1"
RAID0_MODE = "raid0"

UUID = "uuid"
UPDATE_TIME = "updatetime"
EVENTS = "events"

NOT_FOUND = "NotFound"
EMTPTY_RAID_SIZE = "0"
RAIDINFO_SECTOR_SIZE = "256"
COMPLETED = "COMPLETED"
UNKNOWN = "Unknown"

ACTICVEREPAIR_CRON_ENTRY_NAME = "activerepair"
DAYS = ('Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat')

RAID_LOCK_NAME = "raid_info_cache"


class PalRaidCtrl(palcmd.PalCmd):
    """Pal Raid Ctrl command implementation class."""

    #
    # General Utilities functions
    #

    class __GetPartRaidInfoError(Exception):
        """Corresponding getting partition raid error exception."""
        pass

    class __CreateError(Exception):
        "Create Error"
        pass

    class __StartError(Exception):
        "Start Error"
        pass

    class __StopError(Exception):
        "Stop Error"
        pass

    class __DeviceNotFound(Exception):
        "Device Not Found"
        pass

    class __RemoveError(Exception):
        "Remove Error"
        pass

    class __ActiverepairError(Exception):
        "Activerepair Error"
        pass

    class __AlreadyStop(Exception):
        "Already Stop"
        pass

    def __validate_root_tag_name(self, entry):
        """Validate the root tag name."""

        root_tag_name = entry.get().get_node_name()
        if RAIDCTRL_NAME != root_tag_name:
            raise InvalidArgError("invalid root tag..")

    def __get_min_raid_element_info(self):

        _RAID_TO_ELEMENT = util.get_raid_element_info()
        min_raid_element_info = {}
        for raid_name in config.RAID_DEV_LIST:

            symbl_raid_name = None
            for symbl, real in config.RAID_TO_REAL_NAME.iteritems():
                if raid_name == real:
                    symbl_raid_name = symbl

            if symbl_raid_name:
                element_list = _RAID_TO_ELEMENT[symbl_raid_name]
            else:
                element_list = _RAID_TO_ELEMENT[raid_name]

            element_max_num = \
                len(element_list) - element_list.count(MISSING_DEVICE)

            min_raid_element_info[raid_name] = \
                {RAID6_MODE: element_max_num - 2 if element_max_num > 2 else 2,
                 RAID5_MODE: element_max_num - 1 if element_max_num > 1 else 1,
                 RAID1_MODE: 1,
                 RAID0_MODE: element_max_num}

        return min_raid_element_info

    def __get_max_raid_element_info(self):

        _RAID_TO_ELEMENT = util.get_raid_element_info()
        max_raid_element_info = {}
        for raid_name in config.RAID_DEV_LIST:

            symbl_raid_name = None
            for symbl, real in config.RAID_TO_REAL_NAME.iteritems():
                if raid_name == real:
                    symbl_raid_name = symbl

            if symbl_raid_name:
                element_list = _RAID_TO_ELEMENT[symbl_raid_name]
            else:
                element_list = _RAID_TO_ELEMENT[raid_name]

            element_max_num = \
                len(element_list) - element_list.count(MISSING_DEVICE)

            max_raid_element_info[raid_name] = \
                {RAID6_MODE: element_max_num,
                 RAID5_MODE: element_max_num,
                 RAID1_MODE: element_max_num,
                 RAID0_MODE: element_max_num}

        return max_raid_element_info

    def __replace_env_arg(self):

        # CGIから実行されたときにsbinのパスが存在しないため

        if "PATH" in os.environ:
            self.default_env_path = os.environ["PATH"]

        os.environ["PATH"] = \
            '/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin'

    def __restore_env_arg(self):

        if self.default_env_path:
            os.environ["PATH"] = self.default_env_path

    def __devname2real_devpath(self, devname):

        return os.path.realpath(os.path.join(DEV_DIR, devname))

    def __convert_real_devname(self, dev_name):
        """Convert synbol device name to real device name.
           If dev_name is already real this method just returns it.
        """

        dev_path = os.path.join(config.DEV_DIR, dev_name)
        return os.path.basename(os.path.realpath(dev_path))

    def __part_name2raw_part_name(self, part_name):

        kdev_name = self.dev_mgr.get_kdev_name(os.path.basename(part_name))
        if not kdev_name:
            raise NotFoundError
        return kdev_name

    def __disk_name2raw_disk_name(self, disk_name):

        kdev_name = self.dev_mgr.get_kdev_name(os.path.basename(disk_name))
        if not kdev_name:
            raise NotFoundError
        return kdev_name

    def __partition2disk(self, part_name):
        """Convert partition name to corresponding disk name."""

        match = re.match("([a-zA-Z]+[0-9]+)", part_name)

        if not match:
            raise NotFoundError

        disk_name = match.group(1)

        if disk_name not in config.SATA_HDD_DEV_LIST + config.USB_HDD_DEV_LIST:
            raise NotFoundError

        return disk_name

    def __is_exist_node(self, node, node_path):
        """Check whether node_path exists in specified node"""

        try:
            node.get_node(node_path)
        except NotFoundError:
            return False
        return True

    def __set_raid_cache_value(self, raid_name, file_name, value):
        """Save raid status to temp file"""

        file_path = os.path.join(RAID_CACHE_FILE_TMPL % (raid_name, file_name))
        dir_path = os.path.dirname(file_path)
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

        with fileut.FLock(config.LOCK_DIR, RAID_LOCK_NAME, exclusive=False):
            with open(file_path, "w") as fh:
                fh.write(value)
        return

    def __get_raid_cache_value(self, raid_name, file_name):
        """Load raid status to temp file"""

        file_path = \
            os.path.join(RAID_CACHE_FILE_TMPL % (raid_name, file_name))

        if not os.path.isfile(file_path):
            return STOP_STATUS

        with fileut.FLock(config.LOCK_DIR, RAID_LOCK_NAME, exclusive=False):
            with open(file_path, "r") as fh:
                return fh.read().rstrip()

    def __get_product_real_raiddev_name(self, raiddev_name):
        """Get real raiddev name from devinfo.py"""

        if raiddev_name in config.RAID_TO_REAL_NAME:
            return config.RAID_TO_REAL_NAME[raiddev_name]

        return raiddev_name

    def __get_product_symbl_raiddev_name(self, raiddev_name):
        """Get real raiddev name from devinfo.py"""

        for symbl, real in config.RAID_TO_REAL_NAME.iteritems():
            if raiddev_name == real:
                return symbl

        raise NotFoundError

    def __get_partition_path_list(self, partlist_node):

        PART_NAME_XPATH = "./partition/name"
        name_node_list = partlist_node.get_node_list(PART_NAME_XPATH)

        return [self.__devname2real_devpath(name_node.get_text()) \
                for name_node in name_node_list]

    def __force_make_dirs(self, dir_path):
        """make specified directory."""
        if not os.path.isdir(dir_path):
            os.makedirs(dir_path)

    def __execute_cmd(self, is_enable_exception, is_wait,
                      std_in, std_out, std_err,
                      cmd_fmt, cmd_var_idx_list=[], variable_list=[]):
        """Execute command and return Popen object
              cmd_fmt - specify command format list
              cmd_var_idx_list - specify variable index of cmd_fmt
              variable_list - variables corresponding to cmd_var_idx_list
        """

        cmd = cmd_fmt[:]

        if len(cmd_var_idx_list) != len(variable_list):
            raise ExecError

        for (loop_idx, var_idx) in enumerate(cmd_var_idx_list):
            cmd[var_idx] = cmd[var_idx] % variable_list[loop_idx]

        return CmdExec(cmd,
                       stdout=std_out,
                       stderr=std_err,
                       stdin=std_in,
                       wait=is_wait,
                       raise_exc=is_enable_exception,
                       close_fds=True).execute()

    def __is_running_process(self, process_name):
        """ Check whether specified process is running now"""

        try:
            self.__execute_cmd(True, True,
                               NULL, NULL, NULL,
                               PS_ARG_FORMAT,
                               PS_VAR_INDEX_LIST,
                               [process_name])
        except:
            return False

        return True

    def __get_device_path(self, raid_node, dev_type):

        dev_name = \
            raid_node.get_node(NAME_NAME, {TYPE_ATTR: dev_type}).get_text()

        return os.path.realpath(os.path.join(DEV_DIR, dev_name))

    def __get_device_name(self, raid_node, dev_type):

        return raid_node.get_node(NAME_NAME, {TYPE_ATTR: dev_type}).get_text()

    def __get_min_raiddev_num(self, raiddev_name, raidmode):
        """return minimum number of raid disk number
           of specified raid mode
        """

        try:
            return self.__get_min_raid_element_info()[raiddev_name][raidmode]
        except:
            # 要素に存在しない
            self.exception("cant get min num (%s - %s)" % \
                               (raiddev_name, raidmode))

        return UNKNOWN

    def __get_max_raiddev_num(self, raiddev_name, raidmode):
        """Get max raid device number"""

        try:
            return self.__get_max_raid_element_info()[raiddev_name][raidmode]
        except:
            # 要素に存在しない
            self.exception("cant get min num (%s - %s)" % \
                               (raiddev_name, raidmode))

        return UNKNOWN

    def __get_current_raid_level(self, raiddev_name):
        """Get current raid level"""

        fh = open(RAID_X_LEVEL_FILE % raiddev_name)
        level = fh.read().rstrip()
        fh.close()
        return level

    def __get_current_raid_element_num(self, raiddev_name):
        """Get current raid device number"""

        dir_path = RAID_X_SLAVES_DIR % raiddev_name
        slave_list = os.listdir(dir_path)
        return len(slave_list)

    def __is_faulty_element(self, raiddev_name, real_part_name):

        FAULTY_STATE = "faulty"

        element_state_file_path = \
            RAID_X_ELEMENT_STATE_FILE % (raiddev_name, real_part_name)

        if not os.path.exists(element_state_file_path):
            raise NotFoundError

        with open(element_state_file_path, "r") as fh:
            if FAULTY_STATE == fh.read().rstrip():
                return True
        return False

    def __is_spare_element(self, raiddev_name, real_part_name):

        SPARE_STATE = "spare"

        element_state_file_path = \
            RAID_X_ELEMENT_STATE_FILE % (raiddev_name, real_part_name)

        if not os.path.exists(element_state_file_path):
            self.debug("Not found " + element_state_file_path)
            return False

        with open(element_state_file_path, "r") as fh:
            if SPARE_STATE == fh.read().rstrip():
                return True
        return False

    def __get_element_name_list(self, raiddev_name):

        dir_path = RAID_X_SLAVES_DIR % raiddev_name

        if not os.path.exists(dir_path):
            return []

        real_element_list = os.listdir(dir_path)

        symbl_element_list = []
        for real_element_name in real_element_list:

            if real_element_name.startswith("md"):
                symbl_element_list.append(real_element_name)
                continue

            symbl_part_name = \
                self.__real_part_name2symbl_part_name(real_element_name)
            symbl_element_list.append(symbl_part_name)
        return symbl_element_list

    def __is_min_raiddev_num(self, raiddev_name):
        """Check whether min raid dev number"""

        cur_num = self.__get_current_raid_element_num(raiddev_name)
        cur_level = self.__get_current_raid_level(raiddev_name)
        min_num = self.__get_min_raiddev_num(raiddev_name, cur_level)
        if cur_num < min_num:
            return False
        return True

    def __is_redudancy(self, level, raiddev_name, element_list):
        """Check redudancy"""

        num = len(element_list)

        for element in element_list:
            real_element_name = self.__convert_real_devname(element)
            if self.__is_spare_element(raiddev_name, real_element_name):
                num -= 1

        min_num = self.__get_min_raiddev_num(raiddev_name, level)
        if num > min_num:
            return True
        return False

    def __is_max_raiddev_num(self, raiddev_name):
        """Check whether max raid dev number"""

        cur_num = self.__get_current_raid_element_num(raiddev_name)
        cur_level = self.__get_current_raid_level(raiddev_name)
        max_num = self.__get_max_raiddev_num(raiddev_name, cur_level)
        if cur_num < max_num:
            return False
        return True

    def __is_partition_join_to_raid(self, raiddev_name, part_name):

        sys_md_part_path = \
            os.path.join(RAID_X_SLAVES_DIR % raiddev_name, part_name)

        if os.path.exists(sys_md_part_path):
            return True

        return False

    def __get_current_raiddev_size(self, raiddev_name):
        """Get current raid device size"""

        try:
            fh = open(RAID_X_SIZE_FILE % raiddev_name)
            size = fh.read().rstrip()
            fh.close()
        except:
            return NOT_FOUND

        return size

    def __get_current_raiddev_action(self, raiddev_name):
        """Get current raid device size"""

        if not os.path.exists(RAID_X_SYNCACT_FILE % raiddev_name):
            raise NotFoundError

        fh = open(RAID_X_SYNCACT_FILE % raiddev_name, "r")
        action = fh.read().rstrip()
        fh.close()
        return action

    def __get_progress(self, raiddev_name):
        """Get the progress[%] of background command. """

        progress_file_path = RAID_X_SYNCCOMPLETE_FILE % raiddev_name

        if not os.path.exists(progress_file_path):
            return "0"

        progress_fh = open(progress_file_path, "r")
        progress = progress_fh.read()
        progress_fh.close()

        match = re.match(r"([0-9]+) *\/ *([0-9]+)$", progress)
        if match:
            # return progress as percentage
            #   ex. "12345/1000000" = 12345*100 / 1000000 [%]
            return (int(match.group(1)) * 100) / int(match.group(2))

        if re.match("delayed", progress):
            return "0"

        if re.match("none", progress):
            return COMPLETED

        return NOT_FOUND

    def __is_during_activerepair(self):

        for raiddev_name in config.RAID_DEV_LIST:

            try:
                action = self.__get_current_raiddev_action(raiddev_name)
            except NotFoundError:
                continue

            if CHECK_ACTION == action:
                return True

        return False

    def __is_rebuilding(self, raiddev_name):

        try:
            action = self.__get_current_raiddev_action(raiddev_name)
        except:
            return False

        if RECOVER_ACTION == action or RESYNC_ACTION == action:
            return True

        return False

    def __get_current_raid_status(self, raiddev_name):
        """Return current raid status"""

        if not os.path.exists(
                   os.path.join(DEV_DIR,
                       self.__get_product_real_raiddev_name(raiddev_name))):
            return STOP_STATUS

        if not self.__is_min_raiddev_num(raiddev_name):
            return CRASH_STATUS

        if not self.__is_max_raiddev_num(raiddev_name):
            return DEGRADE_STATUS

        if self.__is_during_activerepair():
            return ACTIVEREPAIR_STATUS

        size = self.__get_current_raiddev_size(raiddev_name)
        if NOT_FOUND == size or EMTPTY_RAID_SIZE == size:
            return STOP_STATUS

        # RAID0はsync_actionファイルが存在しないので
        if RAID0_MODE == self.__get_current_raid_level(raiddev_name):
            return RUNNING_STATUS

        action = self.__get_current_raiddev_action(raiddev_name)
        if IDLE_ACTION == action:
            return RUNNING_STATUS

        if RECOVER_ACTION == action:
            return REBUILDING_STATUS

        if RESYNC_ACTION == action:
            return REBUILDING_STATUS

        return STOP_STATUS

    def __disk_name2raid_element_name(self, raiddev_name, disk_name):

        try:
            symbl_raiddev_name = \
                self.__get_product_symbl_raiddev_name(raiddev_name)
        except:
            symbl_raiddev_name = raiddev_name

        for element in util.get_raid_element_info()[symbl_raiddev_name]:
            if element.startswith(disk_name):
                return element

        return ""

    def __check_raid_status_if_remove(self, raiddev_name, remove_target_name):

        CRASH_RESULT = "crash"
        NO_CRASH_RESULT = "nocrash"
        ALREADY_STOP = "alreadystop"
        ALREADY_CRASH = "alreadycrash"

        slave_dir = os.path.join(RAID_X_SLAVES_DIR % raiddev_name)
        slave_list = os.listdir(slave_dir)

        real_element_name = ""
        if remove_target_name in config.RAID_DEV_LIST:  # ex. md13の中のmd6
            real_element_name = remove_target_name
        else:
            symbl_element_name = \
                self.__disk_name2raid_element_name(raiddev_name,
                                                   remove_target_name)

            real_element_name = \
                self.__part_name2raw_part_name(symbl_element_name)

        cur_level = self.__get_current_raid_level(raiddev_name)

        # すでにcrashかstopしてる
        raid_info = self.__get_raid_info(raiddev_name)
        if STOP_STATUS == raid_info[STATUS_NAME]:
            return ALREADY_STOP

        if CRASH_STATUS == raid_info[STATUS_NAME]:
            return ALREADY_CRASH

        if not real_element_name in slave_list:
            # 対象のRAIDデバイスに参加してない場合は崩壊せず
            return NO_CRASH_RESULT

        # 冗長性がなく且つ対象ディスクがリビルド先でない場合はcrash
        if not self.__is_redudancy(cur_level, raiddev_name, slave_list):

            if not self.__is_spare_element(raiddev_name, real_element_name):
                if real_element_name in slave_list:
                    return CRASH_RESULT

        return NO_CRASH_RESULT

    def __get_raid_info(self, raiddev_name):
        """get raid information and save file"""

        def set_resync_attr(raiddev_name, info_dict):

            try:
                action = self.__get_current_raiddev_action(raiddev_name)
                if RESYNC_ACTION == action:
                    info_dict.update({RESYNC_ATTR: True})
                else:
                    info_dict.update({RESYNC_ATTR: False})
            except:
                pass

        info_dict = {}

        product_real_name = \
            self.__get_product_real_raiddev_name(raiddev_name)

        if product_real_name not in config.SYSTEM_RAID_DEV_LIST:
            saved_status = \
                self.__get_raid_cache_value(product_real_name, STATUS_NAME)
            if saved_status == STOP_STATUS or saved_status == CRASH_STATUS:
                info_dict.update({STATUS_NAME: saved_status})
                return info_dict

        # raid status
        status = self.__get_current_raid_status(raiddev_name)

        if status == STOP_STATUS or status == CRASH_STATUS:
            info_dict.update({STATUS_NAME: status})
            return info_dict

        info_dict.update({STATUS_NAME: status})
        if status == REBUILDING_STATUS or status == ACTIVEREPAIR_STATUS:
            progress = self.__get_progress(raiddev_name)
            info_dict.update({PROGRESS_NAME: str(progress)})
            set_resync_attr(raiddev_name, info_dict)

        if status == DEGRADE_STATUS:
            if self.__is_rebuilding(raiddev_name):
                progress = self.__get_progress(raiddev_name)
                info_dict.update({PROGRESS_NAME: str(progress)})
                set_resync_attr(raiddev_name, info_dict)

        element_list = self.__get_element_name_list(raiddev_name)

        # raid device number
        #element_num = self.__get_current_raid_element_num(raiddev_name)
        element_num = len(element_list)
        info_dict.update({NUMBER_NAME: str(element_num)})

        # raid device level
        raid_level = self.__get_current_raid_level(raiddev_name)
        info_dict.update({LEVEL_NAME: raid_level})

        # raid device size
        raiddev_size = self.__get_current_raiddev_size(raiddev_name)
        info_dict.update({SIZE_NAME: raiddev_size})

        # check redudancy
        if self.__is_redudancy(raid_level, raiddev_name, element_list):
            info_dict.update({REDUDANCY_ATTR: True})
        else:
            info_dict.update({REDUDANCY_ATTR: False})

        return info_dict

    def __append_raid_info(self, raid_node):
        """Append raid info to specified node"""

        raid_path = self.__get_device_path(raid_node, RAID_ATTR)

        real_raiddev_name = os.path.basename(raid_path)
        info_dict = self.__get_raid_info(real_raiddev_name)

        for tag_name, value in info_dict.items():
            if tag_name in [REDUDANCY_ATTR, RESYNC_ATTR]:
                continue
            raid_node.append_child(tag_name, value)

        if REDUDANCY_ATTR in info_dict.keys():
            raid_node.get_node(STATUS_NAME).\
                set_bool_attr(REDUDANCY_ATTR, info_dict[REDUDANCY_ATTR])

        if RESYNC_ATTR in info_dict.keys():
            raid_node.get_node(STATUS_NAME).\
                set_bool_attr(RESYNC_ATTR, info_dict[RESYNC_ATTR])

        return

    def __append_element_list(self, raid_node):

        raiddev_name = self.__get_device_name(raid_node, RAID_ATTR)
        real_raiddev_name = self.__convert_real_devname(raiddev_name)
        element_name_list = self.__get_element_name_list(real_raiddev_name)

        elementlist_node = XmlNode(ELEMENTLIST_NAME)
        element_xml = '<element><name>%s</name></element>'
        for element_name in element_name_list:

            element_node = XmlNode(element_xml % element_name)
            elementlist_node.append_child(element_node)

        raid_node.append_child(elementlist_node)

    def __parse_partition_raidinfo(self, result):
        """Parse the result of mdadm -E command"""

        info_dict = {}
        for line in result.split("\n"):
            match = re.search("Array UUID : (\S+)", line)

            if match:
                info_dict[UUID] = match.group(1)
                continue

            match = re.search(r"Update Time : (\S+\s+\S+\s+\S+\s+\S+\s+\S+)",
                              line)
            if match:
                info_dict[UPDATE_TIME] = match.group(1)
                continue

            match = re.search("Events : (\S+)", line)
            if match:
                info_dict[EVENTS] = match.group(1)
                continue

        if (UUID not in info_dict) or \
           (UPDATE_TIME not in info_dict) or \
           (EVENTS not in info_dict):
            raise self.__GetPartRaidInfoError

        return info_dict

    def __append_partition_raidinfo(self, raidctrl_node):
        """Get RAID info of specified partition """

        RAID_NAME_PATH = "./raid/name[@type='raid']"
        PART_XPATH = "./partitionlist/partition"

        raiddev_name = ""
        if self.__is_exist_node(raidctrl_node, RAID_NAME_PATH):
            raiddev_name = raidctrl_node.get_node(RAID_NAME_PATH).get_text()

        part_node_list = raidctrl_node.get_node_list(PART_XPATH)

        for part_node in part_node_list:

            part_name = part_node.get_node(NAME_NAME).get_text()

            # join属性
            if raiddev_name:
                real_raidev_name = self.__convert_real_devname(raiddev_name)
                real_part_name = self.__convert_real_devname(part_name)
                if self.__is_partition_join_to_raid(real_raidev_name,
                                                    real_part_name):
                    part_node.set_bool_attr(JOIN_ATTR, True)
                else:
                    part_node.set_bool_attr(JOIN_ATTR, False)

            # element情報
            part_path = self.__devname2real_devpath(part_name)
            proc = self.__execute_cmd(False, False,
                                      NULL, PIPE, NULL,
                                      MDADM_GETINFO_ARG_FORMAT,
                                      MDADM_GETINFO_VAR_INDEX_LIST,
                                      [part_path])
            result = proc.stdout.read()
            if 0 != proc.wait():
                part_node.set_bool_attr(ELEMENT_ATTR, False)
                continue

            try:
                info_dict = self.__parse_partition_raidinfo(result)
                for key, value in info_dict.items():
                    part_node.append_child(key, value)
                part_node.set_bool_attr(ELEMENT_ATTR, True)
            except self.__GetPartRaidInfoError:
                part_node.set_bool_attr(ELEMENT_ATTR, False)

        return

    def __execute_start_raid(self,
                             is_force,
                             raid_path,
                             partition_name_list):
        """Execute start RAID"""

        if is_force:
            cmd_list = list(MDADM_FORCE_START_ARG)
        else:
            cmd_list = list(MDADM_START_ARG)

        cmd_list.append(raid_path)
        cmd_list.extend(partition_name_list)

        self.__execute_cmd(True, True,
                           NULL, NULL, NULL,
                           cmd_list)
        return

    def __start_raid(self, raid_node, is_force):
        """Start RAID"""

        raid_path = self.__get_device_path(raid_node, RAID_ATTR)
        partlist_node = raid_node.get_node(PARTITIONLIST_NAME)
        part_path_list = self.__get_partition_path_list(partlist_node)

        time.sleep(5)
        self.__execute_start_raid(is_force, raid_path, part_path_list)
        self.__wait_raiddev_ready(raid_path)
        return

    def __do_udev_settle(self):
        """Excecute udev settle"""

        try:
            self.__execute_cmd(False, True,
                               NULL, NULL, NULL,
                               UDEV_SETTLE_ARG)
        except:
            pass

        return

    def __wait_raiddev_ready(self, raid_path):

        self.__do_udev_settle()

        for idx in range(5):
            try:
                fh = open(raid_path, "r")
                fh.close()
            except:
                time.sleep(2)
                continue
            break
        return

    def __mdadm_start_daemon(self):
        """Start mdadm daemon start """

        self.__replace_env_arg()

        self.__execute_cmd(True, True,
                           NULL, NULL, NULL,
                           MDADM_START_DAEMON_ARG)

        #for idx in range(5):
        for idx in range(1):
            time.sleep(2)
            if self.__is_running_process(os.path.basename(MDADM_INITD)):
                self.__restore_env_arg()
                return

            self.__execute_cmd(True, True,
                               NULL, NULL, NULL,
                               MDADM_START_DAEMON_ARG)

        self.__restore_env_arg()
        return

    def __mdadm_stop_daemon(self):
        """Start mdadm daemon stop"""

        self.__replace_env_arg()

        self.__execute_cmd(True, True,
                           NULL, NULL, NULL,
                           MDADM_STOP_DAEMON_ARG)
        self.__restore_env_arg()

        time.sleep(1)

        if not self.__is_running_process(os.path.basename(MDADM_INITD)):
            return

        self.__execute_cmd(True, True,
                           NULL, NULL, NULL,
                           KILLALL_MDADM_DAEMON_ARG)
        return

    def __stop_rebuild(self, raiddev_name):

        if not os.path.exists(RAID_X_SYNCACT_FILE % raiddev_name):
            return

        with open(RAID_X_SYNCACT_FILE % raiddev_name, "r") as fh:
            action = fh.read().rstrip()
            if not RECOVER_ACTION == action and \
               not RESYNC_ACTION == action:
                return

        with open(RAID_X_SYNCACT_FILE % raiddev_name, "w") as fh:
            fh.write(IDLE_ACTION)

        time.sleep(1)
        return

    def __execute_create_raid(self, raid_path, level,
                              part_num, part_path_list):

        proc = self.__execute_cmd(False, False,
                                  PIPE, NULL, NULL,
                                  MDADM_CREATE_ARG_FORMAT + part_path_list,
                                  MDADM_CREATE_VAR_INDEX_LIST,
                                  [raid_path, level, part_num])

        proc.stdin.write("y\n")  # mdadm ask whether cotinuing(y/n)
        proc.stdin.close()

        if 0 != proc.wait():
            return False

        return True

    def __create_raid(self, raid_node, is_silent=False):
        """Create RAID"""

        raid_path = self.__get_device_path(raid_node, RAID_ATTR)

        # level
        level = raid_node.get_node(LEVEL_NAME).get_text()

        # partition number
        number_node = raid_node.get_node(NUMBER_NAME)
        part_num = number_node.get_text()
        missing_attr = number_node.get_attr_value(MISSING_ATTR)

        # partition path list
        partlist_node = raid_node.get_node(PARTITIONLIST_NAME)
        part_path_list = self.__get_partition_path_list(partlist_node)
        if TRUE_ATTR_VAL == missing_attr:
            for idx in range(int(part_num) - len(part_path_list)):
                part_path_list.append("missing")

        if not is_silent:
            self.__mdadm_stop_daemon()

        create_result = self.__execute_create_raid(raid_path, level,
                                                   part_num, part_path_list)
        self.__wait_raiddev_ready(raid_path)

        if not is_silent:
            self.__mdadm_start_daemon()

        if False == create_result:
            raise self.__CreateError

        return

    def __stop_raid(self, raid_path):
        """Stop RAID"""

        if not os.path.exists(raid_path):
            raise self.__AlreadyStop

        # XRの処理をマネる
        for try_num in range(1, 10):
            proc = self.__execute_cmd(False, True,
                                      NULL, NULL, NULL,
                                      MDADM_STOP_ARG_FORMAT,
                                      MDADM_STOP_VAR_INDEX_LIST,
                                      [raid_path])
            if 0 == proc.returncode:
                break
        else:
            raise self.__StopError

        self.__wait_raiddev_ready(raid_path)
        return

    def __add_partition_to_raiddev(self, raid_path, part_path):
        """Add partition to raid device"""

        self.__execute_cmd(True, True,
                           NULL, NULL, NULL,
                           MDADM_ADD_ARG_FORMAT,
                           MDADM_ADD_VAR_INDEX_LIST,
                           [raid_path, part_path])
        return

    def __remove_partition_by_sysfs(self, real_raid_name, real_part_name):

        REMOVE_STATE = "remove"

        element_state_file = \
            RAID_X_ELEMENT_STATE_FILE % (real_raid_name, real_part_name)

        with open(element_state_file, "w") as fh:
            fh.write(REMOVE_STATE)

    def __fail_partition_by_sysfs(self, real_raid_name, real_part_name):

        FAULTY_STATE = "faulty"

        element_state_file = \
            RAID_X_ELEMENT_STATE_FILE % (real_raid_name, real_part_name)

        with open(element_state_file, "w") as fh:
            fh.write(FAULTY_STATE)

    def __remove_partition_from_raiddev(self, raid_path, part_path):
        """Remove partition from raid device"""

        # 強制取り外しした際はデバイスノードが消える.
        # ダミーのデバイスノードを使って取り外す場合構成ディスクの状態が
        # 更新されていないとfailマークがつけれないのでmdの状態を最新の状態に
        # する目的でmdadm --detailを行う

        if not os.path.exists(raid_path):
            return

        self.__stop_rebuild(os.path.basename(raid_path))

        self.__execute_cmd(False, False,
                           NULL, PIPE, NULL,
                           MDADM_DETAIL_ARG_FORMAT,
                           MDADM_DETAIL_VAR_INDEX_LIST,
                           [raid_path])

        time.sleep(3)

        self.__execute_cmd(False, True,
                           NULL, NULL, NULL,
                           BLOCKDEV_FLSUSH_ARG_FORMAT,
                           BLOCKDEV_FLSUSH_VAR_INDEX_LIST,
                           [raid_path])

        self.__execute_cmd(False, True,
                           NULL, NULL, NULL,
                           BLOCKDEV_FLSUSH_ARG_FORMAT,
                           BLOCKDEV_FLSUSH_VAR_INDEX_LIST,
                           [part_path])

        try:
            if not self.__is_faulty_element(os.path.basename(raid_path),
                                            os.path.basename(part_path)):
                self.__execute_cmd(True, True,
                                   NULL, NULL, NULL,
                                   MDADM_FAIL_ARG_FORMAT,
                                   MDADM_FAIL_VAR_INDEX_LIST,
                                   [raid_path, part_path])
        except:
            self.error("fail to set fail flag and try removing by sysfs")
            self.__fail_partition_by_sysfs(os.path.basename(raid_path),
                                           os.path.basename(part_path))

        self.__stop_rebuild(os.path.basename(raid_path))

        try:
            self.__execute_cmd(True, True, NULL, NULL, NULL,
                               MDADM_REMOVE_ARG_FORMAT,
                               MDADM_REMOVE_VAR_INDEX_LIST,
                               [raid_path, part_path])
        except:
            self.error("fail to remove partition and try removing by sysfs")
            self.__remove_partition_by_sysfs(os.path.basename(raid_path),
                                             os.path.basename(part_path))

        try:
            dev_path = self.__dummy_dev_path2real_dev_path(part_path)
            if not os.path.exists(dev_path):
                return
        except:
            self.exception("fail to __dummy_dev_path2real_dev_path")
            pass

        try:
            if not os.path.basename(part_path).startswith("md"):
                self.__clear_raidinfo(part_path)
        except:
            self.exception("fail to __clear_raidinfo")
            pass

        return

    def __clear_raid_superblock(self, part_path):
        """Zero clear RAID super block"""

        self.__execute_cmd(True, True,
                           NULL, NULL, NULL,
                           MDADM_ZEROCLR_ARG_FORMAT,
                           MDADM_ZEROCLR_VAR_INDEX_LIST,
                           [part_path])

        self.__execute_cmd(True, True,
                           NULL, NULL, NULL,
                           BLOCKDEV_FLSUSH_ARG_FORMAT,
                           BLOCKDEV_FLSUSH_VAR_INDEX_LIST,
                           [part_path])
        return

    def __get_raidinfo_top_sector(self, part_path):
        """Get the top sector number of RAID infomation block"""

        proc = self.__execute_cmd(True, True,
                                  NULL, PIPE, NULL,
                                  BLOCKDEV_GETSIZE_ARG_FORMAT,
                                  BLOCKDEV_GETSIZE_VAR_INDEX_LIST,
                                  [part_path])
        partition_size = proc.stdout.read().rstrip()

        proc = self.__execute_cmd(True, True,
                                  NULL, PIPE, NULL,
                                  BLOCKDEV_SECTOR_SIZE_ARG_FORMAT,
                                  BLOCKDEV_SECTOR_SIZE_VAR_INDEX_LIST,
                                  [part_path])
        sector_size = proc.stdout.read().rstrip()

        return str(int(partition_size) / int(sector_size))

    def __zero_clear_device(self, device_path, top_sector, count):
        """Zero clear specified range"""

        self.__execute_cmd(True, True,
                           NULL, NULL, NULL,
                           DD_ZERO_CLEAR_ARG_FORMAT,
                           DD_ZERO_CLEAR_VAR_INDEX_LIST,
                           [device_path, top_sector, count])
        return

    def __clear_raidinfo(self, part_path):
        """Clear RAID info"""

        try:
            self.__clear_raid_superblock(part_path)
        except:
            self.exception("fail to __clear_raid_superblock " + part_path)
            pass

        raidinfo_sector = self.__get_raidinfo_top_sector(part_path)

        self.__zero_clear_device(part_path, raidinfo_sector,
                                 RAIDINFO_SECTOR_SIZE)
        return

    def __set_activerepair_schedule(self, entry):

        sche_node = entry.get().get_node(SCHEDULE_XPATH)

        if not sche_node.is_enabled():
            crond.delete_entry(ACTICVEREPAIR_CRON_ENTRY_NAME)
            return

        day = sche_node.get_node(DAY_NAME).get_text()
        hour = sche_node.get_node(TIME_NAME).get_node(HOUR_NAME).get_text()
        min = sche_node.get_node(TIME_NAME).get_node(MINUTE_NAME).get_text()
        crond.add_entry(ACTICVEREPAIR_CRON_ENTRY_NAME,
                        CRON_ACTIVEREPAIR_CMD,
                        user='root',
                        week=str(dict(zip(DAYS, range(0, 7)))[day]),
                        hour=hour,
                        minutes=min)
        return

    def __get_devnode_major_minor_num(self, disk_path):

        proc = self.__execute_cmd(True, True,
                                  NULL, PIPE, NULL,
                                  GET_DEVNODE_INFO_ARG,
                                  GET_DEVNODE_INFO_VAR_INDEX_LIST,
                                  [disk_path])

        for line in proc.stdout.readlines():
            match = re.match('\S+\s+' + \
                             '\S+\s+' + \
                             '\S+\s+' + \
                             '\S+\s+' + \
                             '([0-9]+),\s+' + \
                             '([0-9]+)\s+',
                             line)
            if not match:
                continue

            major = match.group(1)
            minor = match.group(2)
            return major, minor

    def __make_device_node(self, node_path, major, minor):

        node_dir_path = os.path.dirname(node_path)
        if os.path.isdir(node_dir_path):
            old_node_list = glob.glob(os.path.join(node_dir_path, "*"))
            for old_node_path in old_node_list:
                os.remove(old_node_path)
        else:
            os.makedirs(node_dir_path)

        proc = self.__execute_cmd(False, True,
                                  NULL, NULL, NULL,
                                  MAKE_DEVNODE_ARG,
                                  MAKE_DEVNODE_VAR_INDEX_LIST,
                                  [node_path, major, minor])
        if 0 != proc.wait():
            # 作れなくてもエラーにするほどのことではない
            pass

        return

    def __dummy_dev_path2real_dev_path(self, dummy_dev_path):

        if not dummy_dev_path:
            raise NotFoundError

        return os.path.join(DEV_DIR, os.path.basename(dummy_dev_path))

    def __get_dummy_dev_node_file_path(self, dev_name):

        dev_path = self.dev_mgr.get_kdev_path(dev_name)
        if not dev_path:
            temp_dev_name = self.dev_mgr.get_dev_name(dev_name)
            if not temp_dev_name:
                return ""

            dev_path = self.dev_mgr.get_kdev_path(temp_dev_name)

        return dev_path

    def __real_part_name2symbl_part_name(self, real_part_name):

        try:
            return self.dev_mgr.get_dev_name(real_part_name)
        except:
            self.warning("Not found device name for " + real_part_name)
            return ""

    #
    # Request handlers
    #

    def __query_handler(self, entry):
        """Handle "query" request."""

        self.__validate_root_tag_name(entry)

        is_element = False
        try:
            is_element = \
                entry.get().get_node(OPTION_NAME).is_set_to_true(ELEMENT_ATTR)
        except:
            pass

        try:
            self.__append_raid_info(entry.get().get_node(RAID_NAME))
        except self.__DeviceNotFound:
            entry.set_result(Results.DEVICE_NOT_FOUND)
            return entry

        if is_element:
            self.__append_element_list(entry.get().get_node(RAID_NAME))

        entry.set_result(Results.OK)
        return entry

    def __get_element_info_handler(self, entry):
        """Handle "get_element_info" request."""

        self.__validate_root_tag_name(entry)

        try:
            self.__append_partition_raidinfo(entry.get())
        except self.__GetPartRaidInfoError:
            entry.set_result(Results.GET_ERROR)
        else:
            entry.set_result(Results.OK)

        return entry

    def __sub_start_handler(self, entry):
        """Handle "start" request."""

        self.__validate_root_tag_name(entry)

        force_opt = entry.get_options()[FORCE_OPT]

        if force_opt == TRUE_OPT_VAL:
            is_force = True
        elif force_opt == FALSE_OPT_VAL:
            is_force = False
        else:
            raise InvalidArgError("force option is invalid..")

        raiddev_name = \
            self.__get_device_name(entry.get().get_node(RAID_NAME), RAID_ATTR)
        real_raiddev_name = self.__get_product_real_raiddev_name(raiddev_name)

        try:
            self.__start_raid(entry.get().get_node(RAID_NAME), is_force)
        except:
            entry.set_result(Results.START_ERROR)
            self.__set_raid_cache_value(real_raiddev_name,
                                        STATUS_NAME, CRASH_STATUS)
            try:
                self.__stop_raid(os.path.join(DEV_DIR, real_raiddev_name))
            except:
                pass
        else:
            entry.set_result(Results.OK)
            self.__set_raid_cache_value(real_raiddev_name,
                                        STATUS_NAME, RUNNING_STATUS)

        return entry

    def __start_handler(self, entry):

        with util.RaidOperation(ReqTypes.START):
            try:
                return self.__sub_start_handler(entry)
            except:
                self.exception("fail to __sub_start_handler()")
                raise

    def __dummy_read_raid_info(self, raid_path):
        """Sometime the raid device cannot be removed from /dev after executing
           the command 'mdadm --stop /dev/md*'.
           However then we can remove the raid device by executing the command
           'mdadm --detail /dev/md*' info.
        """

        try:
            self.__execute_cmd(False, False,
                               NULL, PIPE, NULL,
                               MDADM_DETAIL_ARG_FORMAT,
                               MDADM_DETAIL_VAR_INDEX_LIST,
                               [raid_path])

            self.__execute_cmd(False, False,
                               NULL, PIPE, NULL,
                               MDADM_DETAIL_ARG_FORMAT,
                               MDADM_DETAIL_VAR_INDEX_LIST,
                               [raid_path])
        except:
            pass

    def __sub_stop_handler(self, entry):
        """Handle "stop" request."""

        self.__validate_root_tag_name(entry)

        is_crash = False
        try:
            is_crash = \
                entry.get().get_node(OPTION_NAME).is_set_to_true(CRASH_ATTR)
        except:
            pass

        raiddev_name = \
            self.__get_device_name(entry.get().get_node(RAID_NAME), RAID_ATTR)
        real_raiddev_name = self.__get_product_real_raiddev_name(raiddev_name)

        try:
            raid_path = \
                self.__get_device_path(entry.get().get_node(RAID_NAME),
                                       RAID_ATTR)
            self.__stop_raid(raid_path)
            self.__dummy_read_raid_info(raid_path)
        except self.__StopError:
            entry.set_result(Results.STOP_ERROR)
        except self.__AlreadyStop:
            entry.set_result(Results.ALREADY_STOP)
        else:
            entry.set_result(Results.OK)
        finally:
            if is_crash:
                self.__set_raid_cache_value(real_raiddev_name,
                                            STATUS_NAME, CRASH_STATUS)
            else:
                saved_status = \
                    self.__get_raid_cache_value(real_raiddev_name, STATUS_NAME)

                if CRASH_STATUS != saved_status:
                    self.__set_raid_cache_value(real_raiddev_name,
                                                STATUS_NAME, STOP_STATUS)

        return entry

    def __stop_handler(self, entry):

        with util.RaidOperation(ReqTypes.STOP):
            try:
                return self.__sub_stop_handler(entry)
            except:
                self.exception("fail to __sub_stop_handler()")
                raise

    def __sub_create_handler(self, entry):
        """Handle "create" request."""

        self.__validate_root_tag_name(entry)

        is_silent = False
        try:
            is_silent = \
                entry.get().get_node(OPTION_NAME).is_set_to_true(SILENT_ATTR)
        except:
            pass

        raiddev_name = \
            self.__get_device_name(entry.get().get_node(RAID_NAME), RAID_ATTR)
        real_raiddev_name = self.__get_product_real_raiddev_name(raiddev_name)

        try:
            self.__create_raid(entry.get().get_node(RAID_NAME), is_silent)
        except:
            self.__set_raid_cache_value(real_raiddev_name,
                                        STATUS_NAME, CRASH_STATUS)
            entry.set_result(Results.CREATE_ERROR)
        else:
            self.__set_raid_cache_value(raiddev_name,
                                        STATUS_NAME, RUNNING_STATUS)
            entry.set_result(Results.OK)

        return entry

    def __create_handler(self, entry):

        with util.RaidOperation(ReqTypes.CREATE):
            try:
                return self.__sub_create_handler(entry)
            except:
                self.exception("fail to __sub_create_handler()")
                raise

    def __sub_add_handler(self, entry):
        """Handle "add" request."""

        self.__validate_root_tag_name(entry)

        raid_node = entry.get().get_node(RAID_NAME)
        raid_path = self.__get_device_path(raid_node, RAID_ATTR)

        PART_NAME_XPATH = "./partition/name"
        part_name = raid_node.get_node(PART_NAME_XPATH).get_text()
        part_path = self.__devname2real_devpath(part_name)

        try:
            self.__clear_raid_superblock(part_path)
        except:
            self.exception("fail to __clear_raid_superblock " + part_path)
            pass

        try:
            self.__add_partition_to_raiddev(raid_path, part_path)
        except:
            entry.set_result(Results.ADD_ERROR)
        else:
            entry.set_result(Results.OK)

        return entry

    def __add_handler(self, entry):

        with util.RaidOperation(ReqTypes.ADD):
            try:
                return self.__sub_add_handler(entry)
            except:
                self.exception("fail to __sub_add_handler()")
                raise

    def __get_raiddev_vs_part_name_table(self, disk_name):

        table = {}
        symbl_disk_name = self.dev_mgr.get_dev_name(disk_name)
        if not symbl_disk_name:
            # Assume "disk_name" is device name.
            symbl_disk_name = disk_name

        symbl_disk_name = self.dev_mgr.get_base_dev_name(symbl_disk_name)

        for element_info in config.RAID_TO_ELEMENT.values():
            for raiddev_name, element_list in element_info.iteritems():
                for element in element_list:
                    if element.startswith(symbl_disk_name):
                        real_raiddev_name = \
                            self.__get_product_real_raiddev_name(raiddev_name)
                        table.update({real_raiddev_name: element})
        return table

    def __remove_all_raiddev_from_raidev(self, child_raid_name):

        # RAIDの構成要素に更に別のRAIDデバイスが存在し
        # そのRAIDデバイスをremoveする
        # 通常のHDDのremoveとは別処理とする

        for parent_raid, element_raid_list in \
            util.get_raid_element_info().iteritems():

            if child_raid_name not in element_raid_list:
                continue

            try:
                self.__remove_partition_from_raiddev(parent_raid,
                                                     child_raid_name)
            except:
                self.exception("__remove_partition_from_raiddev(%s,%s)" \
                    % (parent_raid, child_raid_name))
                is_fail = True
                continue

        if is_fail:
            raise self.__RemoveError
        return

    def __remove_from_all_raiddev(self, entry):

        # ディスク名を指定された場合はそのディスクが参加しているすべての
        # RAIDからそのディスクのパーティションを削除する
        # ディスクが強制的に取り外されてディスクのデバイスノードが見えなく
        # なった場合などに使用する

        is_fail = False
        DISK_NAME_XPATH = "./disk/name"
        disk_name = entry.get().get_node(DISK_NAME_XPATH).get_text()

        # RAIDに構成要素が別のRAIDデバイスの場合は別処理
        if disk_name.startswith("md"):
            self.__remove_all_raiddev_from_raidev(disk_name)
            return

        # 通常のHDDのremove処理
        raid_part_table = self.__get_raiddev_vs_part_name_table(disk_name)
        for raiddev_name, part_name in raid_part_table.iteritems():

            # dummyノードのパスを取得
            dummy_devnode_path = self.__get_dummy_dev_node_file_path(part_name)

            is_join = True
            if not self.__is_partition_join_to_raid(
                raiddev_name, os.path.basename(dummy_devnode_path)):
                is_join = False

            raid_path = os.path.join(DEV_DIR, raiddev_name)
            try:
                self.__remove_partition_from_raiddev(raid_path,
                                                     dummy_devnode_path)
            except:
                if is_join:
                    is_fail = True

        if is_fail:
            raise self.__RemoveError

        return

    def __check_crash(self, entry):
        """Check whether raid status becomes crash if
           removing specified disk.
           result tag value is 'crash' of 'nocrash'"""

        RESULT_XML_TMPL = \
            "<raid><name>%(name)s</name><result>%(result)s</result></raid>"

        remove_target = entry.get().get_value("./disk/name", None)
        if remove_target is None:
            remove_target = entry.get().get_value("./raid/name", None)
            if remove_target is None:
                raise InvalidArgError("removetarget is not specified !!")
        else:
            if entry.get().get_value("./raid/name", None) is not None:
                raise InvalidArgError("multiple remove targets is specified!!")

        entry.get().append_child(RESULT_NAME)
        entry.get().get_node(RESULT_NAME).append_child(RAIDLIST_NAME)
        raidlist_node = \
            entry.get().get_node(RESULT_NAME).get_node(RAIDLIST_NAME)

        for raiddev_name, element_list in \
                util.get_raid_element_info().iteritems():

            try:
                real_raiddev_name = \
                    self.__get_product_real_raiddev_name(raiddev_name)

                for element in element_list:

                    if not element.startswith(remove_target):
                        continue

                    raid_status = self.__check_raid_status_if_remove(
                                      real_raiddev_name, remove_target)

                    result_node = \
                        XmlNode(RESULT_XML_TMPL % {"name": real_raiddev_name,
                                                   "result": raid_status})
                    raidlist_node.append_child(result_node)
                    break
            except:
                continue

        entry.set_result(Results.OK)
        return entry

    def __check_operation(self, entry):

        op = util.get_cur_raid_operation()
        entry.get().append_child(OPERATION_NAME, op)
        entry.set_result(Results.OK)
        return entry

    def __sub_remove_handler(self, entry):
        """Handle "remove" request."""

        RAID_NAME_XPATH = "./raid/name[@type='raid']"
        PART_NAME_XPATH = "./partition/name"

        self.__validate_root_tag_name(entry)

        # ディスク名を指定された場合はそのディスクが参加しているすべてのRAIDから
        # そのディスクのパーティションを削除
        if self.__is_exist_node(entry.get(), "./disk"):
            try:
                self.__remove_from_all_raiddev(entry)
            except:
                entry.set_result(Results.DELETE_ERROR)
            else:
                entry.set_result(Results.OK)
            return entry

        raid_node = entry.get().get_node(RAID_NAME)

        raid_name = entry.get().get_value(RAID_NAME_XPATH)
        raid_path = self.__get_device_path(raid_node, RAID_ATTR)

        part_name = raid_node.get_value(PART_NAME_XPATH)
        part_path = self.__devname2real_devpath(part_name)

        element_list = self.__get_element_name_list(raid_name)
        if part_name not in element_list:
            entry.set_result(Results.OK)
            return entry

        try:
            self.__remove_partition_from_raiddev(raid_path, part_path)
        except:
            entry.set_result(Results.DELETE_ERROR)
        else:
            entry.set_result(Results.OK)

        return entry

    def __remove_handler(self, entry):

        with util.RaidOperation(ReqTypes.REMOVE):
            try:
                return self.__sub_remove_handler(entry)
            except:
                self.exception("fail to __sub_remove_handler()")
                raise

    def __activerepair_handler(self, entry):
        """Handle "activerepair" request."""

        self.__validate_root_tag_name(entry)

        # スケジュール設定の場合
        op = entry.get_options()[OPERATION_OPT]
        if SCHEDULE_OPT_VALUE == op:
            self.__set_activerepair_schedule(entry)
            entry.set_result(Results.OK)
            return entry

        # アクティブリペア実行の場合
        return self.send_entry(entry,
                               ALBG_ACTIVEREPAIR_PATH,
                               preserve_file=True)

    def __control_handler(self, entry):
        """Handle "control" request."""

        self.__validate_root_tag_name(entry)

        ctrl_funcs = {START_SERVICE_OP: self.__mdadm_start_daemon,
                      STOP_SERVICE_OP: self.__mdadm_stop_daemon}

        op = entry.get_options()[OPERATION_OPT]

        try:
            ctrl_funcs[op]()
            entry.set_result(Results.OK)
        except self.__StartError:
            entry.set_result(Results.START_ERROR)
        except self.__StopError:
            entry.set_result(Results.STOP_ERROR)

        return entry

    def __chk_exist_handler(self, entry):

        is_exist = False
        is_rebuild_target = False

        # ディスク指定の場合
        disk_name = entry.get().get_value("./disk/name", "")
        if disk_name:

            try:
                real_disk_name = self.__disk_name2raw_disk_name(disk_name)
            except NotFoundError:
                real_disk_name = disk_name

            for raiddev_name in config.RAID_DEV_LIST:

                real_raiddev_name = \
                    self.__get_product_real_raiddev_name(raiddev_name)

                try:
                    slave_dir = \
                        os.path.join(RAID_X_SLAVES_DIR % real_raiddev_name)
                    slave_list = os.listdir(slave_dir)
                except:
                    continue

                for slave in slave_list:
                    if slave.startswith(real_disk_name):
                        is_exist = True
                        if self.__is_spare_element(real_raiddev_name,
                                                   real_disk_name):
                            if not real_raiddev_name in \
                                   config.SYSTEM_RAID_DEV_LIST:
                                is_rebuild_target = True
            if is_exist:
                disk_node = entry.get().get_node(DISK_NAME)
                disk_node.set_bool_attr(REBUILD_TARGET_ATTR,
                                        is_rebuild_target)
                entry.set_result(Results.OK)
                return entry
            else:
                entry.set_result(Results.DEVICE_NOT_FOUND)
                return entry

        # パーティション指定の場合
        part_name = entry.get().get_value("./partition/name", "")
        if "" == part_name:
            raise InvalidArgError(str(entry))

        try:
            real_part_name = self.__part_name2raw_part_name(part_name)
        except NotFoundError:
            entry.set_result(Results.DEVICE_NOT_FOUND)
            return entry

        raiddev_name = entry.get().get_value("./raid/name", "")
        real_raiddev_name = self.__get_product_real_raiddev_name(raiddev_name)

        if not os.path.exists(os.path.join(DEV_DIR, real_raiddev_name)):
            entry.set_result(Results.DEVICE_NOT_FOUND)
            return entry

        slave_list = \
            os.listdir(os.path.join(RAID_X_SLAVES_DIR % real_raiddev_name))

        if real_part_name in slave_list:
            entry.set_result(Results.OK)
        else:
            entry.set_result(Results.DEVICE_NOT_FOUND)

        return entry

    def __update_handler(self, entry):

        self.__validate_root_tag_name(entry)

        raiddev_name = entry.get().get_value("./raid/name", "")
        real_raiddev_name = self.__get_product_real_raiddev_name(raiddev_name)
        status = entry.get().get_value("./raid/status", "")
        self.__set_raid_cache_value(real_raiddev_name, STATUS_NAME, status)
        entry.set_result(Results.OK)
        return entry

    def __check_handler(self, entry):

        chk_funcs = {CRASH_ATTR_VAL: self.__check_crash,
                     GET_OPERATION_ATTR_VAL: self.__check_operation}

        chk_type = entry.get_options()[TYPE_OPT]

        try:
            return chk_funcs[chk_type](entry.get_clone())
        except:
            entry.set_result(Results.CHECK_ERROR)

        return entry

    def __delete_handler(self, entry):

        RAID_NAME_XPATH = "./raid/name[@type='raid']"
        raid_name = entry.get().get_value(RAID_NAME_XPATH)
        
        real_raiddev_name = self.__convert_real_devname(raid_name)
        element_name_list = self.__get_element_name_list(real_raiddev_name)

        is_fail = False        
        for element_name in element_name_list:
            if element_name.startswith("satad"):
                #try:
                #    self.__clear_raid_superblock(
                #        self.__devname2real_devpath(element_name))
                #except:
                #    self.exception(
                #        "fail to __clear_raid_superblock(%s)" % element_name)
                #    is_fail = True

                disk_name = self.__partition2disk(element_name)
                
                self.__zero_clear_device(self.__devname2real_devpath(disk_name),
                                         0, 512)


        if is_fail:
            entry.set_result(Results.DELETE_ERROR)
        else:
            entry.set_result(Results.OK)
        return

    #
    # Constractor
    #

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.default_env_path = ""
        self.dev_mgr = device.DevMgr(config.DEV_DIR, config.DEV_LINK_DIR,
                                config.DEV_FILE_DIR, config.HDD_DEV_LIST)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.GET_ELEMENT_INFO,
                         self.__get_element_info_handler)
        self.set_handler(ReqTypes.START, self.__start_handler)
        self.set_handler(ReqTypes.CREATE, self.__create_handler)
        self.set_handler(ReqTypes.STOP, self.__stop_handler)
        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.REMOVE, self.__remove_handler)
        self.set_handler(ReqTypes.ACTIVEREPAIR, self.__activerepair_handler)
        self.set_handler(ReqTypes.CONTROL, self.__control_handler)
        self.set_handler(ReqTypes.CHK_EXIST, self.__chk_exist_handler)
        self.set_handler(ReqTypes.UPDATE, self.__update_handler)
        self.set_handler(ReqTypes.CHECK, self.__check_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)

#
# Main
#

if __name__ == "__main__":
    PalRaidCtrl(option=palcmd.PalCmdOption()).handle_request()
