#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for Selfcheck."""
from __future__ import with_statement

import os
import shutil
import errno

import libpal.palcmd as palcmd
import pal_config as config
from iobase.xmldata import XmlConfNode
from pal_info import (ReqTypes, Results)
from albg_info import ALBG_SELFCHECK_PATH

NAME_NAME = "name"
STATUS_NAME = "status"
RESULT_NAME = "result"
TIMESTAMP_NAME = "timestamp"
CHECK_NODE_LIST_PATH = "checklist/check"

ROOT_TAG_FMT_STR = "<selfcheck>%s</selfcheck>"
CHECK_NODE_FMT_STR = "<check name='%s'><status/><result/><timestamp/></check>"

STATUS_SEP = ":"

DEFAULT_SET_STATUS = "COMPLETE"


class PalSelfcheck(palcmd.PalCmd):
    """Pal Selfcheck command implementation class."""

    def __get_checklist_node(self):

        checklist_node = XmlConfNode("checklist")

        for check in os.listdir(config.SELFCHECK_STATUS_DIR):
            checklist_node.append_child(
                XmlConfNode(CHECK_NODE_FMT_STR % check))

        return checklist_node

    def __set_status_entry(self, entry):

        for check_node in entry.get().get_node_list(CHECK_NODE_LIST_PATH):
            status_file_path = os.path.join(config.SELFCHECK_STATUS_DIR,
                check_node.get_attr_value(NAME_NAME))
            try:
                with open(status_file_path, "rb") as status_file:
                    text = status_file.read()
                    status = text.split(STATUS_SEP, 1)[0]
                    result = text.split(STATUS_SEP, 1)[1]

                check_node.get_node(STATUS_NAME).set_text(status)
                if status == "COMPLETE":
                    check_node.get_node(RESULT_NAME).set_text(result)
                    check_node.get_node(TIMESTAMP_NAME).set_text(
                        str(int(os.stat(status_file_path).st_mtime)))
            except (IOError, OSError), err:
                if err.errno != errno.ENOENT:
                    raise
                check_node.get_node(STATUS_NAME).set_text("NOT_EXEC")

    def __set_status_file(entry, check_node, dir_path):

        check_name = check_node.get_attr_value(NAME_NAME)
        check_status = check_node.get_value(STATUS_NAME, "")
        check_result = check_node.get_value(RESULT_NAME, "")

        if not check_status:
            check_status = DEFAULT_SET_STATUS

        file_path = os.path.join(dir_path, check_name)
        tmp_file_path = os.path.join(
            config.TEMP_DIR, check_name + str(os.getpid()))

        with open(tmp_file_path, "wb") as file:
            file.write(":".join([check_status, check_result]))

        shutil.move(tmp_file_path, file_path)

    def __set_status_files(self, entry):

        dir_path = os.path.join(os.path.join(config.CONF_DIR,
                "selfcheck", "status"))

        try:
            os.makedirs(dir_path)
        except:
            pass

        for check_node in entry.get().get_node_list(CHECK_NODE_LIST_PATH):
             self.__set_status_file(check_node, dir_path)

    def __selfcheck_handler(self, entry):
        """Selfcheck Handler."""

        result_entry = self.send_entry(entry, ALBG_SELFCHECK_PATH)
        return result_entry

    def __query_handler(self, entry):
        """Query Handler."""

        entry.set(
            XmlConfNode(ROOT_TAG_FMT_STR % self.__get_checklist_node()))

        self.__set_status_entry(entry)

        entry.set_result(Results.OK)

        return entry

    def __set_handler(self, entry):
        """Set Handler."""

        self.__set_status_files(entry)

        entry.set_result(Results.OK)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.SELFCHECK, self.__selfcheck_handler)
        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.SET, self.__set_handler)

#
# Main
#

if __name__ == "__main__":
    PalSelfcheck(option=palcmd.PalCmdOption()).handle_request()
