#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for managemnet the system state."""


from __future__ import with_statement

import os

import iobase.xmldata as xmldata
import iobase.fileut as fileut
from iobase.exception import (InvalidArgError,)

import libpal.palcmd as palcmd
import pal_config as config
from pal_info import (ReqTypes, Results)


STATEMGR_NAME = "statemgr"

STATE_NAME = "state"
STATELIST_NAME = "statelist"
STATE_PATH = ".//" + STATE_NAME
CATEGORY_ATTR_NAME = "category"

SYSTEM_CATEGORY = "system"
RAID_CATEGORY = "raid"
RAIDVOL_CATEGORY = "raidvol"
FILESYSTEM_CATEGORY = "filesystem"
ACTIVEREPAIR_CATEGORY = "activerepair"
BACKUP_CATEGORY = "backup"
QUICKCOPY_CATEGORY = "quickcopy"

BOOTING_STATE = "booting"
RUNNING_STATE = "running"
BUSY_STATE = "busy"
UPDATING_STATE = "updating"
PREPARE_INIT_STATE = "prepareinit"
SHUTDOWN_STATE = "shutdown"
PREPARING_STATE = "preparing"
STOP_STATE = "stop"
DEGRADE_STATE = "degrade"
FAIL_STATE = "fail"
CRASH_STATE = "crash"
REBUILDING_STATE = "rebuilding"
MOUNTFAIL_STATE = "mountfail"
KEYWAITING_STATE = "keywaiting"
NOOPERATION_STATE = "nooperation"
FORMATTING_STATE = "formatting"
CHECKING_STATE = "checking"
NONCHANGEABLE_STATE = "nonchangeable"

CATEGORY_LIST = (SYSTEM_CATEGORY, RAID_CATEGORY, RAIDVOL_CATEGORY,
                    FILESYSTEM_CATEGORY, ACTIVEREPAIR_CATEGORY,
                    BACKUP_CATEGORY, QUICKCOPY_CATEGORY)

DEFAULT_STATE_DICT = {
        SYSTEM_CATEGORY: BOOTING_STATE,
        RAID_CATEGORY: PREPARING_STATE,
        RAIDVOL_CATEGORY: PREPARING_STATE,
        FILESYSTEM_CATEGORY: NOOPERATION_STATE,
        ACTIVEREPAIR_CATEGORY: NOOPERATION_STATE,
        BACKUP_CATEGORY: NOOPERATION_STATE,
        QUICKCOPY_CATEGORY: NOOPERATION_STATE,
    }

STATE_LIST_MAP = {
        SYSTEM_CATEGORY: [BOOTING_STATE, RUNNING_STATE, BUSY_STATE,
                        UPDATING_STATE, PREPARE_INIT_STATE, SHUTDOWN_STATE],
        RAID_CATEGORY: [PREPARING_STATE, RUNNING_STATE, STOP_STATE,
                        DEGRADE_STATE, FAIL_STATE, CRASH_STATE],
        RAIDVOL_CATEGORY: [PREPARING_STATE, RUNNING_STATE, REBUILDING_STATE,
                        FAIL_STATE, MOUNTFAIL_STATE, KEYWAITING_STATE],
        FILESYSTEM_CATEGORY: [NOOPERATION_STATE, FORMATTING_STATE,
                                CHECKING_STATE],
        ACTIVEREPAIR_CATEGORY: [NOOPERATION_STATE, RUNNING_STATE],
        BACKUP_CATEGORY: [NOOPERATION_STATE, RUNNING_STATE],
        QUICKCOPY_CATEGORY: [NOOPERATION_STATE, RUNNING_STATE],
    }

STATE_CHANGEABLE_CONDITION_MAP = {
        SYSTEM_CATEGORY: {
                BOOTING_STATE: {
                        # Always fail to change "booting"
                        SYSTEM_CATEGORY: [NONCHANGEABLE_STATE],
                    },
                RUNNING_STATE: {
                        SYSTEM_CATEGORY: [BOOTING_STATE, RUNNING_STATE,
                                        BUSY_STATE, UPDATING_STATE,
                                        PREPARE_INIT_STATE],
                    },
                BUSY_STATE: {
                        SYSTEM_CATEGORY: [BOOTING_STATE, RUNNING_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        BACKUP_CATEGORY: [NOOPERATION_STATE],
                        QUICKCOPY_CATEGORY: [NOOPERATION_STATE],
                    },
                UPDATING_STATE: {
                        SYSTEM_CATEGORY: [BOOTING_STATE, RUNNING_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        BACKUP_CATEGORY: [NOOPERATION_STATE],
                        QUICKCOPY_CATEGORY: [NOOPERATION_STATE],
                    },
                PREPARE_INIT_STATE: {
                        SYSTEM_CATEGORY: [BOOTING_STATE, RUNNING_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        ACTIVEREPAIR_CATEGORY: [NOOPERATION_STATE],
                        BACKUP_CATEGORY: [NOOPERATION_STATE],
                        QUICKCOPY_CATEGORY: [NOOPERATION_STATE],
                    },
                SHUTDOWN_STATE: {
                        RAIDVOL_CATEGORY: [PREPARING_STATE, RUNNING_STATE,
                                        FAIL_STATE, MOUNTFAIL_STATE,
                                        KEYWAITING_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        ACTIVEREPAIR_CATEGORY: [NOOPERATION_STATE],
                        BACKUP_CATEGORY: [NOOPERATION_STATE],
                        QUICKCOPY_CATEGORY: [NOOPERATION_STATE],
                    },
            },
        RAID_CATEGORY: {
                PREPARING_STATE: {},
                RUNNING_STATE: {},
                STOP_STATE: {},
                DEGRADE_STATE: {},
                FAIL_STATE: {},
                CRASH_STATE: {},
            },
        RAIDVOL_CATEGORY: {
                PREPARING_STATE: {},
                RUNNING_STATE: {},
                REBUILDING_STATE: {},
                FAIL_STATE: {},
                MOUNTFAIL_STATE: {},
                KEYWAITING_STATE: {
                        RAID_CATEGORY: [RUNNING_STATE, STOP_STATE,
                                        DEGRADE_STATE, FAIL_STATE,
                                        CRASH_STATE],
                    },
            },
        FILESYSTEM_CATEGORY: {
                NOOPERATION_STATE: {},
                FORMATTING_STATE: {
                        SYSTEM_CATEGORY: [RUNNING_STATE, BUSY_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        BACKUP_CATEGORY: [NOOPERATION_STATE],
                        QUICKCOPY_CATEGORY: [NOOPERATION_STATE],
                    },
                CHECKING_STATE: {
                        SYSTEM_CATEGORY: [RUNNING_STATE, BUSY_STATE],
                        RAID_CATEGORY: [RUNNING_STATE, DEGRADE_STATE,
                                        FAIL_STATE],
                        RAIDVOL_CATEGORY: [RUNNING_STATE, REBUILDING_STATE,
                                        FAIL_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        BACKUP_CATEGORY: [NOOPERATION_STATE],
                        QUICKCOPY_CATEGORY: [NOOPERATION_STATE],
                    },
            },
        ACTIVEREPAIR_CATEGORY: {
                NOOPERATION_STATE: {},
                RUNNING_STATE: {
                        SYSTEM_CATEGORY: [RUNNING_STATE, BUSY_STATE],
                        RAIDVOL_CATEGORY: [RUNNING_STATE, MOUNTFAIL_STATE,
                                            KEYWAITING_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        ACTIVEREPAIR_CATEGORY: [NOOPERATION_STATE],
                    },
            },
        BACKUP_CATEGORY: {
                NOOPERATION_STATE: {},
                RUNNING_STATE: {
                        SYSTEM_CATEGORY: [RUNNING_STATE, BUSY_STATE],
                        RAIDVOL_CATEGORY: [RUNNING_STATE, REBUILDING_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        BACKUP_CATEGORY: [NOOPERATION_STATE],
                        QUICKCOPY_CATEGORY: [NOOPERATION_STATE],
                    },
            },
        QUICKCOPY_CATEGORY: {
                NOOPERATION_STATE: {},
                RUNNING_STATE: {
                        SYSTEM_CATEGORY: [RUNNING_STATE, BUSY_STATE],
                        RAIDVOL_CATEGORY: [RUNNING_STATE, REBUILDING_STATE],
                        FILESYSTEM_CATEGORY: [NOOPERATION_STATE],
                        BACKUP_CATEGORY: [NOOPERATION_STATE],
                        QUICKCOPY_CATEGORY: [NOOPERATION_STATE],
                    },
            },
    }


class PalStateMgr(palcmd.PalCmd):
    """Pal State manager command implementation class."""

    def __check_body_node(self, node):
        """Check if entry's body node is valid."""

        if node.get_node_name() != STATEMGR_NAME:
            raise InvalidArgError(node.get_node_name())

    def __get_all_statelist(self):
        """Return statelist node for all category."""

        statelist = xmldata.XmlConfNode(STATELIST_NAME)
        for category in CATEGORY_LIST:
            statelist.append_child(STATE_NAME,
                                attrs={CATEGORY_ATTR_NAME: category})
        return statelist

    def __get_category_state(self, category):
        """Return the state of "category" category."""

        state_path = os.path.join(config.STATUS_TEMP_DIR, category)
        try:
            with open(state_path) as state_file:
                return state_file.read().strip()
        except:
            return DEFAULT_STATE_DICT[category]

    def __set_category_state(self, category, new_state):
        """Set the state of "category" category to "new_state"."""

        state_path = os.path.join(config.STATUS_TEMP_DIR, category)
        with open(state_path, "w") as state_file:
            return state_file.write(new_state)

    def __is_possible_change_state(self, category, new_state):
        """Return if it is possible to change state to "new_state"."""

        state_changeable_condition = \
                STATE_CHANGEABLE_CONDITION_MAP[category][new_state]
        for check_category in state_changeable_condition:
            if (not self.__get_category_state(check_category) in
                    state_changeable_condition[check_category]):
                return False

        return True

    def __query_handler(self, entry):
        """Handler for "query" request."""

        self.__check_body_node(entry.get())

        state_list = entry.get().get_node_list(STATE_PATH)
        if not len(state_list):
            entry.get().clear()
            entry.get().append_child(self.__get_all_statelist())
            state_list = entry.get().get_node_list(STATE_PATH)

        with fileut.FLock(config.LOCK_DIR, STATEMGR_NAME, exclusive=False):
            for state in state_list:
                state_str = self.__get_category_state(
                                state.get_attr_value(CATEGORY_ATTR_NAME))
                state.set_text(state_str)

        entry.set_result(Results.OK)
        return entry

    def __set_handler(self, entry):
        """Handler for "set" request."""

        self.__check_body_node(entry.get())

        state = entry.get().get_node(STATE_NAME)
        category = state.get_attr_value(CATEGORY_ATTR_NAME)
        new_state = state.get_text()
        if not category in STATE_LIST_MAP:
            raise InvalidArgError('Invalid category "' + category + '"')

        if not new_state in STATE_LIST_MAP[category]:
            raise InvalidArgError('Invalid state "' + new_state + \
                                    '" for ' + category)

        with fileut.FLock(config.LOCK_DIR, STATEMGR_NAME, exclusive=True):
            try:
                self.__set_category_state(category, new_state)
                entry.set_result(Results.OK)
            except:
                self.exception("Set error")
                entry.set_result(Results.SET_ERROR)

        return entry

    def __change_handler(self, entry):
        """Handler for "change" request."""

        self.__check_body_node(entry.get())

        state = entry.get().get_node(STATE_NAME)
        category = state.get_attr_value(CATEGORY_ATTR_NAME)
        new_state = state.get_text()
        if not category in STATE_LIST_MAP:
            raise InvalidArgError('Invalid category "' + category + '"')

        if not new_state in STATE_LIST_MAP[category]:
            raise InvalidArgError('Invalid state "' + new_state + \
                                    '" for ' + category)

        with fileut.FLock(config.LOCK_DIR, STATEMGR_NAME, exclusive=True):
            if self.__is_possible_change_state(category, new_state):
                try:
                    self.__set_category_state(category, new_state)
                    entry.set_result(Results.OK)
                except:
                    entry.set_result(Results.SET_ERROR)
            else:
                    entry.set_result(Results.CHANGE_ERROR)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.QUERY, self.__query_handler)
        self.set_handler(ReqTypes.SET, self.__set_handler)
        self.set_handler(ReqTypes.CHANGE, self.__change_handler)


#
# Main
#

if __name__ == "__main__":
    PalStateMgr(option=palcmd.PalCmdOption()).handle_request()
