#!/usr/bin/env python
# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""PAL command for USER configuration."""

import os
import crypt
import pwd

from iobase.cmdexec import CmdExec
from iobase.cmdexec import NULL, PIPE
from iobase.exception import (ExecError, InvalidArgError)

import libpal.palcmd as palcmd
from libpal.util import RemountRoot
from pal_info import (ReqTypes, Results)

import pal_config as config


USERS_NAME = "users"
USER_NAME = "user"
NAME_NAME = "name"
FULLNAME_NAME = "fullname"
UID_NAME = "uid"
GID_NAME = "gid"
PASSWD_NAME = "passwd"
SHELL_NAME = "shell"
HOMEDIR_NAME = "homedir"

CHAR_LIST = "0123456789" + \
            "abcdefghijklmnopqrstuvwxyz" + \
            "ABCDEFGHIJKLMNOPQRSTUVWXYZ./"
URAND_NUM = 8
SALT_HEADER = "$1$"
SALT_FOOTER = "$"

DEFAULT_SHELL = "/bin/false"
DEFAULT_HOMEDIR = "/nonexistent"

INVALID_ROOT_TAG_MSG = "Invalid tag name!!"
USERNAME_EMPTY_MSG = "User name empty!!"
ADD_FAILED_MSG = "Fail adding user!!"
MODIFY_FAILED_MSG = "Fail modifying user!!"
DELETE_FAILED_MSG = "Fail deleting user!!"


class PalUsers(palcmd.PalCmd):
    """PAL Users command implementation class."""

    def __check_root_tag(self, entry):
        """Check root tag."""

        root_name = entry.get().get_node_name()

        if root_name != USERS_NAME:
            raise InvalidArgError(INVALID_ROOT_TAG_MSG)

    def __empty(self, val):
        if val in (None, "", [], (), {}):
            return True
        return False

    def __crypt(self, string):
        """Encrypt password."""

        salt_string = SALT_HEADER

        for rand in os.urandom(URAND_NUM):
            salt_string += CHAR_LIST[ord(rand) % len(CHAR_LIST)]

        salt_string = salt_string + SALT_FOOTER

        return crypt.crypt(string, salt_string)

    def __userdel(self, name):
        cmd = [config.USERDEL_CMD, name]
        with RemountRoot():
            CmdExec(cmd).execute(stdout=NULL, stderr=NULL)

    def __add_handler(self, entry):
        """Handle add request."""

        try:
            self.__check_root_tag(entry)

            user_node = entry.get().get_node(USER_NAME)

            name = user_node.get_value(NAME_NAME, None)
            if self.__empty(name):
                raise InvalidArgError(USERNAME_EMPTY_MSG)
            uid = user_node.get_value(UID_NAME, None)
            gid = user_node.get_value(GID_NAME, str(config.DEFAULT_GID))
            passwd = user_node.get_value(PASSWD_NAME, "", do_strip=False)
            comment = user_node.get_value(FULLNAME_NAME, "")
            shell = user_node.get_value(SHELL_NAME, DEFAULT_SHELL)
            homedir = user_node.get_value(HOMEDIR_NAME, DEFAULT_HOMEDIR)

            cmd = [config.USERADD_CMD]
            cmd = cmd + ['-p', self.__crypt(passwd)]
            cmd = cmd + ['-c', comment]
            cmd = cmd + ['-s', shell]
            cmd = cmd + ['-d', homedir]
            if not self.__empty(uid):
                cmd = cmd + ['-u', uid]
            if not self.__empty(gid):
                cmd = cmd + ['-g', gid]
            cmd = cmd + [name]
            with RemountRoot():
                CmdExec(cmd).execute(stdout=NULL, stderr=NULL)

            info = pwd.getpwnam(name)
            uid = info.pw_uid
            gid = info.pw_gid
            comment = info.pw_gecos
            shell = info.pw_shell
            homedir = info.pw_dir

            user_node.clear()
            user_node.append_child(NAME_NAME, name)
            user_node.append_child(UID_NAME, text=str(uid))
            user_node.append_child(GID_NAME, text=str(gid))
            user_node.append_child(PASSWD_NAME, text=passwd)
            user_node.append_child(FULLNAME_NAME, text=comment)
            user_node.append_child(SHELL_NAME, text=shell)
            user_node.append_child(HOMEDIR_NAME, text=homedir)

            entry.set_result(Results.OK)

        except ExecError:
            try:
                self.__userdel(name)
            except:
                pass
            entry.set_result(Results.ADD_ERROR)
            self.exception(ADD_FAILED_MSG)

        return entry

    def __modify_handler(self, entry):
        """Handle modify request."""

        try:
            self.__check_root_tag(entry)

            user_node = entry.get().get_node(USER_NAME)

            name = user_node.get_value(NAME_NAME, None)
            if self.__empty(name):
                raise InvalidArgError(USERNAME_EMPTY_MSG)
            uid = user_node.get_value(UID_NAME, None)
            gid = user_node.get_value(GID_NAME, None)
            passwd = user_node.get_value(PASSWD_NAME, None, do_strip=False)
            comment = user_node.get_value(FULLNAME_NAME, None)
            shell = user_node.get_value(SHELL_NAME, None)
            homedir = user_node.get_value(HOMEDIR_NAME, None)

            cmd = []
            if not self.__empty(uid):
                cmd = cmd + ['-u', uid]
            if not self.__empty(gid):
                cmd = cmd + ['-g', gid]
            if passwd is not None:
                cmd = cmd + ['-p', self.__crypt(passwd)]
            if not self.__empty(comment):
                cmd = cmd + ['-c', comment]
            if not self.__empty(shell):
                cmd = cmd + ['-s', shell]
            if not self.__empty(homedir):
                cmd = cmd + ['-d', homedir]
            # 変更項目が無ければusermodしない
            if not self.__empty(cmd):
                cmd = [config.USERMOD_CMD] + cmd + [name]
                with RemountRoot():
                    CmdExec(cmd).execute(stdout=NULL, stderr=NULL)

            info = pwd.getpwnam(name)
            uid = info.pw_uid
            gid = info.pw_gid
            comment = info.pw_gecos
            shell = info.pw_shell
            homedir = info.pw_dir

            user_node.clear()
            user_node.append_child(NAME_NAME, name)
            user_node.append_child(UID_NAME, text=str(uid))
            user_node.append_child(GID_NAME, text=str(gid))
            user_node.append_child(PASSWD_NAME, text=passwd)
            user_node.append_child(FULLNAME_NAME, text=comment)
            user_node.append_child(SHELL_NAME, text=shell)
            user_node.append_child(HOMEDIR_NAME, text=homedir)

            entry.set_result(Results.OK)

        except ExecError:
            entry.set_result(Results.MODIFY_ERROR)
            self.exception(MODIFY_FAILED_MSG)

        return entry

    def __delete_handler(self, entry):
        """Handle delete request."""

        try:
            self.__check_root_tag(entry)

            user_node = entry.get().get_node(USER_NAME)
            name = user_node.get_value(NAME_NAME, None)
            if self.__empty(name):
                raise InvalidArgError(USERNAME_EMPTY_MSG)

            self.__userdel(name)
#            cmd = [config.USERDEL_CMD, name]
#            with RemountRoot():
#                CmdExec(cmd).execute(stdout=NULL, stderr=NULL)

            entry.set_result(Results.OK)

        except ExecError:
            entry.set_result(Results.DELETE_ERROR)
            self.exception(DELETE_FAILED_MSG)

        return entry

    def __init__(self, option=None):
        """Constructor."""

        palcmd.PalCmd.__init__(self, option=option)

        self.set_handler(ReqTypes.ADD, self.__add_handler)
        self.set_handler(ReqTypes.MODIFY, self.__modify_handler)
        self.set_handler(ReqTypes.DELETE, self.__delete_handler)

#
# Main
#

if __name__ == "__main__":
    PalUsers(option=palcmd.PalCmdOption()).handle_request()
