# -*- mode: landisk-python; coding: utf-8; -*-
# vim:ts=4 sw=4 sts=4 ai si et sta

"""IOPortal utility module.

Define functions for communicating IOPortal.
"""

try:
    import json
except:
    import simplejson as json

import time
import ssl
import base64
import urllib2
import socket


SSL_CREATE_UNVERIFIED_CONTEXT = "_create_unverified_context"
MAC_SEP = ":"

CONTENT_TYPE_HEADER_DICT = {"Content-Type": "application/json"}

AUTH_URL_PATH = "/rl3/pkg/auth"
GET_KEY_URL_PATH = "/rl3/pkg/get_key"
REGIST_URL_PATH = "/rl3/service/regist"
DELETE_URL_PATH = "/rl3/service/remove"

AUTH_KEY = "auth"
USERID_KEY = "userId"
PASSWD_KEY = "passwd"
PRODUCT_KEY = "product"
SERIAL_NUMBER_KEY = "serialNumber"
MAC_ADDRESS_KEY = "macAddress"
RESULT_KEY = "result"
ERROR_KEY = "error"
CODE_KEY = "code"
MESSAGE_KEY = "message"
RESPONSE_KEY = "response"
SESSIONID_KEY = "sessionId"
ENCODEDKEY_KEY = "encodedKey"
SERVICE_KEY = "service"
TYPE_KEY = "type"
PIN_KEY = "PIN"

RESULT_OK = "OK"
RESULT_WARN = "WARN"
RESULT_ERROR = "ERROR"

TIMEOUT = 30

AUTH_ERROR = 403
INVALID_MAC_SERIAL_ERROR = 410
REGISTRATION_FAIL_ERROR = 502
ALREADY_DISABLED_ERROR = 204
ALREADY_REGISTERED_ERROR = 280


class RequestError(Exception):
    """Class for error regarding to IOPortal Request."""

    def __init__(self, request_url, result, code, message):
        """Constructor."""

        self.request_url = request_url
        self.result = result
        self.code = code
        self.message = message

    def __str__(self):
        """Return the exception information as string."""

        return "IOPortal Request Error: URL: %s  result: %s  code: %s" \
                    "  message: %s" % \
                (str(self.request_url), str(self.result), str(self.code),
                    str(self.message),)

    def get_request_url(self):
        """Return the "request_url"."""

        return self.request_url

    def get_result(self):
        """Return the "result"."""

        return self.result

    def get_code(self):
        """Return the "code"."""

        return self.code

    def get_message(self):
        """Return the "message"."""

        return self.message


class InternalError(Exception):
    """Class for internal error."""

    pass


class IOPortal:
    """Class for communicating IOPortal."""

    def _request(self, url, param, headers={}):
        """Send "data" to "url"."""

        if isinstance(param, dict):
            param = json.dumps(param)

        req_headers = {}
        req_headers.update(self.default_headers)
        req_headers.update(CONTENT_TYPE_HEADER_DICT)
        req_headers.update(headers)

        old_timeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(self.timeout)

        request = urllib2.Request(url, param, req_headers)
        response = self.opener.open(request)

        socket.setdefaulttimeout(old_timeout)

        response_dict = json.loads(response.read())

        if response_dict[RESULT_KEY] != RESULT_OK:
            raise RequestError(url, response_dict[RESULT_KEY],
                        response_dict[ERROR_KEY][CODE_KEY],
                        response_dict[ERROR_KEY][MESSAGE_KEY])

        return response_dict

    def _communicate(self, func, *args, **key_args):
        """Communicate to IOPortal with "func" function.
           If exception is raised, then it is retried.
        """

        retry = 0
        while True:
            try:
                if retry > 0:
                    time.sleep(self.retry_period)
                return func(*args, **key_args)
            except RequestError, err:
                # Do not retry
                raise
            except:
                if retry >= self.retry_count:
                    raise
                retry += 1

    def _get_session_id(self, mac_address, serial, headers):
        """Get a sessoion ID and return it."""

        mac_address = mac_address.replace(MAC_SEP, "").upper()

        param = {
                AUTH_KEY: {
                        USERID_KEY: self.mail_address,
                        PASSWD_KEY: base64.urlsafe_b64encode(self.password),
                    },
                PRODUCT_KEY: {
                        SERIAL_NUMBER_KEY: serial,
                        MAC_ADDRESS_KEY: mac_address,
                    },
            }

        response_dict = self._request(self.base_url + AUTH_URL_PATH,
                                            param, headers)
        return response_dict[RESPONSE_KEY][SESSIONID_KEY]

    def _get_encoded_key(self, mac_address, serial, headers):
        """Get a encoded key for "mac_address" and return it."""

        session_id = self._get_session_id(mac_address, serial, headers)

        param = {
                SESSIONID_KEY: session_id,
            }

        response_dict = self._request(self.base_url + GET_KEY_URL_PATH,
                                            param, headers)
        return response_dict[RESPONSE_KEY][ENCODEDKEY_KEY]

    def _regist_pin_code(self, mac_address, serial, pin_code, headers):
        """Regist PIN code to IOPortal server."""

        session_id = self._get_session_id(mac_address, serial, headers)

        param = {
            SESSIONID_KEY: session_id,
            SERVICE_KEY: {
                TYPE_KEY: "RL3",
                PIN_KEY: pin_code,
                }
            }

        response_dict = self._request(
            self.base_url + REGIST_URL_PATH, param, headers)
        return response_dict[RESPONSE_KEY][SESSIONID_KEY]

    def _delete_pin_code(self, mac_address, serial, pin_code, headers):
        """Delete PIN code from IOPortal server."""

        session_id = self._get_session_id(mac_address, serial, headers)

        param = {
            SESSIONID_KEY: session_id,
            SERVICE_KEY: {
                TYPE_KEY: "RL3",
                PIN_KEY: pin_code,
                }
            }

        response_dict = self._request(
            self.base_url + DELETE_URL_PATH, param, headers)
        return response_dict[RESPONSE_KEY][SESSIONID_KEY]

    def __init__(self, base_url, mail_address, password, proxies=None,
            default_headers={}, timeout=TIMEOUT, retry_period=5, retry_count=3):
        """Constructor."""

        self.base_url = base_url
        self.mail_address = mail_address
        self.password = password
        self.opener = None
        self.set_proxies(proxies)
        self.default_headers = default_headers
        self.timeout = timeout
        self.retry_period = retry_period
        self.retry_count = retry_count

    def set_proxies(self, proxies):
        """Build a opener to access to the server with "proxies"."""

        handlers = []
        if hasattr(ssl, SSL_CREATE_UNVERIFIED_CONTEXT):
            handlers += [urllib2.HTTPSHandler(
                                    context=ssl._create_unverified_context())]
        if proxies != None:
            handlers += [urllib2.ProxyHandler(proxies)]

        self.opener = urllib2.build_opener(*handlers)
        return proxies

    def get_encoded_key(self, mac_address, serial, headers={}):
        """Get a encoded key for "mac_address" and return it."""

        return self._communicate(self._get_encoded_key,
                                    mac_address, serial, headers)

    def regist_pin_code(self, mac_address, serial, pin_code, headers={}):
        """Regist PIN code to IOPortal server."""

        return self._communicate(self._regist_pin_code,
                                mac_address, serial, pin_code, headers)

    def delete_pin_code(self, mac_address, serial, pin_code, headers={}):
        """Delete PIN code from IOPortal server."""

        return self._communicate(self._delete_pin_code,
                                mac_address, serial, pin_code, headers)
